(*
 File          : ast.ml
 Last updated  : 10.27.10

 AST declarations for the LogLog language.
 *)
module Ast =
struct
  module IntMap =
    Hashtbl.Make
    (struct type t = int
      let equal = (==)
      let hash v = v
    end)

  type varname = int
  type typename = int
  type fieldname = int

  type typerep = IntType
    | BoolType
    | StringType
    | ListType of typerep
    | RelationType of typename
    | RecordType of typename
    | UnitType
    | UnknownType
  type typedvar = varname * typerep

  (* Statements and expressions *)
  type binop = Plus | Minus | Times | Divide
  type boolop = Equiv | Nequiv | Or | And | Lt | Gt
  type primexp =
      Integer of int
    | Bool of bool
    | String of string
  (* Core AST expressions. *)
  and exp =
  | PrimExp of primexp
  | VarExp of var
  | Lambda of typedvar * exp ref
  | FunAppExp of string * (exp ref list)
  | Arithop of binop * exp ref * exp ref
  | Boolop of boolop * exp ref * exp ref
  | LetExp of varname * typerep * exp ref * exp ref
  | FoldExp of string        (* fold function *)
             * exp ref      (* initial value *)
             * varname      (* expression to fold over *)
  | CondExp of exp ref * exp ref * exp ref
  | Symbolic of varname (* new variable. *) * exp ref (* constraint *)
  and cdecl =
  | CExp of exp ref
  | ConstraintForall of varname        (** variable *)
                      * var
                      * exp ref
                      * cdecl
  | ConstraintForMatch of varname * varname   (* First record. *)
                        * varname * varname   (* Second record. *)
                        * exp ref * exp ref
  | ConstraintExistsMatch of varname * varname
                           * varname * varname
                           * var * exp ref * exp ref
  and var = Var of varname
          | RecordIndexVar of exp ref   (* Primary key value *)
                            * varname   (* Record name *)
                            * fieldname (* Field name *)
          | FieldIndexVar of varname * fieldname
          | HiddenVarAccess of var

  (* Some constants. *)
  let ref_true = ref (PrimExp (Bool true))
  let ref_false = ref (PrimExp (Bool false))
  let ref_zero = ref (PrimExp (Integer 0))

  (**************************)
  (* Top-level declarations *)
  (**************************)
  (* User-defined types. *)
  type userty = { dataname : varname
                ; fieldinfo : (fieldname * typerep) list }

  (* Function declaration *)
  (* TODO: Add type here. *)
  type fundecl = { funargs : (varname * typerep) list;
                   funbody : exp ref }

  type datarhs =
      RhsExp of exp
    | Concretize of int * exp

  type topdecl =
      TypeDecl of userty * varname * cdecl list
    | FunDecl of typedvar * fundecl
    | AssignDecl of varname * typerep * datarhs
end
