(**
    File          : ast.ml
    Last updated  : 11.10.10
    
    AST declarations for the LogLog language.
*)
module Ast :
sig
  module IntMap : Hashtbl.S with type key = int

  type typename = int
  type varname = int
  type fieldname = int

  (** Type representations. *)
  type typerep =
      IntType
    | BoolType
    | StringType
    | ListType of typerep
    | RelationType of typename
    | RecordType of typename
    | UnitType
    | UnknownType
  type typedvar = varname * typerep

  (** LogLog basics. *)
  type binop = Plus | Minus | Times | Divide
  type boolop = Equiv | Nequiv | Or | And | Lt | Gt
  type primexp = 
      Integer of int
    | Bool of bool
    | String of string

  (** Expressions. *)
  and exp =
  | PrimExp of primexp
  | VarExp of var
  | Lambda of typedvar * exp ref
  | FunAppExp of string * exp ref list
  | Arithop of binop * exp ref * exp ref
  | Boolop of boolop * exp ref * exp ref
  | LetExp of varname * typerep * exp ref * exp ref
  | FoldExp of string * exp ref * varname
  | CondExp of exp ref * exp ref * exp ref
  | Symbolic of varname (* Variable name. *)
              * exp ref (* constraint *)

  and cdecl =
  | CExp of exp ref
  | ConstraintForall of varname         (** bound variable *)
                      * var             (** what we're matching in *)
                      * exp ref
                      * cdecl
  | ConstraintForMatch of varname * varname
                        * varname * varname
                        * exp ref * exp ref
  | ConstraintExistsMatch of varname * varname
                           * varname * varname
                           * var
                           * exp ref * exp ref
  and var =
    Var of varname
  | RecordIndexVar of exp ref (* Primary key value. *)
                    * varname * fieldname
  | FieldIndexVar of varname * fieldname
  | HiddenVarAccess of var

  val ref_true : exp ref
  val ref_false : exp ref

  (** User-defined types can only be relation/record types. *)
  type userty = { dataname : varname
                ; fieldinfo : (fieldname * typerep) list }

  (* Function declaration *)
  type fundecl = { funargs : (varname * typerep) list;
                   funbody : exp ref }

  (** Things that can appear on the right-hand side of a top-level data
    * declaration. *)
  type datarhs =
      RhsExp of exp
    | Concretize of int * exp

  (** Top-level declarations. *)
  type topdecl =
      TypeDecl of userty * varname * cdecl list
      (** Type declaration contains description of record type with additional
        * assumptions/assertions. *)
    | FunDecl of typedvar * fundecl
      (** Procedure declaration. *)
    | AssignDecl of varname * typerep * datarhs
end
