(**
  Module      : AstBackend
  Maintainer  : jeanyang@csail.mit.edu
  Updated     : 2.9.10

 *)
open Ast
open AstUtils
open BackendInterface
open Error

module AstBackend =
struct
  let error str =
    raise (Unexpected ("AstBackend: " ^ str))

  let get_arithop op : BackendInterface.arithop =
    match op with
      | Ast.Plus -> (0 : BackendInterface.arithop)
      | Ast.Minus -> (1 : BackendInterface.arithop)
      | Ast.Times -> (2 : BackendInterface.arithop)
      | Ast.Divide -> (3 : BackendInterface.arithop)
  (*    | Ast.Div -> 3 *)
  let get_boolop op : BackendInterface.boolop =
    match op with
       | Ast.And -> 0
       | Ast.Or -> 1
       | Ast.Lt -> 3
       | Ast.Gt -> 5
       | Ast.Equiv -> 7
       | Ast.Nequiv -> 8

  (* This function takes an AST expression and stores it in the back-end.
   * We assume *)
  let rec mkAstExpNode (exp : Ast.exp ref) : unit =
    match !exp with
      | Ast.PrimExp pe ->
          (match pe with
             | Ast.Integer i ->
                 debug_print 3
                   ("AstBackend: Creating integer EID in backend of " ^
                    string_of_int i ^ ".");
                 BackendInterface.createAtomEid (BackendInterface.Int i)
             | Ast.Bool b ->
                 BackendInterface.createAtomEid (BackendInterface.Bool b)
             | Ast.String s -> 
                 BackendInterface.createAtomEid (BackendInterface.String s))
      | Ast.VarExp v ->
          error "Var in back-end"
        (* TODO: Where do we store lambdas then? *)
      | Ast.Lambda (_, _) ->
          error "Trying to store lambda."
      | Ast.FunAppExp (_, _) ->
          error "Trying to store unapplied function."
      | Ast.Arithop (op, exp1, exp2) ->
          let (arithop : BackendInterface.arithop) = get_arithop op in
          (mkAstExpNode exp1;
           mkAstExpNode exp2;
           BackendInterface.basicArithop arithop)
      | Ast.Boolop (op, exp1, exp2) ->
          let (boolop : BackendInterface.boolop) = get_boolop op in
          (mkAstExpNode exp1;
           mkAstExpNode exp2;
           BackendInterface.basicBoolop boolop)
      | Ast.LetExp (_, _, _, _) ->
          error "Trying to store let expression."
      | Ast.FoldExp (_, _, _) ->
          error "Trying to store fold expression."
      | Ast.CondExp (cond, t, f) ->
          (mkAstExpNode cond;
           mkAstExpNode t;
           mkAstExpNode f;
           BackendInterface.createCondEid ())
      | Ast.Symbolic (var, _) ->
          debug_print 3
            ("AstBackend: Creating symbolic EID in backend for " ^
             string_of_int var ^ ".");
          BackendInterface.createAtomEid (BackendInterface.Sym var)
end
