(*
 File         : astUtils.ml
 Last updated : 10.7.09
 Maintainer   : jeanyang@csail.mit.edu
 *)
open Ast.Ast
open AstBackend
open BackendInterface
open Error

(************)
(* Equality *)
(************)
module AstEq =
struct
  let var_compare = compare
  let type_compare = compare
  let rec eq_types t1 t2 = match t1, t2 with
    | IntType, IntType -> true
    | BoolType, BoolType -> true
    | StringType, StringType -> true
    | (RecordType r1, RecordType r2
         | RelationType r1, RelationType r2) -> type_compare r1 r2 == 0
    | UnknownType, _ -> true
    | _, UnknownType -> true
    | _, _ -> false

  let rec eq_lists eq_fun lst1 lst2 = match lst1, lst2 with
    | [], [] -> true
    | x::xs, y::ys -> (eq_fun x y) && (eq_lists eq_fun xs ys)
    | _, _ -> false
  let eq_primexps p1 p2 : bool =
    (* TODO *)
    match p1, p2 with
    | Integer i1, Integer i2 -> i1 == i2
    | Bool b1, Bool b2 -> b1 == b2
    | _, _ -> false

  let eq_binops b1 b2 = match b1, b2 with
    | Plus, Plus -> true
    | Minus, Minus -> true
    | Times, Times -> true
    | Divide, Divide -> true
    | _, _ -> false

  let eq_boolops b1 b2 = match b1, b2 with
    | Equiv, Equiv -> true
    | Nequiv, Nequiv -> true
    | Or, Or -> true
    | And, And -> true
    | Lt, Lt -> true
    | Gt, Gt -> true
    | _, _ -> false

  (*
   * Function that determines whether two variables are equal: returns 0 if
   * they are equal, -1 if v2 is larger, and 1 if v1 is larger.
   *)
  let rec eq_vars (v1 : var) (v2 : var) : int =
    match v1, v2 with
    | Var v1, Var v2 -> compare v1 v2
    | RecordIndexVar (p1, r1, f1), RecordIndexVar (p2, r2, f2) ->
        let pc = compare p1 p2 in
          if pc != 0
          then pc
          else
            let rc = var_compare r1 r2 in
              if rc != 0
              then rc
              else let fc = compare f1 f2 in
                if fc != 0
                then fc
                else 0
    | FieldIndexVar (r1, f1), FieldIndexVar (r2, f2) ->
        let rc = var_compare r1 r2 in
          if rc != 0
          then rc
          else compare f1 f2
    (* Arbitrary ordering on these: Vars come before RecordIndexVars, which
     * come before FieldIndexVars. *)
    | Var _, RecordIndexVar (_, _, _) -> -1
    | Var _, FieldIndexVar (_, _) -> -1
    | RecordIndexVar (_, _, _), Var _ -> 1
    | RecordIndexVar (_, _, _), FieldIndexVar (_, _) -> -1
    | FieldIndexVar (_, _), Var _ -> 1
    | FieldIndexVar (_, _), RecordIndexVar (_, _, _) -> 1

  (* Expression equality. *)
  and eq_exps (e1 : exp ref) (e2 : exp ref) : bool =
    match !e1, !e2 with
    (* | ExpId e1, ExpId e2 -> e1 == e2
    | ExpId e1, PrimExp (Integer i) -> BackendInterface.eqExpToVal i e1
    | ExpId e1, _ ->
        debug_print 2
          "AstEq.eq_exps: Comparing an expression ID with an expression.";
        BackendInterface.eqExps e1 (AstBackend.mkAstExpNode e2)
    | PrimExp (Integer i), ExpId e -> eq_exps e2 e1
    | _, ExpId _ -> eq_exps e2 e1 *)
    | PrimExp p1, PrimExp p2 -> eq_primexps p1 p2
    | VarExp v1, VarExp v2 -> eq_vars v1 v2 == 0
    | FunAppExp (f1, args1), FunAppExp (f2, args2) ->
        (var_compare f1 f2 == 0) &&
        (eq_lists (fun x -> fun y -> eq_exps x y) args1 args2)
    | Arithop (b1, e1_1, e2_1), Arithop (b2, e1_2, e2_2) ->
        (eq_binops b1 b2) && (eq_exps e1_1 e2_1) && (eq_exps e1_2 e2_2)
    | Boolop (b1, e1_1, e2_1), Boolop (b2, e1_2, e2_2) ->
        (eq_boolops b1 b2) && (eq_exps e1_1 e2_1) && (eq_exps e1_2 e2_2)
    | LetExp (v1, t1, a1, body1), LetExp (v2, t2, a2, body2) ->
        (var_compare v1 v2 == 0) &&
        (eq_types t1 t2) && (eq_exps a1 a2) &&
        (eq_exps body1 body2)
    | FoldExp (v1, init1, var1), FoldExp (v2, init2, var2) ->
        (var_compare v1 v2 == 0) && 
        (eq_exps init2 init2) && (var1 == var2)
    | CondExp (c1, t1, f1), CondExp (c2, t2, f2) ->
        (eq_exps c1 c2) && (eq_exps t1 t2) && (eq_exps f1 f2)
    | Symbolic (v1, _), Symbolic (v2, _) -> v1 == v2
    | _, _ -> false

  let esimplify (e : exp ref) : exp ref = e
end

module Var =
struct
  type t = var;;
  let compare = AstEq.eq_vars
  let rec v_str var = match var with
    | Var v -> string_of_int v
    | RecordIndexVar (pkey, r, f) ->
        let pkey_as_str =
          match !pkey with
            | Ast.Ast.PrimExp e ->
                (match e with
                   | Ast.Ast.Integer i -> string_of_int i
                   | _ -> raise (Unexpected "non-int"))
            | Ast.Ast.VarExp v -> v_str v
            | _ -> raise (Unexpected "non-primexp") in
        String.concat "$" [pkey_as_str; string_of_int r; string_of_int f]
    | FieldIndexVar (r, f) -> string_of_int r ^ "." ^ string_of_int f
    | HiddenVarAccess v -> "?" ^ v_str v
  (* Converting from a string, possibly an encoded record index, to a
   * structured AST variable. *)
  let str2var str =
    try int_of_string (String.sub str 1 ((String.length str) - 1))
    with
        | Failure _ -> raise (InternalError "str2var failed")
        | _ -> raise (Unexpected "?")
end
module VarMap = Map.Make (Var) ;;
module VarSet = Set.Make (struct
                            type t = var
                            let compare = AstEq.eq_vars
                          end)

(************)
module AstPrint =
struct
  let rec exp_depth e =
    match !e with
      | PrimExp _ -> 1
      | VarExp _ -> 1
      | FunAppExp (_, _) -> 1
      | Arithop (_, e1, e2) -> max (exp_depth e1) (exp_depth e2)
      | Boolop (_, e1, e2) -> max (exp_depth e1) (exp_depth e2)
      | Lambda (_, _) -> raise (Unexpected "AstPrint.exp_depth: Lambda")
      | FoldExp (_, _, _) ->
          raise (Unexpected "AstPrint.exp_depth: Fold")
      | LetExp (_, _, e1, e2) | CondExp (_, e1, e2) ->
          let d1 = exp_depth e1 in
          let d2 = exp_depth e2 in
          let d' = if d1 > d2 then d1 else d2 in
            1 + d'
      | Symbolic _ -> 1

  let parens exp = String.concat "" ["("; exp; ")"]
  
  (* Serialize a type. *)
  let rec pickle_type (ty : typerep) : string =
    match ty with
      | IntType -> "int"
      | BoolType -> "bool"
      | UnitType -> "unit"
      | StringType -> "string"
      | RelationType rty -> string_of_int rty ^ " relation"
      | RecordType rty -> string_of_int rty ^ " record"
      | UnknownType -> "unknown"
  
  let pickle_boolop op =
    match op with
      | Equiv -> "=="
      | Nequiv -> "!="
      | Or -> "||"
      | And -> "&&"
      | Lt -> "<"
      | Gt -> ">"

  let pickle_binop op = match op with
    | Plus -> "+"
    | Minus -> "-"
    | Times -> "*"
    | Divide -> "/"

  let rec pickle_var (v : var) = match v with
    | Var v_name -> "v" ^ string_of_int v_name
    | RecordIndexVar (p_key, r_name, f_name) ->
        string_of_int r_name ^ "." ^ string_of_int f_name ^
        "[" ^ pickle_exp p_key ^ "]"
    | FieldIndexVar (r_name, f_name) ->
        string_of_int r_name ^ "." ^ string_of_int f_name

  (* Expression equality. *)
  and pickle_exp (e : exp ref) : string =
    match !e with
      | PrimExp p ->
          (match p with
             | Integer i -> string_of_int i
             | Bool b -> string_of_bool b
             | String s -> s)
      | VarExp v -> Var.v_str v
      | Lambda (_, _) -> raise (Unexpected "AstPrint.pickle_exp: lambda")
      | FunAppExp (f, args) ->
          let arg_str =
            String.concat ", " (List.map pickle_exp args) in
          f ^ " (" ^ arg_str ^ ")"
      | Arithop (op, e1, e2) -> 
          (pickle_exp e1) ^ " " ^ (pickle_binop op) ^ " " ^ (pickle_exp e2)
      | Boolop (op, e1, e2) ->
          (pickle_exp e1) ^ " " ^ (pickle_boolop op) ^ " " ^ (pickle_exp e2)
      | LetExp (_, _, _, _) -> "let"
      | FoldExp (_, _, _) -> "fold"
      | CondExp (ce, te, fe) ->
          "(if " ^ parens (pickle_exp ce) ^
          " then (" ^ pickle_exp te ^ ")" ^ 
          " else (" ^ pickle_exp fe ^ "))"
      | Symbolic (v, _) -> "symbolic " ^ string_of_int v

  let print_topdecls tdecls =
    let print_topdecl _accum tdecl =
      match tdecl with
        | TypeDecl (t, v, c's) ->
            print_endline
              ("typedecl with " ^ string_of_int (List.length c's) ^ " constraints")
        | FunDecl ((fname, fty), { funargs = args; funbody = fbody }) ->
          let print_arg (vname, _ty) = string_of_int vname in
          let args_str =
            String.concat ", " (List.map print_arg args) in
          print_endline ("function " ^ string_of_int fname ^
                         "(" ^ args_str ^ ")" ^
                         " = \n" ^ pickle_exp fbody)
        | AssignDecl (v, ty, data) ->
            let data_str =
              match data with
                | RhsExp e -> pickle_exp (ref e)
                | Concretize (_ctxt, e) -> "concretize " ^ pickle_exp (ref e) in
            print_endline ("v" ^ string_of_int v ^ " := " ^ data_str) in
    List.fold_left print_topdecl () tdecls
end
