#include <fstream>

#include "backend_defs.h"
#include "BackendEnv.h"

using namespace std;

/**
 * Construct a back-end environment.
 */
BackendEnv::BackendEnv() {
#ifdef __DEBUG_1
  cout << "BackendEnv::BackendEnv(): Constructor called." << endl;
#endif
  // cout << "BackendEnv::BackendEnv(): Yices version " << yices_version() << endl;

  /* Create a new Yices context. */
  __yctxt = yices_mk_context ();

  __dag = new ExpDAG();
  __dag_creator = new ExpDAGCreator(__dag);

  for (int i = 0; i < 50; ++i) {
	  __named_nodes.push_back(EnvNode(0));
  }

  /* TODO: Coordinate front-end variables with back-end variables. */
  __sym_count = 0;
}

/**
 * DAG functions.
 */
void BackendEnv::new_node(
    ExpNode* mother, ExpNode* father, ExpNode* newnode) {
	ExpNode* opt_node = __dag_creator->new_node(mother, father, newnode);
	push_node(EnvNode(opt_node));
#ifdef __DEBUG_1
	cout << "stack size " << __cur_nodes.size() + 1 << " -> ";
	cout << __cur_nodes.size() << endl;
#endif
}

int BackendEnv::stack_size() {
	return __cur_nodes.size();
}

void BackendEnv::push_node(EnvNode node) {
	if (EXPNODE == node.nodetype) {
		node.get_node()->inc_ref_count();
	}
	__cur_nodes.push(node);
}

EnvNode BackendEnv::pop_node() {
#ifdef __TEST
	Assert(__cur_nodes.size() > 0, "Nodes should contain something to pop");
#endif
  EnvNode cur = __cur_nodes.top();
  __cur_nodes.pop();
  if (EXPNODE == cur.nodetype) {
	  cur.get_node()->dec_ref_count();
  }
#ifdef __DEBUG_1
	cout << "BackendEnv::pop_node: " << __cur_nodes.size() + 1
  		 << "->" << __cur_nodes.size() << endl;
#endif
  return cur;
}

EnvNode BackendEnv::peek_node() {
#ifdef __DEBUG_1
	Assert(__cur_nodes.size() > 0, "Cur nodes needs to have elements to peek");
#endif
	return __cur_nodes.top();
}

EnvNode BackendEnv::getExpByName(vname name) {
#ifdef __DEBUG_1
	cout << "BackendEnv::getExpByName: " << name << endl;
#endif
	return __named_nodes[name];
}

yices_var_decl BackendEnv::get_yices_var(vname name) {
  /* Update the Yices environment. */
  yices_var_decl v;
  // Make variable name.
  stringstream str;
  str << "v" << name;  
  string cstr = str.str();

  v = yices_get_var_decl_from_name(__yctxt, (char*)cstr.c_str()); 
  if (v == 0) {
#ifdef __DEBUG_1
    cout << "Declaring var " << name << endl;
#endif
    // Get type.
    yices_type intty = yices_mk_type(__yctxt, "int");

    // Make variable declaration.
    v = yices_mk_var_decl(__yctxt, (char*)cstr.c_str(), intty);
  }
  return v;
}

/**
 * The reference count doesn't change between we're transferring a node from
 * being on the stack to having a variable binding.
 */
void BackendEnv::setExpByName(vname name) {
#ifdef __DEBUG_1
  cout << "BackendEnv::setExpByName: " << name << endl;
#endif
	EnvNode node = pop_node();

  /* Reference counting book-keeping. */
	if (EXPNODE == node.nodetype) {
		ExpNode* cur = node.get_node();
		if (cur->getType() != ExpNode::LIST) {
			cur =__dag_creator->new_node(cur->get_mother(), cur->get_father(), cur);
			node.set_node(cur);
		}
		cur->inc_ref_count();
	}

  /* Update our map. */
	__named_nodes[name] = node;
}

bool BackendEnv::hasExpByName(vname name) {
	// Assert(false, "hasExpByName unimplemented");
	return (name < __named_nodes.size());
}

void Tokenize(const string& str, vector<string>& tokens,
		const string& delimiters = " ") {
	// Skip delimiters at beginning.
	string::size_type lastPos = str.find_first_not_of(delimiters, 0);
	// Find first "non-delimiter".
	string::size_type pos = str.find_first_of(delimiters, lastPos);

	while (string::npos != pos || string::npos != lastPos) {
		// Found a token, add it to the vector.
		tokens.push_back(str.substr(lastPos, pos - lastPos));
		// Skip delimiters.  Note the "not_of"
		lastPos = str.find_first_not_of(delimiters, pos);
		// Find next "non-delimiter"
		pos = str.find_first_of(delimiters, lastPos);
	}
}

void BackendEnv::listHasElemHelper(ExpNode* lst, ExpNode* elem) {
  /* If we have a list, then we just return whether the list has the
   * element in question. */
  if (ExpNode::LIST == lst->type) {
    ListNode* lstn = dynamic_cast<ListNode*>(lst);
    /* TODO: Change this to be optimized too. */
    push_node(lstn->hasElem(elem));
  /* If we have a ternary node, then we evaluate each side of it. */
  } else if (ExpNode::TERNARY == lst->type) {
    TernaryNode* curlst = dynamic_cast<TernaryNode*>(lst);

    listHasElemHelper(curlst->get_mother(), elem);
    EnvNode llst = pop_node();
    listHasElemHelper(curlst->get_father(), elem);
    EnvNode rlst = pop_node();

    ExpNode* left =
      (CONSTVAL == llst.nodetype) ?
        new ConstNode(llst.get_val()) : llst.get_node();
    ExpNode* right =
      (CONSTVAL == rlst.nodetype) ?
        new ConstNode(rlst.get_val()) : rlst.get_node();

    TernaryNode* new_tn = new TernaryNode(curlst->get_guard());
    new_node(left, right, new_tn);
  } else {
    push_node(EnvNode(false));
  }
}

/* Pushes a boolean node that says whether the list has an element. */
void BackendEnv::listHasElem() {
  EnvNode elem = pop_node();
  EnvNode lst = pop_node();

#ifdef __DEBUG_1
  Assert(EXPNODE == lst.nodetype, "list must be exp");
#endif

  ExpNode* elemn;
  if (EXPNODE == elem.nodetype) {
    elemn = elem.get_node();
  } else {
    elemn = new ConstNode(elem.get_val());
  }

#ifdef __DEBUG_1
  cout << "BackendEnv::listHasElem: " << lst.get_node()->lprint() << endl;
  cout << elemn->lprint() << endl;
#endif
  listHasElemHelper(lst.get_node(), elemn);
}

void BackendEnv::readRelationNode(size_t numfields, const char* file_name) {
#ifdef __DEBUG_1
	cout << "ListNode::ListNode: Reading from file " << file_name
		 << endl;
#endif

#ifdef __DEBUG_1
	int initial_stack_size = stack_size();
#endif

	ListNode* cur = new ListNode();

	ifstream file;
	file.open(file_name);
	Assert(file.is_open(), "Was not able to open file.");

	while (!file.eof()) {
		/* Get current line. */
		string curline;
		getline(file, curline);
#ifdef __DEBUG_1
		cout << "ListNode::ListNode: Read line " << curline << endl;
#endif

		/* Scan current line. */
		vector<string> tokens;
		Tokenize(curline, tokens);

		RecordNode* cur_record = new RecordNode(numfields);

		if (tokens.size() == numfields) {
			vector<string>::iterator token_it;
			int field = 0;
			for (token_it = tokens.begin(); token_it != tokens.end(); ++token_it) {
				bool is_digit = true;
				
        string::iterator pch;
				for (pch = token_it->begin(); pch != token_it->end(); ++pch) {
					if (!isdigit(*pch)) {
						is_digit = false;
					}
				}

				if (is_digit) {
					int cur_val = atoi(token_it->c_str());
					cur_record->setValue(field, EnvNode(cur_val));
        } else if (token_it->find(",") != string::npos) {
          vector<string> list_tokens;
          Tokenize(*token_it, list_tokens, ",");

          ListNode* listn = new ListNode();

          vector<string>::iterator list_token_it;
          for (list_token_it = list_tokens.begin();
               list_token_it != list_tokens.end(); ++list_token_it) {
            /* Create a new list node with the elements. */
            if ((list_token_it->size() > 0) && ((*list_token_it)[0] == '?')) {
              /* Get protected integer value. */
              string& cur_str = list_token_it->erase(0, 1);
              int cur_val = atoi(cur_str.c_str());

              /* Create a symbolic node. */
              ExpNode* cnode = new SymNode(__sym_count++);
              ((ConstNode*)cnode)->setVal(cur_val);
#ifdef __DEBUG_1
              Assert(cnode->sym(), "node should be symbolic");
#endif
              cur->merge_sym(cnode);
              EnvNode enode = EnvNode(cnode);

              cur_record->setValue(field, enode);

              /* Map symbolic guy in environment. */
              cnode->inc_ref_count();
              __sym_nodes.push_back(cnode);

              listn->addExpNode(cnode);
            } else {
              ConstNode* list_elt = new ConstNode(atoi(list_token_it->c_str()));
              list_elt->inc_ref_count();
              listn->addExpNode(list_elt);
            }
          }

#ifdef __DEBUG_1
          cout << "MADE LIST: " << listn->lprint() << endl;
#endif
          cur_record->setValue(field, EnvNode(listn));
        /* Case when we have a hidden token. */
				} else if ((token_it->size() > 0) && ((*token_it)[0] == '?')) {
          /* Otherwise see if it's a protected symbolic value and set values
           * according if so. */

          /* Get protected integer value. */
				  string& cur_str = token_it->erase(0, 1);
          int cur_val = atoi(cur_str.c_str());

          /* Create a symbolic node. */
          ExpNode* cnode = new SymNode(__sym_count++);
          ((ConstNode*)cnode)->setVal(cur_val);
#ifdef __DEBUG_1
          Assert(cnode->sym(), "node should be symbolic");
#endif
          cur->merge_sym(cnode);
          EnvNode enode = EnvNode(cnode);
          
          cur_record->setValue(field, enode);

          /* Map symbolic guy in environment. */
          cnode->inc_ref_count();
          __sym_nodes.push_back(cnode);
        } else {
          /* Otherwise create a symbolic node. */
#ifdef __DEBUG_1
					cout << "ListNode::ListNode: Adding symbolic "
						 << "value." << endl;
#endif
					ExpNode* cnode = new SymNode(__sym_count++);
					cur->merge_sym(cnode);
					EnvNode enode = EnvNode(cnode);
					cur_record->setValue(field, enode);
					/* Map symbolic guy in environment. */
					cnode->inc_ref_count();
					__sym_nodes.push_back(cnode);
				}
				++field;
			} // rof token_it
			cur->addExpNode(cur_record);
		}
	} // elihw reading file
	file.close();

	// cur = (ListNode*)__dag_creator->new_node(NULL, NULL, cur);
#ifdef __DEBUG_1
	cout << "Made relation " << cur->lprint() << endl;
#endif
	push_node(EnvNode(cur));
#ifdef __DEBUG_1
	Assert( stack_size() == initial_stack_size + 1
		  , "Stack size should have only increased by 1." );
	cout << "FINISHED READING LIST" << endl;
#endif
}

SymNode* BackendEnv::mkSymNode() {
  SymNode* sn = new SymNode(__sym_count++);
  __sym_nodes.push_back(sn);
  return sn;
}

/**
 * Node constructors.
 */

/**
 *  Assertions.
 */
set<int>& BackendEnv::getVars() {
#ifdef __DEBUG_1
  cout << "BackendEnv::getVars: Returning " << __constraint_vars.size()
	     << " variables." << endl;
#endif
  return __constraint_vars;
}

void BackendEnv::addAssertion(ExpNode* aexp) {
  __assertions.insert(aexp);
}

void BackendEnv::writeAssertionsToFile(string& filename) {
#ifdef __DEBUG_1
  cout << "BackendEnv::writeAssertionsToFile: Writing to " << filename << endl;
#endif
  /* Open file for writing. */
  ofstream file;
  file.open(filename.c_str());

  /* Write variable declarations to file. */
  set<int>::iterator var_it = __constraint_vars.begin();
  for (; var_it != __constraint_vars.end(); ++var_it) {
    file << "(define v" << *var_it << ":: real)\n";
  }

  /* Write assertions themselves to file. */
  set<ExpNode*>::iterator assert_it = __assertions.begin();
  for (; assert_it != __assertions.end(); ++assert_it) {
		ExpNode* cur_node = *assert_it;
		cur_node->ytype = ExpNode::BOOL;
		file << "( assert " << cur_node->get_yicesstr() << " )" << endl;
	}
  file << "(check)" << endl;

  file.close();
}

/* TODO: Don't worrya bout the context yet... */
void BackendEnv::resolveWithContext(int ctxt) {
#ifdef __DEBUG_1
  cout << "BackendEnv::resolveWithContext: " << ctxt << endl;
#endif
  EnvNode en = pop_node();

  if (CONSTVAL == en.nodetype) {
    push_node(en);
  } else {
    /* Just in case, go through and declare all variables. */
    set<int>::iterator var_it = __constraint_vars.begin();
    for (; var_it != __constraint_vars.end(); ++var_it) {
      yices_var_decl vdecl = get_yices_var(*var_it);
    }

    /* Push a context. */
    yices_push(__yctxt);

    /* Make an equality binding the context variable in the yices context. */
    EnvNode ctxt_record = getExpByName(1);
#ifdef __DEBUG_1
    Assert(EXPNODE == ctxt_record.nodetype, "Stack node should be exp node.");
#endif
    ExpNode* expnode = ctxt_record.get_node();
#ifdef __DEBUG_1
    Assert(ExpNode::RECORD == expnode->type, "Node should be a record");
#endif
    RecordNode* rnode = (RecordNode*)expnode;
    EnvNode cval = rnode->lookupField(0);
#ifdef __DEBUG_1
    Assert(EXPNODE == cval.nodetype, "Context node should be symbolic.");
#endif
    ExpNode* symctxtnode = cval.get_node();
    yices_expr sym_expr = symctxtnode->to_yices(__yctxt);
    yices_expr ctxt_num = yices_mk_num(__yctxt, ctxt);
    yices_expr ctxt_eq = yices_mk_eq(__yctxt, sym_expr, ctxt_num);
    yices_assert(__yctxt, ctxt_eq);

    /* Check and get the model. */
    Assert(yices_check(__yctxt), "Context is not satisfiable.");    
    yices_model model = yices_get_model(__yctxt);

    /* set<int>::iterator */ var_it = __constraint_vars.begin();
    for (; var_it != __constraint_vars.end(); ++var_it) {
      yices_var_decl vdecl = get_yices_var(*var_it);
      Assert(vdecl != 0, "Var decl not empty.");
      long intval;
      int vflag = yices_get_int_value(model, vdecl, &intval);
      Assert(vflag == 1, "Got int value for variable");
#ifdef __DEBUG_1
      cout << "Mapping variable " << *var_it << " = " << intval << endl;
#endif

      /* Remap symbolic variables. */
      ExpNode* cur = getSymByName(*var_it);
#ifdef __DEBUG_1
      Assert( cur->getType() == ExpNode::SYM,
          "BackendIface::concretizeEnvVal: Given sym node." );
#endif
      ConstNode* cur_const = (ConstNode*)cur;
      cur_const->type = ExpNode::CONST;
      cur_const->set_sym(false);
      cur_const->setVal((int)intval);
    }

    yices_pop(__yctxt);

    /* Simplify? */
    push_node(en);
  }
}

/**
 * Other.
 */
void BackendEnv::printEnv() {
	cout << "BackendEnv::printEnv: Printing environment..." << endl;
	vector<EnvNode>::iterator env_it;
	int count = 0;
	for (env_it = __named_nodes.begin(); env_it != __named_nodes.end(); ++env_it) {
			cout << count << " |-> ";
		if (EXPNODE == env_it->nodetype) {
			cout << (env_it->get_node()->lprint()) << endl;
		} else {
			cout << (env_it->get_val()) << endl;
		}
		++ count;
	}
}

void BackendEnv::clear() {
  __named_nodes.clear();

  while (!__cur_nodes.empty()) {
	  __cur_nodes.pop();
  }

  __assertions.clear();
  __constraint_vars.clear();
  __dag->clear();
}
