/**
 * In this file we define the different parts of the backend environment.  We
 * have the following:
 * - Nodes, which can be:
 *      - Arithmetic operations.
 *      - Boolean operations.
 *      - Symbolic expressions.
 * - A mapping from expression ID's to
 * - 
 *
 * Questions:
 * - Should relations be different kinds of expressions than expressions?
 */
#ifndef __BACKEND_ENV_H
#define __BACKEND_ENV_H

#include <map>
#include <set>
#include <stack>
#include <yices_c.h>

#include "backend_defs.h"
#include "BackendTypes.h"
#include "ExpDAG.h"
#include "ExpDAGCreator.h"
#include "ExpressionNodes.h"

using namespace std;

/**
 * Backend environment.
 */
class BackendEnv {
  private:
    /* The backend environment consists of the following:
     *  - __named_nodes: a mapping from front-end variable names to pointers to
     *      expression nodes.
     *  - __assertions: a mapping from expression ID's of the assertions.
     */
    vector<EnvNode> __named_nodes;
    vector<ExpNode*> __sym_nodes;
    int __sym_count;
    stack<EnvNode> __cur_nodes;

    set<ExpNode*> __assertions;
    set<int> __constraint_vars;

    ExpDAG* __dag;
    ExpDAGCreator* __dag_creator;

    yices_context __yctxt;

    void listHasElemHelper(ExpNode* lst, ExpNode* elem);
  public:
    yices_context get_context() { return __yctxt; }

    BackendEnv();

    stack<relation_t::iterator> iterators;
    stack<relation_t::iterator> iter_ends;

    void print_stack() {
    	stack<EnvNode> other;
    	cout << "---" << endl;
    	while (__cur_nodes.size() > 0) {
    		EnvNode cur = __cur_nodes.top();
    		if (EXPNODE == cur.nodetype) {
    			cout << cur.get_node()->lprint() << endl;
    		} else {
    			cout << cur.get_val() << endl;
    		}
    		__cur_nodes.pop();
			other.push(cur);
    	}
    	cout << "---" << endl;
    	while (other.size() > 0) {
    		__cur_nodes.push(other.top());
    		other.pop();
    	}
    }

    void new_node(ExpNode* mother, ExpNode* father, ExpNode* node);

    void gc(EnvNode node) {
    	if (EXPNODE == node.nodetype) {
    		__dag_creator->remove_node(node.get_node());
    	}
    }

    int stack_size();
    void push_node(EnvNode node);
    EnvNode pop_node();
    EnvNode peek_node();

    void readRelationNode(size_t num_fields, const char* file_name);
    SymNode* mkSymNode();

    /* Named nodes. */
    /* TODO: Maybe get rid of this too? */
    /* TODO: I think the environment stores these variables to store a mapping
     * from original names to backend internal names.  Figure out if this is
     * true? */
    EnvNode getExpByName(vname name);
    ExpNode* getSymByName(vname name) { return __sym_nodes[name]; }
    void setExpByName(vname name);
    bool hasExpByName(vname name);
    void removeByName(vname name);

    void listHasElem();

    /* Yices. */
    yices_var_decl get_yices_var(vname name);

    /**
     * Assertions.
     */
    set<int>& getVars();
    void addAssertion(ExpNode* aexp);
    void writeAssertionsToFile(string& file);
    void resolveWithContext(int ctxt);

    /* Printing. */
    void printEnv();

    /* Clears the environment. */
    void clear();
};

#endif /* __BACKEND_ENV_H */
