#include "backend_defs.h"
#include "BackendIface.h"
#include "BackendTypes.h"
#include "BasicError.h"
#include "ExpressionNodes.h"


BackendEnv* BackendIface::env = new BackendEnv();

extern "C" void popNode() {
	EnvNode cur = BackendIface::env->pop_node();
	BackendIface::env->gc(cur);
}

extern "C" void dupNode() {
	BackendIface::env->push_node(BackendIface::env->peek_node());
}

extern "C" void pushUnitNode() {
  BackendIface::env->push_node(EnvNode(0));
}

extern "C" void swapNode() {
	EnvNode top = BackendIface::env->pop_node();
	EnvNode second = BackendIface::env->pop_node();
	BackendIface::env->push_node(top);
	BackendIface::env->push_node(second);
}

extern "C" void printEnvExp() {
	EnvNode cur_node = BackendIface::env->pop_node();
	if (EXPNODE == cur_node.nodetype) {
		cur_node.get_node()->printval();
	} else {
		cout << cur_node.get_val() << endl;
	}
	BackendIface::env->gc(cur_node);
}

/**
 * Constant nodes.
 */
extern "C" void createConstNode(int c) {
#ifdef __TEST
  cout << "BackendIface::createConstNode: " << c << endl;
#endif
  BackendIface::env->push_node(EnvNode(c));
}

extern "C" void createStringNode(char* s) {
  BackendIface::env->push_node(EnvNode(new StringNode(string(s))));
}

extern "C" void createBoolNode(bool b) {
  createConstNode(int(b));
}

/* TODO: Figure out what we're going to do with the front-end variable name. */
extern "C" void createSymNode(int varname) {
	SymNode* new_sym_node = new SymNode(varname);
	BackendIface::env->push_node(EnvNode(new_sym_node));
}

extern "C" void createCondNode() {
  EnvNode fnode = BackendIface::env->pop_node();
  EnvNode mnode = BackendIface::env->pop_node();
  EnvNode guard = BackendIface::env->pop_node();

  if (CONSTVAL == guard.nodetype) {
	  if (guard.get_val() == 0) {
		  BackendIface::env->push_node(fnode);
	  } else {
		  BackendIface::env->push_node(mnode);
	  }
  } else {
	  TernaryNode* new_tnode = new TernaryNode(guard.get_node());
	  ExpNode* mother =
			  (EXPNODE == mnode.nodetype) ?
					  mnode.get_node() : new ConstNode(mnode.get_val());
	  ExpNode* father =
			  (EXPNODE == fnode.nodetype) ?
					  fnode.get_node() : new ConstNode(fnode.get_val());
	  new_tnode->set_parents(mother, father);
	  BackendIface::env->push_node(EnvNode(new_tnode));
  }
}

/**
 * Binary operators.
 */
extern "C" void createArithopNode(arithop_t op) {
  EnvNode fnode = BackendIface::env->pop_node();
  EnvNode mnode = BackendIface::env->pop_node();

  ExpNode* cur = NULL;

  if (CONSTVAL == fnode.nodetype && CONSTVAL == mnode.nodetype) {
		int fval = fnode.get_val();
		int mval = mnode.get_val();

		switch (op) {
		case PLUS:
			cur = new ConstNode(mval + fval);
			break;
		case MINUS:
			cur = new ConstNode(mval - fval);
			break;
		case TIMES:
			cur = new ConstNode(mval * fval);
			break;
		case DIV:
			cur = new ConstNode(mval * fval);
			break;
		default:
			cout << "BackendIface::createArithopNode: Got arithmetic operator "
					<< op << "." << endl;
			Assert(false, "Internal error.");
		}
  } else {
		switch (op) {
		case PLUS:
			cur = new PlusNode();
			break;
		case MINUS:
			cur = new PlusNode();
			((PlusNode*) cur)->set_minus(true);
			break;
		case TIMES:
			cur = new TimesNode();
			break;
		case DIV:
			cur = new DivNode();
			break;
		default:
			cout << "BackendIface::createArithopNode: Got arithmetic operator "
					<< op << "." << endl;
			Assert(false, "Internal error.")
			;
		}
		ExpNode* mother =
				(EXPNODE == mnode.nodetype) ?
						mnode.get_node() : new ConstNode(mnode.get_val());
		ExpNode* father =
				(EXPNODE == fnode.nodetype) ?
						fnode.get_node() : new ConstNode(fnode.get_val());
		cur->set_parents(mother, father);
	}
#ifdef __DEBUG_1
	cout << "BackendIface::createArithopNode: Created node " << cur->globalId << endl;
#endif
  BackendIface::env->push_node(EnvNode(cur));
}

extern "C" void createBoolopNode(boolop_t op) {
	EnvNode fnode = BackendIface::env->pop_node();
	EnvNode mnode = BackendIface::env->pop_node();

	if (CONSTVAL == fnode.nodetype && CONSTVAL == mnode.nodetype) {
		int fval = fnode.get_val();
		int mval = mnode.get_val();

		switch (op) {
		case AND:
			BackendIface::env->push_node(EnvNode(mval && fval));
			break;
		case OR:
			BackendIface::env->push_node(EnvNode(mval || fval));
			break;
		case NOT:
			BackendIface::env->push_node(EnvNode(mval ? 0 : 1));
			break;
		case LT:
			BackendIface::env->push_node(EnvNode(mval < fval));
			break;
		case LTE:
			BackendIface::env->push_node(EnvNode(mval <= fval));
			break;
		case GT:
			BackendIface::env->push_node(EnvNode(mval > fval));
			break;
		case GTE:
			BackendIface::env->push_node(EnvNode(mval >= fval));
			break;
		case EQ:
			BackendIface::env->push_node(EnvNode(mval == fval));
			break;
		case NEQ:
			BackendIface::env->push_node(EnvNode(mval != fval));
			break;
		default:
			cout
					<< "BackendIface::createBoolopNode: Received boolean operator "
					<< op << "." << endl;
			Assert(false, "Internal error.")
			;
		}
		return;
	} else if (CONSTVAL == mnode.nodetype) {
		if (op == AND && !(mnode.get_val())) {
			BackendIface::env->gc(fnode);
			BackendIface::env->push_node(EnvNode(0));
			return;
		}

		if (op == OR && mnode.get_val()) {
			BackendIface::env->gc(fnode);
			BackendIface::env->push_node(EnvNode(1));
			return;
		}
	} else if (CONSTVAL == fnode.nodetype) {
		if (op == AND && !(fnode.get_val())) {
			BackendIface::env->gc(mnode);
			BackendIface::env->push_node(EnvNode(0));
			return;
		}

		if (op == OR && fnode.get_val()) {
			BackendIface::env->gc(mnode);
			BackendIface::env->push_node(EnvNode(1));
			return;
		}
	}

	ExpNode* new_node = NULL;
	switch (op) {
	case AND:
		new_node = new AndNode();
		break;
	case OR:
		new_node = new OrNode();
		break;
	case NOT:
		new_node = new NOT_node();
		break;
	case LT:
		new_node = new LtNode();
		break;
	case LTE:
		new_node = new LeNode();
		break;
	case GT:
		new_node = new GtNode();
		break;
	case GTE:
		new_node = new GeNode();
		break;
	case EQ:
		new_node = new EqNode();
		break;
	case NEQ:
		new_node = new NeqNode();
		break;
	default:
		cout << "BackendIface::createBoolopNode: Received boolean operator "
				<< op << "." << endl;
		Assert(false, "Internal error.")
		;
	}
	ExpNode* father =
			(EXPNODE == fnode.nodetype) ?
					fnode.get_node() : new ConstNode(fnode.get_val());
	ExpNode* mother =
			(EXPNODE == mnode.nodetype) ?
					mnode.get_node() : new ConstNode(mnode.get_val());
	new_node->set_parents(mother, father);
#ifdef __TEST
	Assert(new_node != NULL, "Should have initialized new node.");
#endif
	BackendIface::env->push_node(EnvNode(new_node));
}

/**
 * Backend environment values.
 */
extern "C" void bindBackendVal(vname varname) {
  BackendIface::env->setExpByName(varname);
}

extern "C" void lookupBackendVal(vname varname) {
	EnvNode cur = BackendIface::env->getExpByName(varname);
	BackendIface::env->push_node(cur);
	return;
}

extern "C" void lookupBackendValHidden(vname varname) {
  EnvNode cur = BackendIface::env->getExpByName(varname);
  
  /* Swap out node for symbolic value if necessary. */
  EnvNode r = cur;
  if (EXPNODE == cur.nodetype) {
    ExpNode* cur_node = cur.get_node();
    if (ExpNode::SYM == cur_node->type) {
      cout << "Replace symbolic node with constant." << endl;
      r = EnvNode(((ConstNode*)cur_node)->getVal());
    }
  }
  BackendIface::env->push_node(r);
  return;
}

extern "C" void lookupValAsExp(eid_t eid, int* type, int* value) {
	Assert(false, "This is not right");

	ExpNode* cur = NULL; //BackendIface::env->get_node(eid);
	*type = cur->getType();
	if (cur->getType() == ExpNode::CONST) {
		*value = ((ConstNode*)cur)->getVal();
	} else if (cur->getType() == ExpNode::SYM) {
		Assert(false, "Unimplemented sym");
	}
 	return;
}

/*
 * It's the responsibility of this guy to garbage collect the node, because it
 * might have the last remaining pointer to it.
 */
extern "C" int getEnvConcreteBool() {
	EnvNode cur = BackendIface::env->peek_node();
	int retval = -1;
	if (CONSTVAL == cur.nodetype) {
		retval = cur.get_val();
	}
	BackendIface::env->gc(cur);
	return retval;
}

extern "C" int getEnvConcreteInt(bool* concrete) {
	EnvNode cur = BackendIface::env->pop_node();
	if (CONSTVAL == cur.nodetype) {
#ifdef __DEBUG_1
		cout << "BackendIface::getEnvConcreteBool: Returning "
		     << cur.get_val() << endl;
#endif
		*concrete = true;
		return cur.get_val();
	} else {
		BackendIface::env->gc(cur);
		return -1;
	}
}

/**
 * Replace symbolic node with a constant.
 *
 * TODO: All the nodes pointing to this node need to become un-symbolic as well.
 */
extern "C" void concretizeEnvVal(vname vname, int val) {
	/* TODO: This won't work for everything that got a symbolic name in the front end. */
#ifdef __DEBUG_1
	cout << "BackendIface::concretizeEnvVal: Getting expression " << vname << endl;
#endif
	ExpNode* cur = BackendIface::env->getSymByName(vname);
#ifdef __DEBUG_1
	Assert( cur->getType() == ExpNode::SYM,
		    "BackendIface::concretizeEnvVal: Given sym node." );
#endif
	ConstNode* cur_const = (ConstNode*)cur;
	cur_const->type = ExpNode::CONST;
	cur_const->set_sym(false);
	cur_const->setVal(val);
#ifdef __DEBUG_1

#endif
}

/**
 * Relation manipulation.
 */
extern "C" void readExpRelation(int numfields, char* file) {
#ifdef __DEBUG_1
	cout << "BackendIface::readExpRelation: Reading from file " << file
	     << " with " << numfields << " fields." << endl;
#endif
  BackendIface::env->readRelationNode(numfields, file);
}

extern "C" eid_t mkEnvRelation(int name, int numfields) {
#ifdef __DEBUG_1
	cout << "BackendIface::mkEnvRelation: " << name << " with " << numfields
		 << " fields." << endl;
#endif
	ExpNode* new_node = new ListNode();

	BackendIface::env->push_node(EnvNode(new_node));
	BackendIface::env->setExpByName(name);

	return new_node->get_name();
}

extern "C" void bindEnvRelationVal(vname rname, int field) {
	Assert(false, "unimplemented bindEnvRelationVal");
/*
#ifdef __DEBUG_1
  cout << "BackendIface::bindEnvRelationVal: " << rname << ", " << pkey << ", "
	   << field << endl;
#endif
*/
  EnvNode cur_exp = BackendIface::env->getExpByName(rname);
  EnvNode relnode = BackendIface::env->pop_node();

  EnvNode pkey_node = BackendIface::env->pop_node();
  int pkey;

  if (pkey_node.nodetype == CONSTVAL) {
    pkey = pkey_node.get_val();
  } else {
    Assert(false, "Expect a const when getting a primary key");
  }

  ListNode* cur = (ListNode*)(relnode.get_node());
  // cur->addValue(field, cur_exp);
}

extern "C" void lookupEnvRelationVal(vname rname, int field) {
#ifdef __DEBUG_1
  Assert( BackendIface::env->hasExpByName(rname)
        , "Must look up relation bound in environment.");
#endif

  // Get primary key.
  EnvNode pkey_node = BackendIface::env->pop_node();
  EnvNode en = EnvNode(-1);

  ListNode* cur = dynamic_cast<ListNode*>(BackendIface::env->getExpByName(rname).get_node());

  if (pkey_node.nodetype == CONSTVAL) {
    int pkey = pkey_node.get_val();
    en = cur->lookupRelationVal(pkey, field);
  } else if (ExpNode::CONST == pkey_node.get_node()->type) {
    int pkey = (dynamic_cast<ConstNode*>(pkey_node.get_node()))->getVal();
    en = cur->lookupRelationVal(pkey, field);
  } else if (ExpNode:: SYM == pkey_node.get_node()->type) {
    SymNode* sn = dynamic_cast<SymNode*>(pkey_node.get_node());
    en = cur->lookupSymVal(sn, field);
  } else {
    cout << "Node has type: " << pkey_node.get_node()->type << endl;
    Assert(false, "bad");
  }

  BackendIface::env->push_node(en);
}

extern "C" void lookupEnvFieldIndex(int rname, int field) {
#ifdef __DEBUG_1
	cout << "BackendIface::lookupEnvFieldIndex: " << rname << ", " << field
		 << endl;
#endif
	EnvNode cur = BackendIface::env->getExpByName(rname);

#ifdef __DEBUG_1
	Assert( cur.get_node()->getType() == ExpNode::RECORD
			, "Should be looking up field of record." );
#endif
  EnvNode result = ((RecordNode*)(cur.get_node()))->lookupField(field);
  BackendIface::env->push_node(result);
}

extern "C" void lookupEnvFieldIndexHidden(int rname, int field) {
  EnvNode cur = BackendIface::env->getExpByName(rname);

#ifdef __DEBUG_1
  Assert( cur.get_node()->getType() == ExpNode::RECORD
      , "Should be looking up field of record." );
#endif
  EnvNode result = ((RecordNode*)(cur.get_node()))->lookupField(field);

  /* Swap out node for symbolic value if necessary. */
  EnvNode r = result;
  if (EXPNODE == result.nodetype) {
    ExpNode* cur_node = result.get_node();
    if (ExpNode::SYM == cur_node->type) {
      r = EnvNode(((ConstNode*)cur_node)->getVal());
    }
  }
  BackendIface::env->push_node(r);

  return;
}

/**
 * Given an expression ID corresponding to a relation, returns an iterator ID
 * corresponding to a new iterator for that relation.
 */
extern "C" void getRelIterator() {
	EnvNode enode = BackendIface::env->pop_node();
#ifdef __DEBUG_1
	Assert( EXPNODE == enode.nodetype
		  , "BackendIface::getRelIterator: Should have a node on the stack" );
	cout << enode.get_node()->getType() << endl;
	Assert( ExpNode::LIST == enode.get_node()->getType()
			  , "BackendIface::getRelIterator: Should have a relation" );
#endif
	ListNode* relnode = dynamic_cast<ListNode*>(enode.get_node());
	BackendIface::env->iterators.push(relnode->begin());
	BackendIface::env->iter_ends.push(relnode->end());
}

extern "C" void getRelPkeyIterator(int varname) {
	EnvNode pkey = BackendIface::env->pop_node();
	int pkey_val = pkey.get_val();

	EnvNode enode = BackendIface::env->getExpByName(varname);
#ifdef __DEBUG_1
	cout << "BackendIface::getRelPkeyIterator: Getting iterator for "
		 << "primary key" << pkey_val << endl;
	Assert( EXPNODE == enode.nodetype
		  , "BackendIface::getRelPkeyIterator: Should have a node on the stack" );
	Assert( ExpNode::LIST == enode.get_node()->getType()
			  , "BackendIface::getRelPkeyIterator: Should have a relation" );
#endif
	ListNode* relnode = dynamic_cast<ListNode*>(enode.get_node());
	BackendIface::env->iterators.push(relnode->pkey_begin(pkey_val));
	BackendIface::env->iter_ends.push(relnode->pkey_end(pkey_val));
}

/**
 * Gets the next record of the relation at the top of the stack.  When it has
 * run out of records, pops the relation off the stack.
 */
extern "C" bool getNextRecord() {
	relation_it cur_it = BackendIface::env->iterators.top();
	if (cur_it == BackendIface::env->iter_ends.top()) {
		BackendIface::env->iterators.pop();
		BackendIface::env->iter_ends.pop();
		return false;
	} else {
		BackendIface::env->iterators.pop();
		BackendIface::env->push_node(EnvNode(*cur_it));
		BackendIface::env->iterators.push(++cur_it);
		return true;
	}
}

/**
 * Equality.
 */
extern "C" bool eqExpsInEnv() {
	EnvNode e2 = BackendIface::env->pop_node();
	EnvNode e1 = BackendIface::env->pop_node();

	if (e1.nodetype == EXPNODE && e2.nodetype == EXPNODE) {
		ExpNode* e1n = e1.get_node();
		ExpNode* e2n = e2.get_node();

		if (e1n->getType() == ExpNode::LIST &&
			e2n->getType() == ExpNode::LIST) {
			BackendIface::env->gc(e1);
			BackendIface::env->gc(e2);
			return true;
		} else {
			bool retval = ExpNode::eqNodes(e1n, e2n);
			BackendIface::env->gc(e1);
			BackendIface::env->gc(e2);
			return retval;
		}
	} else if (e1.nodetype == CONSTVAL && e2.nodetype == CONSTVAL) {
		return e1.get_val() == e2.get_val();
	}

	return false;
}

extern "C" bool eqExpToValInEnv(int val) {
	Assert(false, "need to reimplement eqExpToValInEnv");
	/*
  ExpNode* e = BackendIface::env->pop_node();
  if (e->getType() == ExpNode::CONST) {
    bool retval = val == ((ConstNode*)e)->getVal();
    BackendIface::env->gc(e);
    return retval;
  } else {
    return false;
  }
  */
}

/**
 * Return whether an expression is symbolic.
 */
extern "C" bool expIsSym() {
	EnvNode e = BackendIface::env->pop_node();
	bool retval = (CONSTVAL == e.nodetype) ? false : e.get_node()->sym();
	BackendIface::env->gc(e);
	return retval;
}

/**
 * Assert an expression.
 */
extern "C" void assertExpEnv() {
	EnvNode exp = BackendIface::env->pop_node();

	if (EXPNODE == exp.nodetype) {
		ExpNode* aexp = exp.get_node();

    /* Assert in Yices context. */
    aexp->ytype = ExpNode::BOOL;
    yices_expr ye = aexp->to_yices(BackendIface::env->get_context());
    yices_assert(BackendIface::env->get_context(), ye);

		BackendIface::env->new_node(aexp->get_mother(), aexp->get_father(),
				aexp);
		exp = BackendIface::env->pop_node();
		if (EXPNODE == exp.nodetype) {
			aexp = exp.get_node();

			if (!(aexp->getType() == ExpNode::CONST)) {
#ifdef __TEST
				Assert(aexp != NULL, "Node we're asserting should not be null.");
				cout << "BackendIface::assertExpEnv: Got node " << aexp->get_name() << endl;
#endif

				/* Only add the assertion if it has a symbolic component. */
				if (aexp->sym()) {
					aexp->inc_ref_count();
					aexp->addVars(BackendIface::env->getVars());
					BackendIface::env->addAssertion(aexp);
					return;
				} else {
					BackendIface::env->gc(EnvNode(aexp));
				}
			} else {
				BackendIface::env->gc(EnvNode(aexp));
			}
		}
	}
}

extern "C" void writeEnvAssertions() {
  string file("constraints.txt");
  BackendIface::env->writeAssertionsToFile(file);
}

extern "C" void resolveWithEnvContext(int ctxt) {
  BackendIface::env->resolveWithContext(ctxt);
}

extern "C" void listElemEnv() {
  BackendIface::env->listHasElem();
}

extern "C" void printEnvVars() {
	BackendIface::env->printEnv();
}

/* Initialize the backend environment. */
extern "C" void initBackendEnv() {
  /* Allocate memory for data structures. */
  BackendIface::init();
#ifdef __DEBUG_1
  Assert(BackendIface::env->stack_size() == 0, "Stack is empty");
  cout << "BackendIface::initBackendEnv: Finished initializing environment."
  	   << endl;
#endif

  /* Make the symbolic context.  1 is the reserved symbolic name for the
     context. */
  RecordNode* rn = new RecordNode(1);
  rn->inc_ref_count();
  BackendIface::env->new_node(NULL, NULL, rn);
  // Pop node from stack?
  ExpNode* sn = BackendIface::env->mkSymNode();
  sn->inc_ref_count();
  rn->setValue(0, EnvNode(sn));

  /* The record is on the stack.  We map the variable 1 (reserved for context
     to it. */
  BackendIface::env->setExpByName(1);
#ifdef __DEBUG_1
  EnvNode en = BackendIface::env->getExpByName(1);
  Assert(EXPNODE == en.nodetype, "Stack node should be exp node when we've just pushed it.");
#endif
}

extern "C" void clearBackendEnv() {
  BackendIface::env->clear();
}

extern "C" void printBackendLive() {
#ifdef SCHECKMEM
	ExpNode::show_still_live();
#endif
}
