#include "DagOptim.h"
#include "SATSolver.h"
#include "CommandLineArgs.h"

/* Removes the node from our common sub-expression elimination data
 * structures. */
/* TODO: Solve the problem where we can't actually delete symbolic
 * expressions... */
void DagOptim::remove_node(ExpNode* node) {
#ifdef __GCCHECK
	cout << "DagOptim::remove_node: Trying to remove node " << node->globalId
		 << " [" << node->get_ref_count() << "] " << node->lprint()
		 << endl;
#endif
	if (node->get_ref_count() > 0) { return; }

	if (node->get_mother() != NULL) {
		node->get_mother()->dec_ref_count();
		remove_node(node->get_mother());
	}
	if (node->get_father() != NULL) {
		node->get_father()->dec_ref_count();
		remove_node(node->get_father());
	}

	/* Special case--garbage college guard of ternary node. */
	if (node->getType() == ExpNode::TERNARY) {
		((TernaryNode*) node)->get_guard()->dec_ref_count();
		remove_node(((TernaryNode*) node)->get_guard());
	}

		/* Clean up after current node. */
	/* Remove node from the ncmap in DagOptim that maps constants. */
	if (storing && node->getType() == ExpNode::CONST) {
#ifdef __GCCHECK
		cout << "DagOptim::remove_node: Unmapping node " << node->globalId
		<< " with value " << ((ConstNode*) node)->getVal() << endl;
#endif
		int val = ((ConstNode*) node)->getVal();
		if (val != 0 && val != 1 &&
			cnmap.find(val) != cnmap.end() && cnmap[val] == node) {
			cnmap.erase(val);
		}
	}

	/* Remove node from the common sub-expression elimination map. */
#ifdef __GCCHECK
	cout << "DagOptim::remove_node: Removing node " << node->globalId
	<< endl;
#endif
	delete node;
}

ExpNode* DagOptim::simplify(ExpNode* node) {
	storing = false;
	node->accept(*this);
	if (rvalue != node) {
		node->dec_ref_count();
		rvalue->inc_ref_count();
		remove_node(node);
	}
	storing = true;
	return rvalue;
}

void DagOptim::visitAndReplace(ExpNode** node) {
	storing = false;
	(*node)->accept(*this);
	storing = true;
	if (*node != rvalue) {
		remove_node(*node);
		*node = rvalue;
	}
}


DagOptim::DagOptim(ExpDAG& dag):cse(dag)
{
	storing = true;
	ALTER_ARRACS = false;
	rvalue = NULL;	
	initialize(dag);
}

DagOptim::~DagOptim()
{
}

extern CommandLineArgs* PARAMS;

ConstNode* DagOptim::getCnode(int val){
	if (!(val == 1 || val == 0)) {
		return new ConstNode(val);
	}

	if( cnmap.find(val) == cnmap.end() ){
		ConstNode* cnode = new ConstNode(val);
#ifdef __DEBUG_1
		cout << "DagOptim::getCnode: Adding constant " << val << endl;
#endif
		cnmap[val] = cnode;
#ifdef __TEST
		Assert( cnmap.find(val) != cnmap.end()
			  , "New value should be in the map now.");
#endif
		return cnode;
	}else{
		// cnmap[val]->inc_ref_count();
		return cnmap[val];
	}
}

ConstNode* DagOptim::getCnode(bool c){
	return getCnode( c ? 1 : 0 );
}


/**
 * TODO: When is this function called?
 *
 * return value: 
 * 	1 : always true.
 * -1 : always false.
 */
template<typename COMP>
int DagOptim::staticCompare(ExpNode* n1, int C , bool reverse ){
	COMP comp;

#ifdef __TEST
	Assert(n1 != NULL, "DagOptim::staticCompare: n1 can't be null");
#endif

	{
		map<ExpNode*, AbstractNodeValue>::iterator it = anv.find(n1);
		if( it != anv.end() ){
			if(it->second.timestamp == n1->globalId){		
				return it->second.staticCompare<COMP>(C, reverse);
			}else{
				anv.erase(it);
			}
		}
	}
	if(  isConst(n1) ){
		anv[n1].init(getIval(n1));
		anv[n1].timestamp = n1->globalId;	
		bool cm = reverse? comp(C, getIval(n1)) : comp(getIval(n1), C);
		return cm ? 1 : -1;	
	}
	
	if(n1->type == ExpNode::NOT){
		int rv = staticCompare<COMP>(n1->get_mother(), C, reverse);
		anv[n1] = anv[n1->get_mother()];
		anv[n1].timestamp = n1->globalId;		
		return rv;
	}

	if(n1->getOtype() == ExpNode::BOOL){
		AbstractNodeValue& nv = anv[n1];
		nv.init(0);	
		nv.insert(1);
		nv.timestamp = n1->globalId;
		return nv.staticCompare<COMP>(C, reverse);
	}

	if(n1->type == ExpNode::SYM){
		SymNode* inode = dynamic_cast<SymNode*>(n1);
		AbstractNodeValue& nv = anv[n1];
		if(inode->getOtype() == ExpNode::BOOL){
			nv.init(0);
			nv.insert(1);
			nv.timestamp = n1->globalId;
			return nv.staticCompare<COMP>(C, reverse);
		}else{
			nv.makeTop();
			nv.timestamp = n1->globalId;
			return 0;
		}
	}


	if( n1->type == ExpNode::PLUS){
		AbstractNodeValue& nv = anv[n1];
		nv.timestamp = n1->globalId;
		{
			ExpNode* parent = n1->get_mother();
			map<ExpNode*, AbstractNodeValue>::iterator it = anv.find(parent);

			if(it == anv.end()){
				staticCompare<COMP>(parent, C, reverse);
			}else{
				if(it->second.timestamp != parent->globalId){
					anv.erase(it);
					staticCompare<COMP>(parent, C, reverse);
				}
			}

			nv.insert( anv[parent] );
		}
		{
			ExpNode* parent = n1->get_father();
			map<ExpNode*, AbstractNodeValue>::iterator it = anv.find(parent);
			if(it == anv.end()){
				staticCompare<COMP>(parent, C, reverse);
			}else{
				if(it->second.timestamp != parent->globalId){
					anv.erase(it);
					staticCompare<COMP>(parent, C, reverse);
				}
			}
			nv.add( anv[parent] );
		}
		return nv.staticCompare<COMP>(C, reverse);
	}

	anv[n1].makeTop();
	anv[n1].timestamp = n1->globalId;
	return 0;
}

template<typename COMP>
void DagOptim::helper(ExpNode* parent, int C, int reverse, int& rv, AbstractNodeValue& nv){
				int tmp = 0;
				
				map<ExpNode*, AbstractNodeValue>::iterator it = anv.find(parent);

				if(it == anv.end()){
					tmp = staticCompare<COMP>(parent, C, reverse);
					nv.insert( anv[parent] );
				}else{
					if(it->second.timestamp == parent->globalId){
						tmp = it->second.staticCompare<COMP>(C, reverse);
						nv.insert(it->second);
					}else{
						anv.erase(it);
						tmp = staticCompare<COMP>(parent, C, reverse);
						nv.insert( anv[parent] );
					}
				}

				if(tmp == 0 || (rv != -2 && tmp != rv)){				
					rv = 0;	
				}				
				if(rv != 0){
					rv = tmp;
				}
}


template<typename COMP, typename NTYPE>
bool DagOptim::compSymplification(NTYPE& node){
	return false;

	if( typeid(*node.get_mother()) == typeid(PlusNode) &&  typeid(*node.get_father()) == typeid(PlusNode) ){
		ExpNode* momo = node.get_mother()->get_mother();
		ExpNode* mofa = node.get_mother()->get_father();
		
		ExpNode* famo = node.get_father()->get_mother();
		ExpNode* fafa = node.get_father()->get_father();
		if(momo == fafa){
			ExpNode* tmp = fafa;
			fafa = famo;
			famo = tmp;	
		}
		
		if(mofa == famo){
			ExpNode* tmp = momo;
			momo = mofa;
			mofa = tmp;
		}
		
		if(mofa == fafa){
			ExpNode* tmp = fafa;
			fafa = famo;
			famo = tmp;	
			
			tmp = momo;
			momo = mofa;
			mofa = tmp;	
		}
		
		if(momo == famo){
			NTYPE* pnode = new NTYPE();
			pnode->set_parents(mofa, fafa);
			rvalue  = pnode;
			visit(*pnode);
			if(rvalue != pnode){
				remove_node(pnode);
				
			}else{
				addNode(pnode);	
			}
			return true;
		}
	}
		
	if( typeid(*node.get_mother()) == typeid(PlusNode) ){
		
		ExpNode* momo = node.get_mother()->get_mother();
		ExpNode* mofa = node.get_mother()->get_father();
		
		if(mofa == node.get_father()){
			momo = 	node.get_mother()->get_father();;
			mofa = node.get_mother()->get_mother();
		}
		//At this point, if any of the two is equal to node.get_father(), it will be momo
		
		if(momo == node.get_father()){
			NTYPE* pnode = new NTYPE();
			pnode->set_parents(mofa, getCnode(0));
			pnode->inc_ref_count();
			rvalue  = pnode;
			visit(*pnode);		
			if(rvalue != pnode){
				remove_node(pnode);
				//cout<<"=================== saved 3"<<endl;
			}else{
				addNode(pnode);	
			}
			return true;
		}			
	}
		
	if( typeid(*node.get_father()) == typeid(PlusNode) ){
		
		ExpNode* momo = node.get_father()->get_mother();
		ExpNode* mofa = node.get_father()->get_father();
		
		if(mofa == node.get_mother()){
			momo = 	node.get_father()->get_father();;
			mofa = node.get_father()->get_mother();
		}
		//At this point, if any of the two is equal to node.get_father(), it will be momo
		
		if(momo == node.get_mother()){
			NTYPE* pnode = new NTYPE();
			pnode->set_parents(getCnode(0), mofa);
			rvalue  = pnode;
			visit(*pnode);
			if(rvalue != pnode){
				remove_node(pnode);
				//cout<<"=================== saved 3"<<endl;
			}else{
				addNode(pnode);	
			}
			return true;
		}			
	}
		
	
	if(isConst(node.get_mother())){
		int tmp = staticCompare<COMP>(node.get_father(), getIval(node.get_mother()), true);
		if(tmp == 1){
			rvalue  = getCnode(1);
			return 	true;
		}
		if(tmp == -1){
			rvalue  = getCnode(0);
			return true;	
		}
	}
	
	if(isConst(node.get_father())){
		int tmp = staticCompare<COMP>(node.get_mother(), getIval(node.get_father()), false);
		
		if(tmp == 1){
			rvalue  = getCnode(1);
			return 	true;
		}
		if(tmp == -1){
			rvalue  = getCnode(0);
			return true;	
		}
	}
	
	return false;
}

template<typename NTYPE>
bool DagOptim::floatTernary(NTYPE& node){
	/* Instead of floating the ternary guys, combine things that have the same
	 * guard.
	 */
	ExpNode* mother = node.get_mother();
	ExpNode* father = node.get_father();

	if (mother->getType() == ExpNode::TERNARY &&
		father->getType() == ExpNode::TERNARY) {
		TernaryNode* tmother = (TernaryNode*)mother;
		TernaryNode* tfather = (TernaryNode*)father;
		if ((tmother->get_guard())->eqNode(tfather->get_guard())) {
			ExpNode* new_exp = new TernaryNode(tmother->get_guard());

			ExpNode* new_mother = new NTYPE();
			new_mother->set_parents(mother->get_mother(), father->get_mother());

			ExpNode* new_father = new NTYPE();
			new_father->set_parents(mother->get_father(), father->get_father());

			visitAndReplace(&new_exp);
			rvalue = new_exp;
			return true;
		} else {
			return false;
		}

	} else {
		return false;
	}

	/*
	if (mother->getType() == ExpNode::TERNARY
			&& father->getType() == ExpNode::TERNARY) {
		ExpNode* cond1 = ((TernaryNode*) mother)->get_guard();
		ExpNode* cond2 = ((TernaryNode*) father)->get_guard();
		ExpNode* leaf1 = mother->get_mother();
		ExpNode* leaf2 = mother->get_father();
		ExpNode* leaf3 = father->get_mother();
		ExpNode* leaf4 = father->get_father();

		ExpNode* new_top_node = new TernaryNode(cond1);
		ExpNode* new_top_left = new TernaryNode(cond2);

		ExpNode* new_top_left_left = new NTYPE();
		new_top_left_left->set_parents(leaf1, leaf3);

		visitAndReplace(&new_top_left_left);

		ExpNode* new_top_left_right = new NTYPE();
		new_top_left_right->set_parents(leaf1, leaf4);
		visitAndReplace(&new_top_left_right);

		new_top_left->set_parents(new_top_left_left, new_top_left_right);

		ExpNode* new_top_right = new TernaryNode(cond2);

		ExpNode* new_top_right_left = new NTYPE();
		new_top_right_left->set_parents(leaf2, leaf3);
		visitAndReplace(&new_top_right_left);

		ExpNode* new_top_right_right = new NTYPE();
		new_top_right_right->set_parents(leaf2, leaf4);
		visitAndReplace(&new_top_right_right);

		new_top_right->set_parents(new_top_right_left, new_top_right_right);

		new_top_node->set_parents(new_top_left, new_top_right);

		// Return the new ternary node.
		rvalue = new_top_node;
		return true;
	} else if (mother->getType() == ExpNode::TERNARY) {
		ExpNode* cur_guard = ((TernaryNode*) mother)->get_guard();
		ExpNode* exp1 = mother->get_mother();
		ExpNode* exp2 = mother->get_father();

		TernaryNode* new_node = new TernaryNode(cur_guard);

		// Make the new left operand and optimize it.
		ExpNode* new_exp1 = new NTYPE();
		new_exp1->set_parents(exp1, father);
		visitAndReplace(&new_exp1);

		// Make the new right operand and optimize it.
		ExpNode* new_exp2 = new NTYPE();
		new_exp2->set_parents(exp2, father);
		visitAndReplace(&new_exp2);

		// We are returning a ternary node with the +'s pushed down.
		new_node->set_parents(new_exp1, new_exp2);

		rvalue = new_node;
		return true;
	} else if (father->getType() == ExpNode::TERNARY) {
		ExpNode* cur_guard = ((TernaryNode*) father)->get_guard();
		ExpNode* exp1 = father->get_mother();
		ExpNode* exp2 = father->get_father();

		TernaryNode* new_node = new TernaryNode(cur_guard);

		// Make the new left operand and optimize it.
		ExpNode* new_exp1 = new NTYPE();
		new_exp1->set_parents(mother, exp1);
		visitAndReplace(&new_exp1);

		// Make the new right operand and optimize it.
		ExpNode* new_exp2 = new NTYPE();
		new_exp2->set_parents(mother, exp2);
		visitAndReplace(&new_exp2);

		new_node->set_parents(new_exp1, new_exp2);

		rvalue = new_node;
		return true;
	}
	rvalue = &node;
	return false;
	*/
}

bool DagOptim::isNegOfEachOther(ExpNode* n1, ExpNode* n2){
	if( typeid(*n1) == typeid(NEG_node) ){
		if(  n1->get_mother() == n2){
			return true;	
		}
	}
	
	if(  typeid(*n2) == typeid(NEG_node) ){
		if( n2->get_mother() == n1){
			return true;
		}
	}
	return false;	
}



bool DagOptim::isNotOfEachOther(ExpNode* n1, ExpNode* n2){
	if( n1->type == ExpNode::NOT){
		if( n1->get_mother() == n2){
			return true;	
		}
	}
	
	if( n2->type == ExpNode::NOT){
		if( n2->get_mother() == n1){
			return true;
		}
	}
	return false;
}
bool DagOptim::isConst(const ExpNode* n1){
	return (n1->type == ExpNode::CONST);
}

bool DagOptim::getBval(const ExpNode* n1){
	return getIval(n1) != 0;
}	

int  DagOptim::getIval(const ExpNode* n1){
	Assert( isConst(n1), "This node is not a constant !!");
	const ConstNode * cn = dynamic_cast<const ConstNode*>(n1);
	return cn->getVal()	;
}

void DagOptim::visit( SymNode& node ){
	rvalue = &node;
}

void DagOptim::visit( ConstNode& node ){
	if (!(node.getVal() == 1 || node.getVal() == 0)) {
		rvalue = &node;
	}

	int val = node.getVal();
	if( cnmap.find(val) == cnmap.end() ){
#ifdef __DEBUG_1
		cout << "DagOptim::visit(ConstNode&): Adding constant " << val << endl;
#endif
		cnmap[val] = &node;
		rvalue = &node;
	}else{
		rvalue = cnmap[val];
	}
}

/*
 * Optimize conditional nodes by first trying to remove the condition.
 */
void DagOptim::visit( TernaryNode& node ) {
	ExpNode* guard = node.get_guard();

	/* First try to remove the condition. */
	if (guard->getType() == ExpNode::CONST) {
		if (((ConstNode*)guard)->getVal() == 1) {
			rvalue = node.get_mother();
			visitAndReplace(&rvalue);
		} else {
			rvalue = node.get_father();
		}
	} else {
		rvalue = &node;
	}
}

void DagOptim::visit( AndNode& node ){
	// if (floatTernary(node)) { return; }

	if( node.get_father() == node.get_mother() ){ // x & x == x
		rvalue = node.get_father();
		// rvalue->inc_ref_count();
		return;	
	}
	if( isNotOfEachOther(node.get_father(), node.get_mother()) ){ // x & !x == false
		rvalue = getCnode(0);
		return;	
	}
	if( isConst(node.get_father()) ){
		if( isConst(node.get_mother()) ){ // const prop
			rvalue = getCnode ( getBval( node.get_father() ) && getBval( node.get_mother() ) );
			return;
		}
		if( !getBval(node.get_father()) ){ // x & false == false;
			rvalue = getCnode(0);
			return;
		}else{
			rvalue = node.get_mother();
			// rvalue->inc_ref_count();
			return;	
		}		
	}
	if( isConst(node.get_mother()) ){
		if( ! getBval(node.get_mother())){ // false & x == false;
			rvalue = getCnode(0);			
			return;
		}else{
			rvalue = node.get_father();
			// rvalue->inc_ref_count();
			return;	
		}
	}

	if(node.get_mother()->type == ExpNode::AND){
		if(isNotOfEachOther(node.get_father(), node.get_mother()->get_mother())){
			rvalue = getCnode(0);
			return;	
		}
		if(isNotOfEachOther(node.get_father(), node.get_mother()->get_father())){
			rvalue = getCnode(0);
			return;	
		}

	}

	if(node.get_father()->type == ExpNode::AND){
		if(isNotOfEachOther(node.get_mother(), node.get_father()->get_mother())){
			rvalue = getCnode(0);
			return;	
		}
		if(isNotOfEachOther(node.get_mother(), node.get_father()->get_father())){
			rvalue = getCnode(0);
			return;	
		}

	}

	rvalue = &node;
}

void DagOptim::visit( OrNode& node ){
	// if (floatTernary(node)) { return; }

	if( node.get_father() == node.get_mother() ){ // x | x == x
		rvalue = node.get_father();
		// rvalue->inc_ref_count();
		return;	
	}
	if( isNotOfEachOther(node.get_father(), node.get_mother()) ){ // x | !x == true
		rvalue = getCnode(1);
		return;	
	}
	if( isConst(node.get_father()) ){
		if( isConst(node.get_mother()) ){ // const prop
			rvalue = getCnode ( getBval( node.get_father() ) || getBval( node.get_mother() ) );
			return;
		}
		if( getBval(node.get_father()) ){ // x | true == true
			rvalue = getCnode(1);			
			return;
		}else{
			rvalue = node.get_mother();
			// rvalue->inc_ref_count();
			return;				
		}		
	}
	if( isConst(node.get_mother()) ){ // true | x == true
		if( getBval(node.get_mother())){
			rvalue = getCnode(1);			
			return;
		}else{
			rvalue = node.get_father();
			// rvalue->inc_ref_count();
			return;	
		}
	}
	rvalue = &node;
}


void DagOptim::visit( NOT_node& node ){
	if( isConst(node.get_mother()) ){ // const prop
		rvalue = getCnode( ! getBval(node.get_mother()) );
		return;
	}
	if( node.get_mother()->type == ExpNode::NOT){// ! ! x == x;
		rvalue = node.get_mother()->get_mother();
		// rvalue->inc_ref_count();
		return;
	}
	rvalue = &node;
}


void DagOptim::visit( PlusNode& node ){
	// if (floatTernary(node)) { return; }

	ExpNode* mother = node.get_mother();
	ExpNode* father = node.get_father();
	int nc = 0;
	
	if(isConst(node.get_father())){
		mother = node.get_father();
		father = node.get_mother();
		nc++;
	}
	if(isConst(node.get_mother())){
		nc++;	
	}
	//At this point, if nc>0, get_mother() is const,
	
	if(nc > 0){
		if(nc > 1){
			rvalue  = getCnode( getIval( mother ) + getIval( father ) );
			return;
		}else{
			if(getIval( mother ) == 0){
				rvalue  = father;
				// rvalue->inc_ref_count();
				return;
			}
			if( typeid(*father) == typeid(node) ){
				PlusNode& fathernode = *dynamic_cast<PlusNode*>(father);
				if( isConst(fathernode.get_father()) ){
					//cout<<" DID REPLACEMENT 1"<<endl;
					ExpNode* nconst = getCnode(getIval( fathernode.get_father() ) + getIval( mother ));
					PlusNode* pnode = new PlusNode();
					pnode->set_parents(nconst, fathernode.get_mother());
					// Assert(!isConst(fathernode.get_mother()), "DagOptim: This can't happen");
					addNode(pnode);		
					// stillPrivate = pnode;
					rvalue  = pnode;
					// rvalue->inc_ref_count();
					return;
				}
				if( isConst(fathernode.get_mother()) ){
					//cout<<" DID REPLACEMENT 2"<<endl;
					ExpNode* nconst = getCnode(getIval( fathernode.get_mother() ) + getIval( mother ));
					PlusNode* pnode = new PlusNode();
					pnode->set_parents(nconst, fathernode.get_father());
					//Assert(!isConst(fathernode.get_father()), "DagOptim: This can't happen");
					addNode(pnode);	
					// stillPrivate = pnode;
					rvalue  = pnode;
					// rvalue->inc_ref_count();
					return;
				}
				
			}
			
		}	
	}
	
	rvalue = &node;
	
	
	if( isNegOfEachOther(node.get_mother(), node.get_father()) ){
		rvalue = getCnode(0);
		return;
	}
	
	
	if( typeid(*node.get_father()) == typeid(node) ){
		PlusNode& parent = *dynamic_cast<PlusNode*>(node.get_father());
		ExpNode* tt1 = parent.get_father();
		ExpNode* tt2 = parent.get_mother();
		ExpNode* tt3 = node.get_mother();
			
		if(isNegOfEachOther( tt1 , tt2)){
			rvalue = tt3;
			// rvalue->inc_ref_count();
			return;	
		}
		if(isNegOfEachOther( tt1 , tt3)){
			rvalue = tt2;
			// rvalue->inc_ref_count();
			return;	
		}
		if(isNegOfEachOther( tt2 , tt3)){
			rvalue = tt1;
			// rvalue->inc_ref_count();
			return;	
		}
	}
	
	
	if( typeid(*node.get_mother()) == typeid(node) ){
		PlusNode& parent = *dynamic_cast<PlusNode*>(node.get_mother());
		ExpNode* tt1 = parent.get_father();
		ExpNode* tt2 = parent.get_mother();
		ExpNode* tt3 = node.get_father();
			
		if(isNegOfEachOther( tt1 , tt2)){
			rvalue = tt3;
			// rvalue->inc_ref_count();
			return;	
		}
		if(isNegOfEachOther( tt1 , tt3)){
			rvalue = tt2;
			// rvalue->inc_ref_count();
			return;	
		}
		if(isNegOfEachOther( tt2 , tt3)){
			rvalue = tt1;
			// rvalue->inc_ref_count();
			return;	
		}
	}
	
	
	
	if( isConst(node.get_mother()) ){ // const prop
		if( isConst(node.get_father()) ){
			rvalue  = getCnode( getIval( node.get_mother() ) + getIval( node.get_father() ) );
			Assert(false, "Optim is useless");
			return;
		}	
		if(  getIval( node.get_mother() ) == 0 ){
			rvalue  = node.get_father();
			// rvalue->inc_ref_count();
			Assert(false, "Optim is useless");
			return;
		}		
		if( typeid(*node.get_father()) == typeid(node) ){
				PlusNode& fathernode = *dynamic_cast<PlusNode*>(node.get_father());
				if( isConst(fathernode.get_father()) ){
					ExpNode* nconst = getCnode(getIval( fathernode.get_father() ) + getIval( node.get_mother() ));
					PlusNode* pnode = new PlusNode();
					pnode->set_parents(nconst, fathernode.get_mother());
					//Assert(!isConst(fathernode.get_mother()), "DagOptim: This can't happen");
					addNode(pnode);
					Assert(false, "Optim is useless");
				}
			
		}
	}
	if( isConst(node.get_father()) ){
		if(  getIval( node.get_father() ) == 0 ){
			rvalue  = node.get_mother();
			// rvalue->inc_ref_count();
			Assert(false, "Optim is useless");
			return;
		}	
	}
	
	rvalue = &node;
}

void DagOptim::visit( TimesNode& node ){
	// if (floatTernary(node)) { return; }

	if( isConst(node.get_mother()) ){ // const prop
		if( isConst(node.get_father()) ){
			rvalue  = getCnode( getIval( node.get_mother() ) * getIval( node.get_father() ) );
			return;
		}
		if(  getIval( node.get_mother() ) == 0 ){
			rvalue  = getCnode( 0 );
			return;
		}
		if(  getIval( node.get_mother() ) == 1 ){
			rvalue  = node.get_father();
			// rvalue->inc_ref_count();
			return;
		}	
	}
	if( isConst(node.get_father()) ){
		if(  getIval( node.get_father() ) == 0 ){
			rvalue  = getCnode( 0 );
			return;
		}
		if(  getIval( node.get_father() ) == 1 ){
			rvalue  = node.get_mother();
			// rvalue->inc_ref_count();
			return;
		}
	}
	rvalue = &node;
}

void DagOptim::visit( DivNode& node ){
	// if (floatTernary(node)) { return; }

	if( isConst(node.get_mother()) ){ // const prop
		if( isConst(node.get_father()) ){
			int c = getIval( node.get_father() );
			if(c != 0){
				rvalue  = getCnode( getIval( node.get_mother() ) / c );
			}else{
				rvalue  = getCnode( 0 );
			}
			return;
		}			
	}
	rvalue = &node;
}

void DagOptim::visit( NEG_node& node ){
	if( isConst(node.get_mother()) ){	// const prop
		rvalue  = getCnode( -getIval( node.get_mother() ) );
		return;
	}
	if( typeid(*(node.get_mother())) == typeid(node)  ){// - -x == x;
		rvalue = node.get_mother()->get_mother();
		// rvalue->inc_ref_count();
		return;
	}
	rvalue = &node;
}
	
	
		
void DagOptim::visit( GtNode& node ){
	// if (floatTernary(node)) { return; }

	if( isConst(node.get_mother()) ){
		if( isConst(node.get_father()) ){
			rvalue  = getCnode( getIval( node.get_mother() ) > getIval( node.get_father() ) );
			return;
		}			
	}
	
	if( node.get_mother() == node.get_father() ){
		rvalue = getCnode(false);
		return;
	} 
	
	

	if( compSymplification<greater<int> , GtNode>(node) ){
		return;	
	}
	

	rvalue = &node;
}
void DagOptim::visit( GeNode& node ){
	// if (floatTernary(node)) { return; }

	if( isConst(node.get_mother()) ){
		if( isConst(node.get_father()) ){
			rvalue  = getCnode( getIval( node.get_mother() ) >= getIval( node.get_father() ) );
			return;
		}			
	}
	if( node.get_mother() == node.get_father() ){
		rvalue = getCnode(true);
		return;
	}
	
	if( compSymplification<greater_equal<int> , GeNode>(node) ){
		return;	
	}
	 
	rvalue = &node;
}


void DagOptim::visit( LtNode& node ){
	// if (floatTernary(node)) { return; }

	if( isConst(node.get_mother()) ){
		if( isConst(node.get_father()) ){
			rvalue  = getCnode( getIval( node.get_mother() ) < getIval( node.get_father() ) );
			return;
		}			
	}
	if( node.get_mother() == node.get_father() ){
		rvalue = getCnode(false);
		return;
	} 
	
	if( compSymplification<less<int> , LtNode>(node) ){
		return;	
	}	
	
	rvalue = &node;
}

void DagOptim::visit( LeNode& node ){
	// if (floatTernary(node)) { return; }

	if( isConst(node.get_mother()) ){
		if( isConst(node.get_father()) ){
			rvalue  = getCnode( getIval( node.get_mother() ) <= getIval( node.get_father() ) );
			return;
		}			
	}
	if( node.get_mother() == node.get_father() ){
		rvalue = getCnode(true);
		return;
	} 
	
	
	if( compSymplification<less_equal<int> , LeNode>(node) ){
		return;	
	}	
	
	
	
	rvalue = &node;
}

void DagOptim::visit( EqNode& node ){
	// if (floatTernary(node)) { return; }

	if( node.get_mother() == node.get_father() ){
		rvalue = getCnode(true);
		return;
	} 

	if( isConst(node.get_mother()) ){
		if( isConst(node.get_father()) ){
			rvalue  = getCnode( getIval( node.get_mother() ) == getIval( node.get_father() ) );
			return;
		}
		int mvalue = getIval( node.get_mother() );
		if( node.get_father()->getOtype() == ExpNode::BOOL ){
			
			if(mvalue == 1){
				rvalue = node.get_father();
				// rvalue->inc_ref_count();
				return;	
			}
			
			if(mvalue == 0){
				NOT_node* nt = new NOT_node();				
				nt->set_parents(node.get_father(), NULL);
				addNode(nt);
				nt->accept(*this);
				return;	
			}
		}

	}

	if( isConst(node.get_father()) ){
		if( node.get_mother()->getOtype() == ExpNode::BOOL ){
			int mvalue = getIval( node.get_father() );
			
			if(mvalue == 1){
				rvalue = node.get_mother();
				// rvalue->inc_ref_count();
				return;	
			}
			
			if(mvalue == 0){
				NOT_node* nt = new NOT_node();
				nt->set_parents(node.get_mother(), NULL);
				addNode(nt);
				nt->accept(*this);
				return;	
			}
		}
	}

	/* TODO: Put this back eventually? */
	if( compSymplification<equal_to<int> , EqNode>(node) ){
#ifdef __DEBUG_1
		cout << "COMPUTED SIMPLIFICATION." << endl;
#endif
		return;	
	}
	rvalue = &node;
}

void DagOptim::visit( NeqNode& node ){
	// if (floatTernary(node)) { return; }

	if( node.get_mother() != node.get_father() ){
		rvalue = getCnode(true);
		return;
	}

	if( isConst(node.get_mother()) ){
		if( isConst(node.get_father()) ){
			rvalue  = getCnode( getIval( node.get_mother() ) != getIval( node.get_father() ) );
			return;
		}
		int mvalue = getIval( node.get_mother() );
		if( node.get_father()->getOtype() == ExpNode::BOOL ){

			if(mvalue == 0){
				rvalue = node.get_father();
				// rvalue->inc_ref_count();
				return;
			}

			if(mvalue == 1){
				NOT_node* nt = new NOT_node();
				nt->set_parents(node.get_father(), NULL);
				addNode(nt);
				nt->accept(*this);
				return;
			}
		}

	}

	if( isConst(node.get_father()) ){
		if( node.get_mother()->getOtype() == ExpNode::BOOL ){
			int mvalue = getIval( node.get_father() );

			if(mvalue == 0){
				rvalue = node.get_mother();
				// rvalue->inc_ref_count();
				return;
			}

			if(mvalue == 1){
				NOT_node* nt = new NOT_node();
				nt->set_parents(node.get_mother(), NULL);
				addNode(nt);
				nt->accept(*this);
				return;
			}
		}
	}

	if( compSymplification<not_equal_to<int> , NeqNode>(node) ){
		return;
	}
	rvalue = &node;
}

void DagOptim::initLight(ExpDAG& dag){
	cnmap.clear();
	cse.clear();
}

void DagOptim::initialize(ExpDAG& dag){
	anv.clear();
	initLight(dag);
}

ExpNode* DagOptim::computeCSE(ExpNode* node){
	return cse.computeCSE(node);
}

ExpNode* DagOptim::computeOptim(ExpNode* node){
#ifdef __DEBUG_1
	cout << "DagOptim::computeOptim: " << node->lprint() << endl;
	Assert( node->globalId != -22
		  , "Internal error: node should not have global ID -22");
#endif
	node->accept(*this);
	bool simple_opt = false;

	if (node != rvalue) {
#ifdef __DEBUG_1
		cout << "DagOptim::computeOptim: Replaced node with "
			 << rvalue->lprint() << endl;
#endif
		node = rvalue;
		node->inc_ref_count();
		simple_opt = true;
	}
#ifdef __DEBUG_1
	Assert( node->globalId != -22
		  , "Basic optimizations should not have returned global ID -22");
#endif

	/* Now perform common sub-expression elimination. */
	ExpNode* tmp = cse.computeCSE(node);

#ifdef __DEBUG_1
	if (tmp != node) {
		cout << "DagOptim::computeOptim: CSE replaced node with "
			 << rvalue->lprint() << endl;
	}
#endif

	/* Need to handle the case that we common sub-expression replaced the new
	 * node we got from our earlier optimizations.
	 */
	if (simple_opt && tmp != node) {
		node->dec_ref_count();
		remove_node(node);
	}

#ifdef __DEBUG_1
	Assert(tmp->globalId != -22, "CSE should not have returned global ID -22");
#endif
	return tmp;
}


void DagOptim::cleanup(ExpDAG& dag){
	dag.removeNullNodes();
	//dag.repOK();
	//dag.sort_graph();
	dag.cleanup();
#ifdef _DEBUG
	dag.repOK();
#endif
}
