#include "ExpressionNodes.h"
#include "ExpDAG.h"

#include <sstream>
#include <algorithm>

bool EnvNode::eq(EnvNode other) {
  if ((CONSTVAL == nodetype) && (CONSTVAL == other.nodetype)) {
    return (get_val() == other.get_val());
  } else if ((EXPNODE == nodetype) && (EXPNODE == other.nodetype)) {
    EnvNode eq_node = get_node()->eqNode(other.get_node());
    return ((CONSTVAL == eq_node.nodetype) && (eq_node.get_val()));
  } else {
    return false;
  }
}

/**
 * ExpNode base class definitions.
 */
int ExpNode::__s_next_global_id = 0;

#ifdef SCHECKMEM
set<ExpNode*> ExpNode::allocated;
#endif 

/* Default expression node constructor. */
ExpNode::ExpNode() :
		globalId(__s_next_global_id++), __mother(NULL), __father(NULL),
		__sym(false), __ref_count(0), flag(0),
		otype(BOTTOM)
{ 
#ifdef __GCCHECK
	cout << "ExpNode::ExpNode: Constructing node " << globalId << endl;
#endif
#ifdef SCHECKMEM
  allocated.insert(this);
#endif
}

/* Default copy constructor. */
ExpNode::ExpNode(const ExpNode& bn) :
		globalId(__s_next_global_id++),
		__mother(bn.__mother), __father(bn.__father),
		__sym(bn.__sym), __ref_count(bn.__ref_count), flag(0),
		otype(bn.otype), type(bn.type)
{
	if (__mother != NULL) {
		__mother->inc_ref_count();
	}
	if (__father != NULL) {
		__father->inc_ref_count();
	}
#ifdef SCHECKMEM
  allocated.insert(this);
#endif
}

ExpNode::~ExpNode(){
#ifdef __GCCHECK
	cout << "ExpNode::ExpNode: Deleting node " << globalId
		 << " with a ref count of " << __ref_count << endl;
#endif
#ifdef SCHECKMEM
  allocated.erase(this);
#endif
}

void ExpNode::resetId() { globalId = __s_next_global_id++; }

/**
 * Getters and setters.
 */
void ExpNode::set_parents(ExpNode* new_mother, ExpNode* new_father) {
	if (new_mother == __mother && new_father == __father) {
		return;
	}

	Assert(__mother == NULL, "Previous mother is null.");
	Assert(__father == NULL, "Previous father is null");

	__mother = new_mother;
	__father = new_father;
	if (__mother != NULL) {
		__mother->inc_ref_count();
		merge_sym(__mother);
	}
	if (__father != NULL) {
		__father->inc_ref_count();
		merge_sym(__father);
	}
}

ExpNode::OutType ExpNode::joinOtype(OutType t1, OutType t2) const{
  if(t1 == BOTTOM){ return t2; }
  if(t2 == BOTTOM){ return t1; }
  if( t2 == t1 ){
    return t1;
  }else{
    return INT;
  }
}

void ExpNode::neighbor_replace(ExpNode* replacement){
  // ExpNode* onode = this;
	Assert(false, "ExpNode::neighbor_replace not implemented.");
}

string ExpNode::get_tname() const{
  switch(type){
    case PLUS: return "PLUS";
    case TIMES: return "TIMES";
    case DIV: return "DIV";
    case NEG: return "NEG";
    case CONST: return "CONST";
    case SYM : return "SYM";
    case LIST: return "LIST";
    case RECORD: return "RECORD";
    case TERNARY: return "TERNARY";
    case AND: return "AND";
    case OR: return "OR";
    case NOT: return "NOT";
    case GT: return "GT";
    case GE: return "GE";
    case LT: return "LT";
    case LE: return "LE";
    case EQ: return "EQ";
    case NEQTY: return "NEQTY";
    case ASSERT: return "ASSERT";
    case STRING: return "STRING";
  }
  cout << "ABOUT TO ABORT BECAUSE OF " << globalId << "  " << type << endl;
  throw BasicError("Err", "Err");
}

string ExpNode::get_sym() const{
  switch(type){
    case PLUS: return "+";
    case TIMES: return "*";
    case DIV: return "/";
    case NEG: return "-";
    case CONST: return "C";
    case SYM: return "S";
    case LIST: return "R";
    case RECORD: return "r";
    case TERNARY: return "T";
    case AND: return "&";
    case OR: return "|";
    case NOT: return "!";
    case GT: return ">";
    case GE: return ">=";
    case LT: return "<";
    case LE: return "<=";
    case EQ: return "==";
    case NEQTY: return "!=";
    case ASSERT: return "ASSERT";
    case STRING: return "STRING";
  }
  cout<<"ABOUT TO ABORT BECAUSE OF "<< globalId <<"  "<<type<<endl;
  throw BasicError("Err", "Err");
}

void ExpNode::printval() {
	cout << this->lprint() << endl;
}

string ExpNode::lprint(){
  stringstream str;
  str << "[" << __ref_count << "] ";
  str << globalId <<"= ";
  if(__mother != NULL){
#ifdef __GCCHECK
	  Assert( __mother->globalId != -22
			, "Mother should not have been garbage collected");
#endif
    str << get_mother()->lprint() << " ";
  }
  str << get_sym()<<" ";

  if(__father != NULL){
#ifdef __GCCHECK
	 Assert( __father->globalId != -22
			, "Father should not have been garbage collected");
#endif
    str << get_father()->lprint() << " ";
  }
  return str.str();
}

yices_expr ExpNode::to_yices(yices_context yctxt) {
  cout << type << endl;
  Assert(false, "Should not get here!");
}

string ExpNode::lid(){
  stringstream str;
  str << globalId;
  return str.str();
}


/**
 * Serialization.
 */
string ExpNode::get_namestr() const {
  stringstream str;
  if(globalId > -1)
    str << globalId << "__"<<get_tname();
  else{      
    str<<"name_"<<abs(globalId)<<"__"<<get_tname();

  }
  return str.str();
}

/**
 * The parent Yices-serialization function should never actually get called.
 */
string ExpNode::get_yicesstr() const {
	cout << "ExpNode::get_yicesstr: Type " << this->getType() << endl;
  Assert(false, "Yices serialization should be defined for this node.");
}

ExpNode::OutType ExpNode::getOtype() const{
  Assert(false, "This shouldn't get called");
  return BOOL;
}

ExpNode::Type ExpNode::getType() const{
  Assert(false, "Somebody didn't implement getType.");
  return AND;
}

int ExpNode::do_dfs(int idx){
  if( flag != 0)
    return idx;
  flag = 1;    
  /* TODO: Change this to do DFS over parents. */
  Assert(idx >= 0, "This shouldn't happen");
  /* TODO: Figure out what we should set here. */
  // id = idx;
  return idx-1;
}

int ExpNode::back_dfs(int idx){
  if( flag != 0)
    return idx;
  flag = 1;    
  if(__father != NULL){
    idx = __father->back_dfs(idx);
  }
  if(__mother != NULL){
    idx = __mother->back_dfs(idx);
  }
  /* TODO: Figure out what to set here. */
  // id = idx;
  return idx+1;
}

void ExpNode::replace_parent(const ExpNode * oldpar, ExpNode* newpar){
  Assert( oldpar != NULL, "oldpar can't be null");
  Assert( newpar != NULL, "oldpar can't be null");
  if(__father == oldpar){
    __father = newpar;
    //oldpar->remove_child(this);
    //we assume the old parent is going to be distroyed, so we don't
    //bother modifying it.
    merge_sym(__father);
  }
  if(__mother == oldpar){
    __mother = newpar;
    //oldpar->remove_child(this);
    //we assume the old parent is going to be distroyed, so we don't
    //bother modifying it.
    merge_sym(__mother);
  }
}

void ExpNode::printSubDAG(ostream& out, set<const ExpNode* >& s)const{
  s.insert(this);
  if(__father != NULL && s.count(__father) == 0 ){
    __father->printSubDAG(out, s);
  }	
  if(__mother != NULL && s.count(__mother) == 0){
    __mother->printSubDAG(out, s);
  }
  outDagEntry(out);	
}

void ExpNode::outDagEntry(ostream& out) const{
  if( __father != NULL){
    out<<" "<<__father->get_namestr()<<" -> "<<get_namestr()<<" ; "<<endl;
  }
  if( __mother != NULL){
    out<<" "<<__mother->get_namestr()<<" -> "<<get_namestr()<<" ; "<<endl;
  }
  if(__father == NULL && __mother == NULL){
    out<<"// orphan node: "<<get_namestr()<<" ; "<<endl ;
  }
}

void ExpNode::addVars(set<int>& vars) { return; }
EnvNode ExpNode::eqNode(ExpNode* other) {
  Assert(false, "eqNode should have been defined.");
}

/**
 * Binary nodes.
 */

/**
 * AndNode
 */
AndNode::AndNode(const AndNode& b): ExpNode() { }
AndNode::AndNode(){
	type = AND;
	otype = BOOL;
}
void AndNode::accept(NodeVisitor& visitor) {
  visitor.visit( *this );
}
ExpNode* AndNode::clone(bool copyChildren = true) {
  return new AndNode(*this);
};

yices_expr AndNode::to_yices(yices_context yctxt) {
#ifdef __TEST
  Assert(get_mother() != NULL, "Plus node should have two parents.");
  Assert(get_father() != NULL, "Plus node should have two parents.");
#endif
  __mother->ytype = ExpNode::BOOL;
  yices_expr ym = get_mother()->to_yices(yctxt);
  __father->ytype = ExpNode::BOOL;
  yices_expr yf = get_father()->to_yices(yctxt);
  yices_expr args[2] = { ym, yf };
  yices_expr sum = yices_mk_and(yctxt, args, 2);

  return sum;
}

string AndNode::get_yicesstr() const {
	stringstream s;
	__mother->ytype = ExpNode::BOOL;
	s << "( " << "and " << __mother->get_yicesstr() << " ";
	__father->ytype = ExpNode::BOOL;
	s << __father->get_yicesstr() << " )";
	return s.str();
}

void AndNode::addVars(set<int>& varset) {
	get_mother()->addVars(varset);
	get_father()->addVars(varset);
}

EnvNode AndNode::eqNode(ExpNode* other) {
	if (other->getType() == ExpNode::AND) {
		EnvNode eq_mothers = get_mother()->eqNode(other->get_mother());
    EnvNode eq_fathers = get_father()->eqNode(other->get_father());
    if ((CONSTVAL == eq_mothers.nodetype) && (eq_mothers.get_val())) {
		    return eq_fathers;
    } else {
      if ((CONSTVAL == eq_fathers.nodetype) && (eq_fathers.get_val())) {
          return eq_mothers;
      } else {
        AndNode* an = new AndNode();
        an->set_parents(eq_mothers.get_node(), eq_fathers.get_node());
        return EnvNode(an);
      }
    }
	}
  return EnvNode(false);
}

/**
 * OrNode
 */
OrNode::OrNode() : ExpNode() { type = OR; otype = BOOL; }
OrNode::OrNode(const OrNode& bn): ExpNode(bn){ }
void OrNode::accept(NodeVisitor& visitor)  { visitor.visit( *this ); }
ExpNode* OrNode::clone(bool copyChildren){return new OrNode(*this);  };

yices_expr OrNode::to_yices(yices_context yctxt) {
#ifdef __TEST
  Assert(get_mother() != NULL, "Plus node should have two parents.");
  Assert(get_father() != NULL, "Plus node should have two parents.");
#endif
  __mother->ytype = ExpNode::BOOL;
  yices_expr ym = get_mother()->to_yices(yctxt);
  __father->ytype = ExpNode::BOOL;
  yices_expr yf = get_father()->to_yices(yctxt);
  yices_expr args[2] = { ym, yf };
  yices_expr sum = yices_mk_or(yctxt, args, 2);

  return sum;
}

string OrNode::get_yicesstr() const {
	stringstream s;
	__mother->ytype = ExpNode::BOOL;
	s << "( " << "or " << __mother->get_yicesstr() << " ";
	__father->ytype = ExpNode::BOOL;
	s << __father->get_yicesstr() << " )";
	return s.str();
}

void OrNode::addVars(set<int>& varset) {
	get_mother()->addVars(varset);
	get_father()->addVars(varset);
}

EnvNode OrNode::eqNode(ExpNode* other) {
	if (other->getType() == ExpNode::OR) {
    EnvNode eq_mothers = get_mother()->eqNode(other->get_mother());
    EnvNode eq_fathers = get_father()->eqNode(other->get_father());
    if ((CONSTVAL == eq_mothers.nodetype) && (eq_mothers.get_val())) {
        return eq_fathers;
    } else {
      if ((CONSTVAL == eq_fathers.nodetype) && (eq_fathers.get_val())) {
          return eq_mothers;
      } else {
        AndNode* an = new AndNode();
        an->set_parents(eq_mothers.get_node(), eq_fathers.get_node());
        return EnvNode(an);
      }
    }
  }
  return EnvNode(false);
}

/**
 * PlusNode
 */
PlusNode::PlusNode() : ExpNode() {
	type = PLUS;
	__is_minus = false;
}

PlusNode::PlusNode(bool minus) {
	__is_minus = minus;
}
PlusNode::PlusNode(const PlusNode& bn) :
	ExpNode(bn), __is_minus(bn.__is_minus) {
}
void PlusNode::set_minus(bool minus) {
	__is_minus = minus;
}
void PlusNode::accept(NodeVisitor& visitor) {
	visitor.visit(*this);
}
ExpNode* PlusNode::clone(bool copyChildren) {
	return new PlusNode(*this);
}
;
ExpNode::OutType PlusNode::getOtype() const {
	return INT;
}
ExpNode::Type PlusNode::getType() const {
	return PLUS;
}

string PlusNode::lprint() {
	stringstream ss;
	ss << get_mother()->lprint() << " + " << get_father()->lprint();
	return ss.str();
}

yices_expr PlusNode::to_yices(yices_context yctxt) {
#ifdef __TEST
  Assert(get_mother() != NULL, "Plus node should have two parents.");
  Assert(get_father() != NULL, "Plus node should have two parents.");
#endif
  yices_expr ym = get_mother()->to_yices(yctxt);
  yices_expr yf = get_father()->to_yices(yctxt);
  yices_expr args[2] = { ym, yf };
  yices_expr sum = yices_mk_sum(yctxt, args, 2);

  return sum;
}

string PlusNode::get_yicesstr() const {
	stringstream s;
	__mother->ytype = ExpNode::INT;
	s << "( " << "+ " << __mother->get_yicesstr() << " ";
	__father->ytype = ExpNode::INT;
	s << __father->get_yicesstr() << " )";
	return s.str();
}

void PlusNode::addVars(set<int>& varset) {
	get_mother()->addVars(varset);
	get_father()->addVars(varset);
}

EnvNode PlusNode::eqNode(ExpNode* other) {
	if (other->getType() == ExpNode::PLUS) {
    EnvNode eq_mothers = get_mother()->eqNode(other->get_mother());
    EnvNode eq_fathers = get_father()->eqNode(other->get_father());
    if ((CONSTVAL == eq_mothers.nodetype) && (eq_mothers.get_val())) {
        return eq_fathers;
    } else {
      if ((CONSTVAL == eq_fathers.nodetype) && (eq_fathers.get_val())) {
          return eq_mothers;
      } else {
        AndNode* an = new AndNode();
        an->set_parents(eq_mothers.get_node(), eq_fathers.get_node());
        return EnvNode(an);
      }
    }
  }
  return EnvNode(false);
}

/**
 * TimesNode
 */
TimesNode::TimesNode() : ExpNode() {
	type = TIMES;
}
TimesNode::TimesNode(const TimesNode& bn) :
	ExpNode(bn) {
}
void TimesNode::accept(NodeVisitor& visitor) {
	visitor.visit(*this);
}
ExpNode* TimesNode::clone(bool copyChildren = true) {
	return new TimesNode(*this);
}

ExpNode::OutType TimesNode::getOtype() const {
	return INT;
}
ExpNode::Type TimesNode::getType() const {
	return TIMES;
}

yices_expr TimesNode::to_yices(yices_context yctxt) {
#ifdef __TEST
  Assert(get_mother() != NULL, "Plus node should have two parents.");
  Assert(get_father() != NULL, "Plus node should have two parents.");
#endif
  yices_expr ym = get_mother()->to_yices(yctxt);
  yices_expr yf = get_father()->to_yices(yctxt);
  yices_expr args[2] = { ym, yf };
  yices_expr sum = yices_mk_mul(yctxt, args, 2);

  return sum;
}

string TimesNode::get_yicesstr() const {
	stringstream s;
	__mother->ytype = ExpNode::INT;
	s << "( " << "* " << __mother->get_yicesstr() << " ";
	__father->ytype = ExpNode::INT;
	s << __father->get_yicesstr() << " )";
	return s.str();
}

void TimesNode::addVars(set<int>& varset) {
	get_mother()->addVars(varset);
	get_father()->addVars(varset);
}

EnvNode TimesNode::eqNode(ExpNode* other) {
	if (other->getType() == ExpNode::TIMES) {
    EnvNode eq_mothers = get_mother()->eqNode(other->get_mother());
    EnvNode eq_fathers = get_father()->eqNode(other->get_father());
    if ((CONSTVAL == eq_mothers.nodetype) && (eq_mothers.get_val())) {
        return eq_fathers;
    } else {
      if ((CONSTVAL == eq_fathers.nodetype) && (eq_fathers.get_val())) {
          return eq_mothers;
      } else {
        AndNode* an = new AndNode();
        an->set_parents(eq_mothers.get_node(), eq_fathers.get_node());
        return EnvNode(an);
      }
    }
  }
  return EnvNode(false);
}

/**
 * DivNode
 */
DivNode::DivNode() : ExpNode() {
	type = DIV;
}
DivNode::DivNode(const DivNode& bn) :
	ExpNode(bn) {
}
void DivNode::accept(NodeVisitor& visitor) {
	visitor.visit(*this);
}
ExpNode* DivNode::clone(bool copyChildren = true) {
	return new DivNode(*this);
}
;
ExpNode::OutType DivNode::getOtype() const { return INT; }
ExpNode::Type DivNode::getType() const { return DIV; }

string DivNode::get_yicesstr() const {
	stringstream s;
	__mother->ytype = ExpNode::INT;
	s << "( " << "/ " << __mother->get_yicesstr() << " ";
	__father->ytype = ExpNode::INT;
	s << __father->get_yicesstr() << " )";
	return s.str();
}

void DivNode::addVars(set<int>& varset) {
	get_mother()->addVars(varset);
	get_father()->addVars(varset);
}

EnvNode DivNode::eqNode(ExpNode* other) {
	if (other->getType() == ExpNode::DIV) {
    EnvNode eq_mothers = get_mother()->eqNode(other->get_mother());
    EnvNode eq_fathers = get_father()->eqNode(other->get_father());
    if ((CONSTVAL == eq_mothers.nodetype) && (eq_mothers.get_val())) {
        return eq_fathers;
    } else {
      if ((CONSTVAL == eq_fathers.nodetype) && (eq_fathers.get_val())) {
          return eq_mothers;
      } else {
        AndNode* an = new AndNode();
        an->set_parents(eq_mothers.get_node(), eq_fathers.get_node());
        return EnvNode(an);
      }
    }
  }
  return EnvNode(false);
}

/**
 * GtNode
 */
GtNode::GtNode() : ExpNode() {
	type = GT;
	otype = BOOL;
}

GtNode::GtNode(const GtNode& bn): ExpNode(bn) {}

void GtNode::accept(NodeVisitor& visitor) {
	visitor.visit(*this);
}
ExpNode* GtNode::clone(bool copyChildren = true) {
	return new GtNode(*this);
}
;
ExpNode::OutType GtNode::getOtype() const {
	return BOOL;
}
ExpNode::Type GtNode::getType() const { return GT; }

yices_expr GtNode::to_yices(yices_context yctxt) {
#ifdef __TEST
  Assert(get_mother() != NULL, "Plus node should have two parents.");
  Assert(get_father() != NULL, "Plus node should have two parents.");
#endif
  yices_expr ym = get_mother()->to_yices(yctxt);
  yices_expr yf = get_father()->to_yices(yctxt);
  yices_expr sum = yices_mk_gt(yctxt, ym, yf);

  return sum;
}

string GtNode::get_yicesstr() const {
	stringstream s;
	__mother->ytype = ExpNode::INT;
	s << "( " << "> " << __mother->get_yicesstr() << " ";
	__father->ytype = ExpNode::INT;
	s << __father->get_yicesstr() << " )";
	return s.str();
}

void GtNode::addVars(set<int>& varset) {
	__mother->addVars(varset);
	__father->addVars(varset);
}

EnvNode GtNode::eqNode(ExpNode* other) {
	if (other->getType() == ExpNode::GT) {
    EnvNode eq_mothers = get_mother()->eqNode(other->get_mother());
    EnvNode eq_fathers = get_father()->eqNode(other->get_father());
    if ((CONSTVAL == eq_mothers.nodetype) && (eq_mothers.get_val())) {
        return eq_fathers;
    } else {
      if ((CONSTVAL == eq_fathers.nodetype) && (eq_fathers.get_val())) {
          return eq_mothers;
      } else {
        AndNode* an = new AndNode();
        an->set_parents(eq_mothers.get_node(), eq_fathers.get_node());
        return EnvNode(an);
      }
    }
  }
  return EnvNode(false);
}

/**
 * GeNode
 */

GeNode::GeNode() : ExpNode() {
	type = GE;
	otype = BOOL;
}
GeNode::GeNode(const GeNode& bn): ExpNode(bn) {}
void GeNode::accept(NodeVisitor& visitor) {
	visitor.visit(*this);
}
ExpNode* GeNode::clone(bool copyChildren = true) {
	return new GeNode(*this);
}
;
ExpNode::OutType GeNode::getOtype() const {
	return BOOL;
}
ExpNode::Type GeNode::getType() const {
	return GE;
}

yices_expr GeNode::to_yices(yices_context yctxt) {
#ifdef __TEST
  Assert(get_mother() != NULL, "Plus node should have two parents.");
  Assert(get_father() != NULL, "Plus node should have two parents.");
#endif
  yices_expr ym = get_mother()->to_yices(yctxt);
  yices_expr yf = get_father()->to_yices(yctxt);
  yices_expr sum = yices_mk_ge(yctxt, ym, yf);

  return sum;
}

string GeNode::get_yicesstr() const {
	stringstream s;
	__mother->ytype = ExpNode::INT;
	s << "( " << ">= " << __mother->get_yicesstr() << " ";
	__father->ytype = ExpNode::INT;
	s << __father->get_yicesstr() << " )";
	return s.str();
}

void GeNode::addVars(set<int>& varset) {
	get_mother()->addVars(varset);
	get_father()->addVars(varset);
}

EnvNode GeNode::eqNode(ExpNode* other) {
	get_mother()->ytype = ExpNode::INT;
	get_father()->ytype= ExpNode::INT;

	if (other->getType() == ExpNode::GE) {
    EnvNode eq_mothers = get_mother()->eqNode(other->get_mother());
    EnvNode eq_fathers = get_father()->eqNode(other->get_father());
    if ((CONSTVAL == eq_mothers.nodetype) && (eq_mothers.get_val())) {
        return eq_fathers;
    } else {
      if ((CONSTVAL == eq_fathers.nodetype) && (eq_fathers.get_val())) {
          return eq_mothers;
      } else {
        AndNode* an = new AndNode();
        an->set_parents(eq_mothers.get_node(), eq_fathers.get_node());
        return EnvNode(an);
      }
    }
  }
  return EnvNode(false);
}

/**
 * LtNode
 */

LtNode::LtNode() : ExpNode() { type = LT; otype = BOOL; }
LtNode::LtNode(const LtNode& bn): ExpNode(bn) {}
void LtNode::accept(NodeVisitor& visitor) {
	visitor.visit(*this);
}
ExpNode* LtNode::clone(bool copyChildren = true) {
	return new LtNode(*this);
}
;
ExpNode::OutType LtNode::getOtype() const {
	return BOOL;
}
ExpNode::Type LtNode::getType() const {
	return LT;
}

yices_expr LtNode::to_yices(yices_context yctxt) {
#ifdef __TEST
  Assert(get_mother() != NULL, "Plus node should have two parents.");
  Assert(get_father() != NULL, "Plus node should have two parents.");
#endif
  yices_expr ym = get_mother()->to_yices(yctxt);
  yices_expr yf = get_father()->to_yices(yctxt);
  yices_expr sum = yices_mk_lt(yctxt, ym, yf);

  return sum;
}

string LtNode::get_yicesstr() const {
	stringstream s;
	__mother->ytype = ExpNode::INT;
	s << "( " << "< " << __mother->get_yicesstr() << " ";
	__father->ytype = ExpNode::INT;
	s << __father->get_yicesstr() << " )";
	return s.str();
}
void LtNode::addVars(set<int>& varset) {
	get_mother()->addVars(varset);
	get_father()->addVars(varset);
}

EnvNode LtNode::eqNode(ExpNode* other) {
  
	if (other->getType() == ExpNode::LT) {
    EnvNode eq_mothers = get_mother()->eqNode(other->get_mother());
    EnvNode eq_fathers = get_father()->eqNode(other->get_father());
    if ((CONSTVAL == eq_mothers.nodetype) && (eq_mothers.get_val())) {
        return eq_fathers;
    } else {
      if ((CONSTVAL == eq_fathers.nodetype) && (eq_fathers.get_val())) {
          return eq_mothers;
      } else {
        AndNode* an = new AndNode();
        an->set_parents(eq_mothers.get_node(), eq_fathers.get_node());
        return EnvNode(an);
      }
    }
  }
  return EnvNode(false);
}

/**
 * LeNode
 */
LeNode::LeNode() : ExpNode() { type = LE; otype = BOOL; }
LeNode::LeNode(const LeNode& bn): ExpNode(bn) {}
void LeNode::accept(NodeVisitor& visitor) {
	visitor.visit(*this);
}
ExpNode* LeNode::clone(bool copyChildren = true) {
	return new LeNode(*this);
}
;
ExpNode::OutType LeNode::getOtype() const {
	return BOOL;
}
ExpNode::Type LeNode::getType() const {
	return LE;
}

yices_expr LeNode::to_yices(yices_context yctxt) {
#ifdef __TEST
  Assert(get_mother() != NULL, "Plus node should have two parents.");
  Assert(get_father() != NULL, "Plus node should have two parents.");
#endif
  yices_expr ym = get_mother()->to_yices(yctxt);
  yices_expr yf = get_father()->to_yices(yctxt);
  yices_expr sum = yices_mk_le(yctxt, ym, yf);

  return sum;
}

string LeNode::get_yicesstr() const {
	stringstream s;
	__mother->ytype = ExpNode::INT;
	s << "( " << "<= " << __mother->get_yicesstr() << " ";
	__father->ytype = ExpNode::INT;
	s << __father->get_yicesstr() << " )";
	return s.str();
}

void LeNode::addVars(set<int>& varset) {
	get_mother()->addVars(varset);
	get_father()->addVars(varset);
}

EnvNode LeNode::eqNode(ExpNode* other) {
	if (other->getType() == ExpNode::LE) {
    EnvNode eq_mothers = get_mother()->eqNode(other->get_mother());
    EnvNode eq_fathers = get_father()->eqNode(other->get_father());
    if ((CONSTVAL == eq_mothers.nodetype) && (eq_mothers.get_val())) {
        return eq_fathers;
    } else {
      if ((CONSTVAL == eq_fathers.nodetype) && (eq_fathers.get_val())) {
          return eq_mothers;
      } else {
        AndNode* an = new AndNode();
        an->set_parents(eq_mothers.get_node(), eq_fathers.get_node());
        return EnvNode(an);
      }
    }
  }
  return EnvNode(false);
}

/**
 * EqNode
 */
EqNode::EqNode() : ExpNode() { type = ExpNode::EQ; otype = BOOL; }
EqNode::EqNode(const EqNode& bn): ExpNode(bn) {}
void EqNode::accept(NodeVisitor& visitor) {
	visitor.visit(*this);
}
ExpNode* EqNode::clone(bool copyChildren = true) {
	return new EqNode(*this);
}
;
ExpNode::OutType EqNode::getOtype() const {
	return BOOL;
}
ExpNode::Type EqNode::getType() const {
	return ExpNode::EQ;
}

yices_expr EqNode::to_yices(yices_context yctxt) {
#ifdef __TEST
  Assert(get_mother() != NULL, "Plus node should have two parents.");
  Assert(get_father() != NULL, "Plus node should have two parents.");
#endif
  yices_expr ym = get_mother()->to_yices(yctxt);
  yices_expr yf = get_father()->to_yices(yctxt);
  yices_expr sum = yices_mk_eq(yctxt, ym, yf);

  return sum;
}

string EqNode::get_yicesstr() const {
	stringstream s;
	__mother->ytype = ExpNode::INT;
	s << "( " << "= " << __mother->get_yicesstr() << " ";
	__father->ytype = ExpNode::INT;
	s << __father->get_yicesstr() << " )";
	return s.str();
}

void EqNode::addVars(set<int>& varset) {
	get_mother()->addVars(varset);
	get_father()->addVars(varset);
}

EnvNode EqNode::eqNode(ExpNode* other) {
	if (other->getType() == ExpNode::EQ) {
    EnvNode eq_mothers = get_mother()->eqNode(other->get_mother());
    EnvNode eq_fathers = get_father()->eqNode(other->get_father());
    if ((CONSTVAL == eq_mothers.nodetype) && (eq_mothers.get_val())) {
        return eq_fathers;
    } else {
      if ((CONSTVAL == eq_fathers.nodetype) && (eq_fathers.get_val())) {
          return eq_mothers;
      } else {
        AndNode* an = new AndNode();
        an->set_parents(eq_mothers.get_node(), eq_fathers.get_node());
        return EnvNode(an);
      }
    }
  }
  return EnvNode(false);
}

/**
 * NeqNode
 */
NeqNode::NeqNode() : ExpNode() { type = NEQTY; otype = BOOL; }
NeqNode::NeqNode(const NeqNode& bn): ExpNode(bn) {}

void NeqNode::accept(NodeVisitor& visitor) {
	visitor.visit(*this);
}
ExpNode* NeqNode::clone(bool copyChildren = true) {
	return new NeqNode(*this);
}
;
ExpNode::OutType NeqNode::getOtype() const {
	return BOOL;
}
ExpNode::Type NeqNode::getType() const {
	return NEQTY;
}

yices_expr NeqNode::to_yices(yices_context yctxt) {
#ifdef __TEST
  Assert(get_mother() != NULL, "Plus node should have two parents.");
  Assert(get_father() != NULL, "Plus node should have two parents.");
#endif
  yices_expr ym = get_mother()->to_yices(yctxt);
  yices_expr yf = get_father()->to_yices(yctxt);
  yices_expr sum = yices_mk_diseq(yctxt, ym, yf);

  return sum;
}

string NeqNode::get_yicesstr() const {
	stringstream s;
	__mother->ytype = ExpNode::INT;
	s << "( " << "/= " << __mother->get_yicesstr() << " ";
	__father->ytype = ExpNode::INT;
	s << __father->get_yicesstr() << " )";
	return s.str();
}

void NeqNode::addVars(set<int>& varset) {
	get_mother()->addVars(varset);
	get_father()->addVars(varset);
}

EnvNode NeqNode::eqNode(ExpNode* other) {
	if (other->getType() == ExpNode::NEQTY) {
    EnvNode eq_mothers = get_mother()->eqNode(other->get_mother());
    EnvNode eq_fathers = get_father()->eqNode(other->get_father());
    if ((CONSTVAL == eq_mothers.nodetype) && (eq_mothers.get_val())) {
        return eq_fathers;
    } else {
      if ((CONSTVAL == eq_fathers.nodetype) && (eq_fathers.get_val())) {
          return eq_mothers;
      } else {
        AndNode* an = new AndNode();
        an->set_parents(eq_mothers.get_node(), eq_fathers.get_node());
        return EnvNode(an);
      }
    }
  }
  return EnvNode(false);
}

/**
 * Predicated nodes.
 */
TernaryNode::TernaryNode() : ExpNode() {
	Assert(false, "Doesn't make sense to be a TernaryNode without a guard");
  type = TERNARY;
  __guard = NULL;
  otype = BOOL;
}

TernaryNode::TernaryNode(ExpNode* guard) : ExpNode() {
  type = TERNARY;
  __guard = guard;
  __guard->inc_ref_count();
  this->merge_sym(guard);
  otype = BOOL;
}

TernaryNode::TernaryNode(const TernaryNode& bn)
: ExpNode(bn), __guard(bn.__guard) {
	__guard->inc_ref_count();
}

TernaryNode::~TernaryNode() { }

void TernaryNode::accept (NodeVisitor &visitor) { visitor.visit(*this); }

ExpNode* TernaryNode::clone(bool copyChildren) {
  return new TernaryNode(*this);
}

string TernaryNode::lprint(){
  stringstream str;
  str << "(" << __guard->lprint() << " ? " << get_mother()->lprint() << " : "
	  << get_father()->lprint() << ")";
  return str.str();
}

ExpNode::Type TernaryNode::getType() const { return TERNARY; }

yices_expr TernaryNode::to_yices(yices_context yctxt) {
#ifdef __TEST
  Assert(get_mother() != NULL, "Plus node should have two parents.");
  Assert(get_father() != NULL, "Plus node should have two parents.");
#endif
  __guard->ytype = ExpNode::BOOL;
  yices_expr yg = __guard->to_yices(yctxt);

  __mother->ytype = ytype;
  yices_expr ym = get_mother()->to_yices(yctxt);

  __father->ytype = ytype;
  yices_expr yf = get_father()->to_yices(yctxt);
  yices_expr sum = yices_mk_ite(yctxt, yg, ym, yf);

  return sum;
}

string TernaryNode::get_yicesstr() const {
  stringstream str;

  __guard->ytype = ExpNode::BOOL;
  str << "( if " << __guard->get_yicesstr() << " ";

  __mother->ytype = ytype;
  str << __mother->get_yicesstr() << " ";

  __father->ytype = ytype;
  str << __father->get_yicesstr() << " )";
  return str.str();
}

void TernaryNode::addVars(set<int>& varset) {
	__guard->addVars(varset);
	__mother->addVars(varset);
	__father->addVars(varset);
}

EnvNode TernaryNode::eqNode(ExpNode* other) {
	if (other->getType() == ExpNode::TERNARY) {
    EnvNode eq_guards = __guard->eqNode(((TernaryNode*)other)->get_guard());
    EnvNode eq_mothers = get_mother()->eqNode(other->get_mother());
    EnvNode eq_fathers = get_father()->eqNode(other->get_father());

    /* Require the guards to be unconditionally equal. */
    if ((CONSTVAL == eq_guards.nodetype) && (eq_guards.get_val())) {
      if ((CONSTVAL == eq_mothers.nodetype) && (eq_mothers.get_val())) {
        return eq_fathers;
      } else if ((CONSTVAL == eq_fathers.nodetype) && (eq_fathers.get_val())) {
            return eq_mothers;
      } else {
        AndNode* an = new AndNode();
        an->set_parents(eq_mothers.get_node(), eq_fathers.get_node());
        return EnvNode(an);
      }
    }
  }
  return EnvNode(false);
}

/**
 * Equality testing.
 */
bool ExpNode::eqNodes(ExpNode* n1, ExpNode* n2) {
	EnvNode eq = n1->eqNode(n2);
  if ((CONSTVAL == eq.nodetype) && eq.get_val()) {
	  return true;
  } else {
		cout << "exps are not equal" << endl;
		n1->printval();
		n2->printval();
    return false;
	}
}
