/**
 *  This file defines different kinds of Expression nodes.
 */
#ifndef EXPNODES_H_
#define EXPNODES_H_

#include <map>
#include <set>
#include <cstdlib>
#include <typeinfo>
#include <yices_c.h>

#include "BackendTypes.h"
#include "FastSet.h"
#include "NodeVisitor.h"
#include "Dllist.h"


#ifdef CONST
#undef CONST
#endif

using namespace std;

class EnvNode;
class ExpDAG;

class ExpNode {

  private:
    /** The unique ID to be assigned to the next ExpNode created. */
    static int __s_next_global_id;

    int __ref_count;

#ifdef SCHECKMEM
    static set<ExpNode*> allocated;
#endif

  protected:
    ExpNode* __mother;
    ExpNode* __father;
    bool __sym;

    ExpNode();
    ExpNode(const ExpNode& bn);


  public:
    typedef enum{ AND = 0, OR /* 1 */, NOT /* 2 */
                , PLUS /* 3 */, TIMES /* 4 */, DIV /* 5 */, NEG /* 6 */
                , CONST /* 7 */, SYM /* 8 */
                , LIST /* 9 */, RECORD /* 10 */
                , TERNARY /* 11 */
                , GT /* 12 */, GE /* 13 */, LT /* 14 */, LE /* 15 */
                , EQ /* 16 */, NEQTY /* 17 */
                , ASSERT /* 18 */
                , STRING /* 19 */ } Type;
    typedef enum{ BOTTOM
                , BOOL, INT, STRINGTY
                , RELTY, RECORDTY } OutType;

    /**
     * The globally unique ID of this node.
     *
     * (Warning: limits number of nodes to 4Gi).
     */
    int globalId;
    mutable int flag;
    /** 
     * The unique ID of this node within a DAG, not guaranteed to be globally
     * unique.
     * (Implementation detail: the DAG contains a vector of nodes, and this 'id'
     * is this nodes position within that vector.
     */

    Type type;
    mutable OutType otype;
    mutable OutType ytype;

    virtual ~ExpNode();

    /*
     * Getters and setters
     */
    eid_t get_name() { return globalId; }
    ExpNode* get_mother() { return __mother; }
    ExpNode* get_father() { return __father; }
    void set_parents(ExpNode* new_mother, ExpNode* new_father);
    bool sym() { return __sym; }
    void set_sym(bool sym) { __sym = sym; }
    void merge_sym(ExpNode* other_node) { __sym = __sym || other_node->sym(); }

    void inc_ref_count() {
#ifdef __GCCHECK
    	cout << "ExpNode::inc_ref_count: Incrementing for node " << globalId
    		 << "[" << __ref_count + 1 << "]" << endl;
#endif
    	++__ref_count;
    }
    void dec_ref_count() {
#ifdef __GCCHECK
    	cout << "ExpNode::dec_ref_count: Decrementing for node " << globalId
    		 << " [" << __ref_count - 1 << "]" << endl;
#endif
    	--__ref_count;
    }
    int get_ref_count() { return __ref_count; }

    void resetId();
    OutType joinOtype(OutType t1, OutType t2) const;

    virtual int do_dfs(int idx);
    virtual int back_dfs(int idx);
    virtual void replace_parent(const ExpNode * oldpar, ExpNode* newpar);
    virtual void outDagEntry(ostream& out) const;

    virtual string get_tname() const;
    virtual string get_sym() const;
    virtual void printval();
    virtual string lprint();
    virtual string lid();

    /* Converting to a Yices expression. */
    virtual yices_expr to_yices(yices_context yctxt);

    /**
     * Serialization.
     */
    virtual string get_namestr() const;
    virtual string get_yicesstr() const;

    virtual void accept(NodeVisitor& visitor) =0;
    virtual ExpNode* clone(bool copyChildren = true)=0;
    virtual void printSubDAG(ostream& out, set<const ExpNode* >& s)const;
    virtual OutType getOtype() const;
    virtual Type getType() const;
    void neighbor_replace(ExpNode* replacement);

    virtual void addVars(set<int>& varset);
    virtual EnvNode eqNode(ExpNode* other);
    static bool eqNodes(ExpNode* n1, ExpNode* n2);

#ifdef SCHECKMEM
    static void show_still_live() {
    	cout << "***" << endl;
    	cout << "STILL LIVE:" << endl;
    	set<ExpNode*>::iterator it;
    	for (it = allocated.begin(); it != allocated.end(); ++it) {
    		cout << "ID: " << (*it)->globalId << " Type: " << (*it)->getType();
    		cout << " Sym: " << (*it)->sym() << endl;
    		cout << " Ref count: " << (*it)->get_ref_count();
    		cout << endl << (*it)->lprint() << endl;
    	}
    	cout << "***" << endl;
    }
#endif
};

/**
 *	Union type
 */
typedef enum enkind { CONSTVAL, EXPNODE } enkind_t;
class EnvNode {

public:
	enkind_t nodetype;
	union {
		int constVal;
		ExpNode* node;
	} envNodeData;

	EnvNode(int val) {
		nodetype = CONSTVAL;
		envNodeData.constVal = val;
	}
	EnvNode(ExpNode* exp) {
		nodetype = EXPNODE;
		envNodeData.node = exp;
	}

	int get_val() { return envNodeData.constVal; }
	ExpNode* get_node() { return envNodeData.node; }
	void set_node(ExpNode* node) { envNodeData.node = node; }
  bool eq(EnvNode other);
};

/**
 * Special FastSet hash function that hashes on the
 * ExpNode's ID (assumed unique) instead of its memory address.
 */
template<>
struct FastSetTraits<ExpNode>
{
  static inline unsigned
    hash (ExpNode *bn, int sz)
    {
      unsigned tmp = bn->globalId; // (unsigned) in>>2;  // (generic version)

      tmp = tmp * (tmp + 4297);
      tmp = tmp + (tmp >> 2) + (tmp >> 5) + (tmp >> 21) + (tmp >> 28);		

      unsigned m = 1 << sz;
      m = m-1;
      return (tmp & m)<<2;
    }
};

class AndNode: public ExpNode{ 	
  public: 
    AndNode(const AndNode& b);
    AndNode();
    virtual void accept(NodeVisitor& visitor);
    virtual ExpNode* clone(bool copyChildren);
    OutType getOtype() const { return BOOL; }
    Type getType() const { return AND; }
    void addVars(set<int>& varset);
    EnvNode eqNode(ExpNode* other);

    virtual yices_expr to_yices(yices_context yctxt);
    string get_yicesstr() const;
};

class OrNode: public ExpNode{	
  public: 
    OrNode();
    OrNode(const OrNode& bn);
    virtual void accept(NodeVisitor& visitor);
    virtual ExpNode* clone(bool copyChildren = true);
    OutType getOtype()const { return BOOL; }
    Type getType() const { return OR; }
    void addVars(set<int>& varset);
    EnvNode eqNode(ExpNode* other);

    virtual yices_expr to_yices(yices_context yctxt);
    string get_yicesstr() const;
};

/* TODO */
class NOT_node: public ExpNode{	
  public: 
    NOT_node(){ type = NOT; }  
    NOT_node(const NOT_node& bn): ExpNode(bn){ }  
    virtual void accept(NodeVisitor& visitor)  { visitor.visit( *this ); }
    virtual ExpNode* clone(bool copyChildren = true){return new NOT_node(*this);  };
    OutType getOtype() const { return BOOL; }
    Type getType() const { return NOT; }
    virtual string lid(){
      stringstream str;
      str<<"(!"<<get_mother()->lid()<<")";
      return str.str();
    }
    void addVars(set<int>& varset) { __mother->addVars(varset); }
    EnvNode eqNode(ExpNode* other) {
		if (other->getType() == NOT) {
			return get_mother()->eqNode(other->get_mother());
		} else {
			return false;
		}
    }
};

class PlusNode: public ExpNode{	
  private:
    bool __is_minus;
  public:
    PlusNode();
    PlusNode(bool minus);
    PlusNode(const PlusNode& bn);

    void set_minus(bool minus);
    bool is_minus() { return __is_minus; }

    virtual void accept(NodeVisitor& visitor);
    virtual ExpNode* clone(bool copyChildren);
    OutType getOtype()const;
    Type getType() const;
    virtual string lprint();
   
    virtual yices_expr to_yices(yices_context yctxt);
    string get_yicesstr() const;
   
    void addVars(set<int>& varset);
    EnvNode eqNode(ExpNode* other);
};

class TimesNode: public ExpNode {
public:
	TimesNode();
	TimesNode(const TimesNode& bn);
	virtual void accept(NodeVisitor& visitor);
	virtual ExpNode* clone(bool copyChildren);
	OutType getOtype() const;
	Type getType() const;

  virtual yices_expr to_yices(yices_context yctxt);
  string get_yicesstr() const;

	void addVars(set<int>& varset);
	EnvNode eqNode(ExpNode* other);
};

class DivNode: public ExpNode{	
  public: 
    DivNode();
    DivNode(const DivNode& bn);
    virtual void accept(NodeVisitor& visitor);
    virtual ExpNode* clone(bool copyChildren);
    OutType getOtype()const;
    Type getType() const;
   
    string get_yicesstr() const;

    void addVars(set<int>& varset);
    EnvNode eqNode(ExpNode* other);
};

/* TODO: What does NEG_node do? */
class NEG_node: public ExpNode{	

  public: 
    NEG_node(){ type = NEG; }
    NEG_node(const NEG_node& bn): ExpNode(bn){ }
    virtual void accept(NodeVisitor& visitor)  { visitor.visit( *this ); }
    virtual ExpNode* clone(bool copyChildren = true){return new NEG_node(*this);  };
    OutType getOtype()const { return INT; }
    Type getType() const { return NEG; }
};


class ConstNode: public ExpNode{ 
  int __val;
  public:
  ConstNode();
  ConstNode(int n);
  ConstNode(const ConstNode& bn);
  virtual void accept(NodeVisitor& visitor) { visitor.visit( *this ); }
  virtual void setVal(int v) { __val = v; }
  virtual int getVal() const { return __val; }

  string get_namestr() const;

  virtual yices_expr to_yices(yices_context yctxt);
  string get_yicesstr() const;

  virtual string lid();
  virtual string lprint();
  virtual ExpNode* clone(bool copyChildren = true);
  OutType getOtype()const { return (otype == BOTTOM) ? INT : otype; }
  Type getType() const { return CONST; }
  EnvNode eqNode(ExpNode* other);
};

class StringNode: public ExpNode{
  string __val;

  public:
  StringNode();
  StringNode(string n);
  StringNode(const StringNode& bn);
  virtual void accept(NodeVisitor& visitor) { visitor.visit( *this ); }
  virtual void setVal(string v) { __val = v; }
  virtual const string& getVal() { return __val; }

  string get_namestr() const;

  virtual yices_expr to_yices(yices_context yctxt);
  string get_yicesstr() const;

  virtual string lid();
  virtual string lprint();
  virtual ExpNode* clone(bool copyChildren = true);
  OutType getOtype()const { return (otype == BOTTOM) ? STRINGTY : otype; }
  Type getType() const { return STRING; }
  EnvNode eqNode(ExpNode* other);
};

typedef std::vector<EnvNode> record_t;
class RecordNode: public ExpNode{
 private:
  record_t* __record;
  size_t __num_fields;

 public:
  RecordNode();
  ~RecordNode();
  RecordNode(size_t numfields);
  RecordNode(const RecordNode& bn);
  virtual void accept(NodeVisitor& visitor);
  string get_namestr() const;
  virtual string lid();
  virtual string lprint();
  virtual ExpNode* clone(bool copyChildren = true);

  OutType getOtype() const { return RECORDTY; }
  Type getType() const { return RECORD; }

  record_t* get_record() { return __record; }

  EnvNode lookupField(size_t field) { return (*__record)[field]; }
  void setValue(size_t field, EnvNode exp);

  void resize(int newsize);
};

typedef vector<ExpNode*> relation_t;
typedef vector<ExpNode*>::iterator relation_it;
class ListNode: public ExpNode{
 private:
  relation_t __rel;
  // Relation mapped with primary keys.
  map<int, relation_t*> __mapped_rel;

 public:
  ListNode();
  ~ListNode();
  ListNode(const ListNode& bn);
 
  virtual void accept(NodeVisitor& visitor);
  string get_namestr() const;
  virtual string lid();
  virtual string lprint();
  virtual ExpNode* clone(bool copyChildren = true);
  OutType getOtype() const { return RELTY; }
  Type getType() const { return LIST; }

  /* Iterating. */
  relation_it begin() { return __rel.begin(); }
  relation_it end() { return __rel.end(); }
  relation_it pkey_begin(int pkey) { return __mapped_rel[pkey]->begin(); }
  relation_it pkey_end(int pkey) { return __mapped_rel[pkey]->end(); }

  /* Looking up record values. */
  EnvNode lookupRelationVal(int pkey, int field);
  EnvNode lookupSymVal(SymNode* pkey, size_t field);
  void addExpNode(ExpNode* record);

  EnvNode hasElem(ExpNode* en);
};

/* TODO: Move this to C file. */
class InterNode: public ExpNode{
  protected:
    InterNode(const InterNode& bn);
    InterNode();
    int nbits;
  public:
    virtual ~InterNode(){}
    int get_nbits() const;
    void set_nbits(int n);
    OutType getOtype() const;
    Type getType() const { Assert(false, "Inter node?"); }
    string get_namestr() const;
    virtual string lprint();
    virtual string lid();
};

class SymNode: public ConstNode{
private:
	int __frontend_var;
	/* TODO: Store a private variable here? */
  public:
    SymNode();
    SymNode(int fvar);
    SymNode(const SymNode& bn);
    virtual void accept(NodeVisitor& visitor);
    virtual ExpNode* clone(bool copyChildren = true);
    Type getType() const { return SYM; }
    OutType getOtype() const { return INT; }

    int get_frontend_var();

    string lprint();

    virtual yices_expr to_yices(yices_context yctxt);
    string get_yicesstr() const;

    void addVars(set<int>& vars);
    EnvNode eqNode(ExpNode* other);
};

class GtNode: public ExpNode{	
  public: 
    GtNode();
    GtNode(const GtNode& bn);
    virtual void accept(NodeVisitor& visitor);
    virtual ExpNode* clone(bool copyChildren);
    OutType getOtype() const;
    Type getType() const;

    virtual yices_expr to_yices(yices_context yctxt);
    string get_yicesstr() const;

    void addVars(set<int>& varset);
    EnvNode eqNode(ExpNode* other);
};

class GeNode: public ExpNode{	
  public: 
    GeNode();
    GeNode(const GeNode& bn);
    virtual void accept(NodeVisitor& visitor);
    virtual ExpNode* clone(bool copyChildren);
    OutType getOtype() const;
    Type getType() const;

    virtual yices_expr to_yices(yices_context yctxt);
    string get_yicesstr() const;

    void addVars(set<int>& varset);
    EnvNode eqNode(ExpNode* other);
};

class LtNode: public ExpNode{	
  public: 
    LtNode();
    LtNode(const LtNode& bn);
    virtual void accept(NodeVisitor& visitor);
    virtual ExpNode* clone(bool copyChildren);
    OutType getOtype() const;
    Type getType() const;

    virtual yices_expr to_yices(yices_context yctxt);
    string get_yicesstr() const;

    void addVars(set<int>& varset);
    EnvNode eqNode(ExpNode* other);
};

class LeNode: public ExpNode{	
  public: 
    LeNode();
    LeNode(const LeNode& bn);
    virtual void accept(NodeVisitor& visitor);
    virtual ExpNode* clone(bool copyChildren);
    OutType getOtype()const;
    Type getType() const;

    virtual yices_expr to_yices(yices_context yctxt);
    string get_yicesstr() const;

    void addVars(set<int>& varset);
    EnvNode eqNode(ExpNode* other);
};

class EqNode: public ExpNode{	
  public: 
    EqNode();
    EqNode(const EqNode& bn);
    virtual void accept(NodeVisitor& visitor);
    virtual ExpNode* clone(bool copyChildren);
		OutType getOtype() const;
    Type getType() const;

    virtual yices_expr to_yices(yices_context yctxt);
    string get_yicesstr() const;

    void addVars(set<int>& varset);
    EnvNode eqNode(ExpNode* other);
};

class NeqNode: public ExpNode{
  public:
    NeqNode();
    NeqNode(const NeqNode& bn);
    virtual void accept(NodeVisitor& visitor);
    virtual ExpNode* clone(bool copyChildren);
		OutType getOtype() const;
    Type getType() const;

    virtual yices_expr to_yices(yices_context yctxt);
    string get_yicesstr() const;

    void addVars(set<int>& varset);
    EnvNode eqNode(ExpNode* other);
};

/* Predicated nodes. */
class TernaryNode: public ExpNode {
 public:
  TernaryNode ();
  TernaryNode(const TernaryNode& bn);
  TernaryNode(ExpNode* guard);
  virtual ~TernaryNode();

  /* TODO: Make destructor? */
  Type getType() const;

  virtual yices_expr to_yices(yices_context yctxt);
  string get_yicesstr() const;

  ExpNode* get_guard() { return __guard; }

  virtual void accept (NodeVisitor &visitor);
  virtual ExpNode* clone(bool copyChildren = true);

  virtual string lprint();

  void addVars(set<int>& varset);
  EnvNode eqNode(ExpNode* other);

 private:
  ExpNode* __guard;
};

/* TODO: Change function name. */
/*
     typedef enum{ AND, OR, NOT
                , PLUS, TIMES, DIV, NEG
                , CONST, BOOLCONST, SYM, LIST, RECORD
                , TERNARY
                , GT, GE, LT, LE, EQ, NEQTY
                , ASSERT} Type;
*/
inline ExpNode* newBoolNode(ExpNode::Type type) {
	switch (type) {
	case ExpNode::AND:
		return new AndNode();
	case ExpNode::OR:
		return new OrNode();
	case ExpNode::NOT:
		return new NOT_node();
	case ExpNode::PLUS:
		return new PlusNode();
	case ExpNode::TIMES:
		return new TimesNode();
	case ExpNode::DIV:
		return new DivNode();
	case ExpNode::NEG:
		return new NEG_node();
	case ExpNode::CONST:
		return new ConstNode();
	case ExpNode::SYM:
		return new SymNode();
	case ExpNode::LIST:
		return new ListNode();
	case ExpNode::RECORD:
		return new RecordNode();
	case ExpNode::TERNARY:
		return new TernaryNode();
	case ExpNode::GT:
		return new GtNode();
	case ExpNode::GE:
		return new GeNode();
	case ExpNode::LT:
		return new LtNode();
	case ExpNode::LE:
		return new LeNode();
	case ExpNode::EQ:
		return new EqNode();
  case ExpNode::STRING:
    return new StringNode();
	default:
		Assert(false, "Creating node of unknown type.");
	}
}

inline bool isDllnode(ExpNode* bn){
	return bn->type == ExpNode::ASSERT;
}

inline DllistNode* getDllnode(ExpNode* bn){
	return dynamic_cast<DllistNode*>(bn);
}
#endif
