#include <iostream>

#include "backend_defs.h"
#include "BackendIface.h"
#include "BasicError.h"
#include "ExpDAG.h"
#include "ExpDAGCreator.h"
#include "ExpressionNodes.h"

using namespace std;

extern "C" void createConstNode(int c);
extern "C" void createSymNode(int varname);
extern "C" void createBoolNode(int val);
extern "C" void createCondNode();
extern "C" void createArithopNode(arithop_t op);
extern "C" void createBoolopNode(boolop_t op);
extern "C" void aliasBackendVal(int varname, int origname);
extern "C" void bindBackendVal(int varname);
extern "C" void lookupBackendVal(int varname);
extern "C" int getEnvConcreteBool();
extern "C" void readExpRelation(int numfields, char* file);
extern "C" void mkEnvRelation(int name, int numfields);
extern "C" void bindEnvRelationVal(int rname, int pkey, int field);
extern "C" void lookupEnvRelationVal(int rname, int pkey, int field);
extern "C" void lookupEnvFieldIndex(int field);
extern "C" int getRelIterator();
extern "C" void getNextRecord(int relit);
extern "C" bool eqExpsInEnv();
extern "C" bool eqExpToValInEnv(int val);
extern "C" void assertExpEnv();
extern "C" void concretizeEnvVal(int vname, int val);

extern "C" void writeEnvAssertions();
extern "C" void printEnvVars();
extern "C" void printEnvExp();

extern "C" void initBackendEnv();
extern "C" void clearBackendEnv();

int next_sym_name = 0;

/**
 * Testing individual nodes.
 */
void test_const() {
	cout << endl;
	cout << "test_const: Testing constant node." << endl;
	cout << "---" << endl;

	Assert(BackendIface::env->stack_size() == 0, "Starting with empty stack.");
	createConstNode(3);
	Assert(BackendIface::env->stack_size() == 1, "Stack size 1.");
	createConstNode(4);
	Assert(BackendIface::env->stack_size() == 2, "Stack size 1.");
	createConstNode(0);
	Assert(BackendIface::env->stack_size() == 3, "Stack size 1.");

	ExpNode* cur = BackendIface::env->pop_node();
	Assert( cur->getType() == ExpNode::CONST
		  , "Constant node should have type CONST" );
	Assert( ((ConstNode*)cur)->getVal() == 0
		  , "Constant node should have the right value.");

	cur = BackendIface::env->pop_node();
	Assert( cur->getType() == ExpNode::CONST
			, "Constant node should have type CONST" );
	Assert( ((ConstNode*)cur)->getVal() == 4
			, "Constant node should have the right value.");


	cur = BackendIface::env->pop_node();
	Assert( cur->getType() == ExpNode::CONST
			, "Constant node should have type CONST" );
	Assert( ((ConstNode*)cur)->getVal() == 3
			, "Constant node should have the right value.");

	Assert(cur->get_ref_count() == 0, "No more references to current node.");

	Assert(BackendIface::env->stack_size() == 0, "Ending with empty stack.");
	delete cur;
}

void test_boolop() {
	cout << endl;
	cout << "test_boolop: Testing boolean nodes." << endl;
	cout << "---" << endl;

	Assert(BackendIface::env->stack_size() == 0, "Beginning with empty stack.");

	createConstNode(0);
	createConstNode(1);
	Assert(BackendIface::env->stack_size() == 2, "Have 2 stack items before we make a boolop");
	createBoolopNode(EQ);
	BackendIface::env->pop_node();

	Assert(BackendIface::env->stack_size() == 0, "Ending with empty stack.");
}

void test_sym() {
	cout << endl;
	cout << "test_sym: Testing symbolic nodes." << endl;
	cout << "---" << endl;

	Assert(BackendIface::env->stack_size() == 0, "Beginning with empty stack.");

	int symname = next_sym_name++;
	createSymNode(symname);
	Assert(BackendIface::env->stack_size() == 1, "Have 1 stack item");
	ExpNode* sym_node = BackendIface::env->peek_node();

	Assert( sym_node->getType() == ExpNode::SYM
		  , "test_sym: Should get sym node back." );
	Assert( ((SymNode*)sym_node)->get_frontend_var() == symname
		  , "test_sym: Should get same varname back." );
	Assert( sym_node->sym(), "Sym node should be symbolic." );

	createConstNode(42);
	ExpNode* const_node = BackendIface::env->peek_node();
	Assert( !const_node->sym(), "Const node should not be symbolic" );

	/* Test creating arith node. */
	createArithopNode(PLUS);
	ExpNode* sum_exp = BackendIface::env->pop_node();
	Assert( sum_exp->getType() == ExpNode::PLUS
		  , "test_sym: Should get unevaluated sum back." );
	Assert( sum_exp->get_mother()->getType() == ExpNode::SYM
		  , "test_sym: Left operand should be symbolic." );
	Assert( sum_exp->sym(), "Plus node should be symbolic" );

	/* Test creating boolop. */
	symname = next_sym_name++;
	createSymNode(symname);
	createConstNode(42);
	createBoolopNode(EQ);
	ExpNode* eq_exp = BackendIface::env->peek_node();
	cout << eq_exp->lprint() << endl;
	Assert( eq_exp->sym(), "Eq node should be symbolic" );

	assertExpEnv();

	/* Test creating conditional. */
	/* TODO */
	/*
	eid_t cond_eid = createCondNode(eq_eid, sum_eid, sum_eid);
	ExpNode* cond_exp = BackendIface::env->get_node(cond_eid);
	Assert(cond_exp->sym(), "Cond exp should be symbolic");

	eid_t true_eid = createBoolNode(true);
	eid_t cond_eid2 = createCondNode(true_eid, sum_eid, sum_eid);
	ExpNode* cond_exp2 = BackendIface::env->get_node(cond_eid2);
	Assert(cond_exp2->sym(), "Cond exp 2 should be symbolic");
	*/
}

void test_relation() {
	cout << endl;
	cout << "test_relation: Testing relation node." << endl;
	cout << "---" << endl;

	BackendIface::env->readRelationNode(2, "test.txt");
	ExpNode* cur_node = BackendIface::env->pop_node();
	Assert( cur_node->getType() == ExpNode::RELATION
		  , "Created a relation" );
	RelationNode* rel_node = (RelationNode*)cur_node;
	Assert( rel_node->lookupRelationVal(0, 0)->getType() == ExpNode::CONST
		  , "relation field");
	Assert( rel_node->lookupRelationVal(0, 1)->getType() == ExpNode::CONST
		  , "relation field");
	Assert( rel_node->sym()
		  , "relation is symbolic");
}

void test_relation_iter() {
	cout << endl;
	cout << "test_relation_iter: Testing relation iterators." << endl;
	cout << "---" << endl;

	BackendIface::env->readRelationNode(2, "test.txt");
	ExpNode* cur_node = BackendIface::env->pop_node();
	Assert( cur_node->getType() == ExpNode::RELATION
		  , "Created a relation" );
	RelationNode* rel_node = (RelationNode*)cur_node;

	int it_id = rel_node->getIterator();
	RecordNode* cur_rec = rel_node->nextRecord(it_id);
	Assert( cur_rec->lookupField(0)->getType() == ExpNode::CONST
		  , "test_relation_iter: val should match" );
	Assert( cur_rec->lookupField(1)->getType() == ExpNode::CONST
		  , "test_relation_iter: val should match" );

	cur_rec = rel_node->nextRecord(it_id);
	Assert( cur_rec->lookupField(0)->getType() == ExpNode::CONST
			, "test_relation_iter: val should match" );
	Assert( cur_rec->lookupField(1)->getType() == ExpNode::SYM
			, "test_relation_iter: val should match" );

	cur_rec = rel_node->nextRecord(it_id);
	Assert( cur_rec->lookupField(0)->getType() == ExpNode::CONST
			, "test_relation_iter: val should match" );
	Assert( cur_rec->lookupField(1)->getType() == ExpNode::CONST
			, "test_relation_iter: val should match" );

	cur_rec = rel_node->nextRecord(it_id);
	Assert( cur_rec->lookupField(0)->getType() == ExpNode::CONST
			, "test_relation_iter: val should match" );
	Assert( cur_rec->lookupField(1)->getType() == ExpNode::CONST
			, "test_relation_iter: val should match" );

	cur_rec = rel_node->nextRecord(it_id);
	Assert( cur_rec == NULL
		  , "test_relation_iter: Should have reached end of the relation." );

	/* Test double iterator. */
	int outer_id = rel_node->getIterator();
	RecordNode* outer_rec = rel_node->nextRecord(outer_id);
	int count = 0;
	for (; outer_rec != NULL; outer_rec = rel_node->nextRecord(outer_id)) {
		int inner_id = rel_node->getIterator();
		RecordNode* inner_rec = rel_node->nextRecord(inner_id);
		for(; inner_rec != NULL; inner_rec = rel_node->nextRecord(inner_id)) {
			++count;
		}
	}
	cout << "test_relation_iter: looped " << count << " times." << endl;
	Assert( count == 16, "Should have gone around the loop 16 times." );
}

/**
 * Tests the DAG functionality.
 */
void test_DAG() {
	cout << endl;
	cout << "test_DAG: Testing DAG." << endl;
	cout << "---" << endl;

	ExpDAG* dag = new ExpDAG();
	ExpNode* new_node = new ConstNode();
	dag->addNewNode(new_node);

	/* TODO */
	/*
	ExpNode* retrieved_node = dag->get_node(new_node->get_name());
	Assert( new_node->get_name() == retrieved_node->get_name()
		  , "test_DAG: node we store should be same as one we get back");
		  */
	delete new_node;
	delete dag;
}

void test_DAGCreator() {
	cout << endl;
	cout << "test_DAGCreator: Testing indirect DAG manipulation." << endl;
	cout << "---" << endl;

	ExpDAG* dag = new ExpDAG();
	ExpDAGCreator* dagc = new ExpDAGCreator(dag);

	/* Test the two ways we can make new nodes. */
	/* Way 1: By passing an ExpNode*. */
	ConstNode* new_node = new ConstNode(0);
	ExpNode* cur1 = dagc->new_node(NULL, NULL, new_node);
	Assert ( cur1->getType() == ExpNode::CONST
		   , "test_DAGCreator: created node has the right type" );
	cout << ((ConstNode*)cur1)->getVal() << endl;
	Assert ( ((ConstNode*)cur1)->getVal() == 0
		   , "test_DAGCreator: Node should have returned unchanged.");
	int node_eid = cur1->get_name();

	/* Make sure we're getting the right thing back from the environment. */
	/* ExpNode* dag_node = dag->get_node(node_eid);
	Assert( dag_node == cur1
		  , "test_DAGCreator: Should be getting right node back when creating via node.");
	ExpNode* dag_creator_node = dagc->get_node(node_eid);
	Assert( dag_creator_node == cur1
		  , "test_DAGCreator: Should be getting right node back when creating via node.");
	*/

	if (new_node != cur1) { delete cur1; }
	delete new_node;
	delete dag;
	delete dagc;
}

/***
 * Interface function tests.
 */
void test_iface_bind_lookup() {
	cout << endl;
	cout << "test_iface_bind_lookup: Testing binding, aliasing, and lookup." << endl;
	cout << "---" << endl;

	createConstNode(0);
	ExpNode* created = BackendIface::env->peek_node();
	int createdval = ((ConstNode*)created)->getVal();

	/* Bind and lookup. */
	bindBackendVal(0);
	lookupBackendVal(0);
	ExpNode* cur = BackendIface::env->pop_node();
	Assert(createdval == ((ConstNode*)cur)->getVal()
		  , "test_iface_bind_lookup: Should get var back when bound.");
}

void test_iface_relation() {
	cout << endl;
	cout << "test_iface_relation: Testing relation and record nodes." << endl;
	cout << "---" << endl;

	int name = 0;
	int numfields = 1;
	mkEnvRelation(name, numfields);
	/* NOTE: This puts the relation onto the stack. */
	ExpNode* cur = BackendIface::env->pop_node();

	Assert( BackendIface::env->hasExpByName(name)
		  , "test_iface_relation: Relation name should be bound.");

	/* TODO: Garbage collect. */
}

/*
 * 	extern "C" bool eqExpsInEnv(eid_t eid1, eid_t eid2);
	extern "C" bool eqExpToValInEnv(int val, eid_t eid);
 */
void test_iface_equality() {
	cout << endl;
	cout << "test_iface_equality: Testing equality functions." << endl;
	cout << "---" << endl;

	createConstNode(0);
	createConstNode(0);
	/* eid_t eid2 = createConstNode(1);
	eid_t eid3 = createSymNode(next_sym_name++); */

	/* TODO */
	Assert( eqExpsInEnv()
		  , "test_iface_equality: expressions should be equal in env.");
	/*
	Assert( eqExpToValInEnv(0, eid0)
		  , "test_iface_equality: expression should be equal to value in env");
	Assert( !eqExpsInEnv(eid0, eid2)
		  , "test_iface_equality: expressions should not be equal" );
	Assert( !eqExpsInEnv(eid0, eid3)
		  , "test_iface_equality: expressions should not be equal" );
	*/
	/* TODO: Garbage collect. */
}

void test_concrete_bool() {
	cout << endl;
	cout << "test_concrete_bool: Testing getting concrete boolean values from "
		 << "the back-end." << endl;
	cout << "---" << endl;

	Assert(BackendIface::env->stack_size() == 0, "Beginning with empty stack.");

	createBoolNode(false);
	createBoolNode(true);

	ExpNode* true_node = BackendIface::env->pop_node();
	Assert( ((ConstNode*)true_node)->getVal() == 1
		  , "True node." );

	ExpNode* false_node = BackendIface::env->pop_node();
	Assert( ((ConstNode*)false_node)->getVal() == 0
		  , "False node should have been first thing we created." );

	Assert( true_node->getType() == ExpNode::CONST
		  , "Node has constant type" );

	createSymNode(0);
	BackendIface::env->push_node(true_node);
	createBoolopNode(AND);
	Assert( getEnvConcreteBool() == -1
		  , "test_concrete_bool: Should not get concrete EID back." );
}

void test_iface_concretize() {
	cout << endl;
	cout << "test_iface_concretize: Test concretization." << endl;
	cout << "---" << endl;
	// concretizeEnvVal(eid_t eid, int val)

	createSymNode(next_sym_name++);
	ExpNode* start_node = BackendIface::env->peek_node();
	Assert ( start_node->getType() == ExpNode::SYM
		   , "test_iface_concretize: Node should have sym type" );

	bindBackendVal(1000);
	concretizeEnvVal(0, 3);

	BackendIface::env->getExpByName(0+1000);
	ExpNode* result_node = BackendIface::env->pop_node();
	Assert ( !result_node->sym(), "Node should be concrete now." );
}

/**
 * Testing common sub-expression elimination.
 */
void test_cse() {
	cout << endl;
	cout << "test_cse: Testing common sub-expression elimination." << endl;
	cout << "---" << endl;

	ExpDAG* dag = new ExpDAG();
	ExpDAGCreator* dagc = new ExpDAGCreator(dag);

	ConstNode* new_node = new ConstNode(2);
	ExpNode* opt_node = dagc->new_node(NULL, NULL, new_node);

	ConstNode* same_node = new ConstNode(2);
	ExpNode* same_opt = dagc->new_node(NULL, NULL, same_node);

	Assert(opt_node == same_opt, "Optimized constants should be same.");

	delete same_opt;
	delete dag;
	delete dagc;
}

void test_if_floating() {
	cout << endl;
	cout << "test_if_floating: Testing optimization for ternary nodes." << endl;
	cout << "---" << endl;

	ExpDAG* dag = new ExpDAG();
	ExpDAGCreator* dagc = new ExpDAGCreator(dag);

	/***************************/
	/* Simple if optimization. */
	/***************************/
	/* Make ternary node. */
	ExpNode* cond_node = new ConstNode(0);
	ConstNode* left_opr = new ConstNode(2);
	ConstNode* right_opr = new ConstNode(3);
	TernaryNode* if_node = new TernaryNode(cond_node);

	/* Now optimize. */
	ExpNode* opt_node = dagc->new_node(left_opr, right_opr, if_node);
	Assert(opt_node == right_opr, "Should be just false branch.");
	delete opt_node;

	/**********************************/
	/* Floating one branch at a time. */
	/**********************************/
	/* Make ternary node. */
	PlusNode* plus_node = new PlusNode();

	cond_node = new SymNode();
	left_opr = new ConstNode(2);
	right_opr = new ConstNode(3);
	if_node = new TernaryNode(cond_node);
	if_node->set_parents(left_opr, right_opr);

	/* Now optimize. */
	opt_node = dagc->new_node(if_node, right_opr, plus_node);
	Assert(opt_node->getType() == ExpNode::TERNARY, "If should have floated");
	Assert(opt_node->get_mother()->getType() == ExpNode::CONST, "Floating if left");
	Assert(opt_node->get_father()->getType() == ExpNode::CONST, "Floating if right");
	// Assert(opt_node == right_opr, "Should be just false branch.");
	delete opt_node;

	/***************************/
	/* Floating both branches. */
	/***************************/
	/* Make ternary node. */
	plus_node = new PlusNode();

	cond_node = new SymNode();
	left_opr = new ConstNode(2);
	right_opr = new ConstNode(3);
	if_node = new TernaryNode(cond_node);
	if_node->set_parents(left_opr, right_opr);

	/* Now optimize. */
	opt_node = dagc->new_node(if_node, if_node, plus_node);
	Assert(opt_node->getType() == ExpNode::TERNARY, "If should have floated");
	Assert(opt_node->get_mother()->getType() == ExpNode::TERNARY, "Floating if left");
	Assert(opt_node->get_father()->getType() == ExpNode::TERNARY, "Floating if right");
	delete opt_node;

	/*************************/
	/* Multi-level floating. */
	/*************************/
	/* Make ternary node. */
	plus_node = new PlusNode();

	cond_node = new SymNode();
	left_opr = new ConstNode(2);
	TernaryNode* right_if = new TernaryNode(cond_node);
	right_if->set_parents(left_opr, left_opr);

	if_node = new TernaryNode(cond_node);
	if_node->set_parents(left_opr, right_if);

	/* Now optimize. */
	opt_node = dagc->new_node(left_opr, if_node, plus_node);
	Assert(opt_node->getType() == ExpNode::TERNARY, "If should have floated");
	Assert(opt_node->get_father()->getType() == ExpNode::TERNARY, "Floating if right");
	delete opt_node;

	delete dag;
	delete dagc;
}

int main () {
	cout << "Running tests..." << endl;
	initBackendEnv();

	/* Single node tests. */
	cout << "*** Testing expression nodes. ***" << endl;
	test_const();
	test_boolop();
	test_sym();
	test_relation();
	test_relation_iter();
	cout << endl << endl;

	/* Infrastructure tests. */
	cout << "*** Testing DAG infrastructure. ***" << endl;
	test_DAG();
	test_DAGCreator();
	cout << endl << endl;

	/* Interface function tests. */
	cout << "*** Testing interface functions. ***" << endl;
	test_iface_bind_lookup();
	test_iface_relation();
	test_iface_equality();
	test_concrete_bool();
	test_iface_concretize();
	cout << endl << endl;

	/* Optimization tests. */
	/*
	cout << "*** Testing optimizations. ***" << endl;
	test_cse();
	test_if_floating();
	cout << endl << endl;
	*/

	/* Garbage collection tests. */
	cout << "*** Testing garbage collection. ***" << endl;


	return 0;
}
