#include "backend_defs.h"
#include "BackendIface.h"
#include "BackendTypes.h"
#include "BasicError.h"
#include "ExpressionNodes.h"

/**
 * Constant nodes.
 */
extern "C" eid_t createConstNode(int c) {
#ifdef __TEST
  Assert(BackendIface::getEnv() != NULL, "Backend env not null.");
  Assert(BackendIface::getEnv()->getDAG() != NULL, "Backend env dag not null.");
#endif

  ExpNode* cur = BackendIface::getEnv()->new_node(NULL, NULL, ExpNode::CONST);
#ifdef __DEBUG_1
  cout << "BackendIface: Creating constant node with ID "
       << cur->get_name() << endl;
#endif // __DEBUG_1
  return cur->get_name();
}

extern "C" eid_t createSymNode(int varname) {
#ifdef __DEBUG_1
  cout << "BackendIface: Creating symbolic node for var " << varname << endl;
#endif
  if (BackendIface::getEnv()->hasExpByName(varname)) {
    return BackendIface::getEnv()->getExpByName(varname)->get_name();
  } else {
    ExpNode* new_node =
      BackendIface::getEnv()->new_node(NULL, NULL, ExpNode::SYM);
    BackendIface::getEnv()->setExpByName(varname, new_node);
    return new_node->get_name();
  }
}

extern "C" eid_t createBoolNode(int val) {
  return BackendIface::getBoolNode((bool)val);
}

extern "C" eid_t createCondNode(eid_t cid, eid_t tid, eid_t fid) {
  ExpNode* guard = BackendIface::getEnv()->get_node(cid);
  ExpNode* mother = BackendIface::getEnv()->get_node(tid);
  ExpNode* father = BackendIface::getEnv()->get_node(fid);

  TernaryNode* cur = (TernaryNode*)
    BackendIface::getEnv()->new_node(mother, father, ExpNode::TERNARY);
  cur->set_guard(guard);
  return cur->get_name();
}

/**
 * Binary operators.
 */
extern "C" eid_t createArithopNode(arithop_t op, eid_t id1, eid_t id2) {
  cout << "Arithmetic operator " << op << endl;
  ExpNode* mother = BackendIface::getEnv()->get_node(id1);
  ExpNode* father = BackendIface::getEnv()->get_node(id2);
  ExpNode* cur = NULL;
  switch (op) {
    case PLUS:
      cur = BackendIface::getEnv()->new_node(mother, father, ExpNode::PLUS);
      break;
    case MINUS:
      cur = BackendIface::getEnv()->new_node(mother, father, ExpNode::PLUS);
      ((PLUS_node*)cur)->set_minus(true);
      break;
    case TIMES:
      cur = BackendIface::getEnv()->new_node(mother, father, ExpNode::TIMES);
      break;
    case DIV:
      cur = BackendIface::getEnv()->new_node(mother, father, ExpNode::DIV);
      break;
    default:
      cout << "BackendIface::createArithopNode: Got arithmetic operator "
           << op << "." << endl;
      Assert(false, "Internal error.");
  }

  Assert(cur != NULL, "Should have initialized new node.");
  return cur->get_name();
}

extern "C" eid_t createBoolopNode(boolop_t op, eid_t id1, eid_t id2) {
  ExpNode* new_node = NULL;

  switch (op) {
    case AND:
      new_node = new AndNode();
      break;
    case OR:
      new_node = new OrNode();
      break;
    case NOT:
      new_node = new NOT_node();
      break;
    case LT:
      new_node = new LT_node();
      break;
    case LTE:
      new_node = new LE_node();
      break;
    case GT:
      new_node = new GT_node();
      break;
    case GTE:
      new_node = new GE_node();
      break;
    case EQ:
      new_node = new EQ_node();
      break;
    case NEQ:
      Assert(false ,"NEQ_node is not implemented.");
      break;
    default:
      cout << "BackendIface::createBoolopNode: Received boolean operator "
           << op << "." << endl;
      Assert(false, "Internal error.");
  }

  Assert(new_node != NULL, "Should have initialized new node.");
  
  ExpNode* mother = BackendIface::getEnv()->get_node(id1);
  ExpNode* father = BackendIface::getEnv()->get_node(id2);
  ExpNode* cur = BackendIface::getEnv()->new_node(mother, father, new_node);  
  return cur->get_name();
}

extern "C" void aliasBackendVal(vname varname, vname origname) {
#ifdef __TEST
  Assert( BackendIface::getEnv()->hasExpByName(origname)
        , "BackendIface::aliasBackendVal: original name must be bound.");
#endif
  ExpNode* orig = BackendIface::getEnv()->getExpByName(origname);
  BackendIface::getEnv()->setExpByName(varname, orig);
}

/**
 * Backend environment values.
 */
extern "C" void bindBackendVal(vname varname, eid_t eid) {
#ifdef __DEBUG_1
  cout << "BackendIface::bindBackendVal: " << varname << " := " << eid << endl;
#endif
  BackendIface::getEnv()->setEidByName(varname, eid);
  cout << "BOUND: " << BackendIface::getEnv()->getEidByName(varname) << endl;
}

extern "C" eid_t lookupBackendVal(vname varname) {
#ifdef __DEBUG_1
  cout << "BackendIface::lookupBackendVal: " << varname << ":= ";
  cout << BackendIface::getEnv()->getEidByName(varname) << endl;
#endif
  return BackendIface::getEnv()->getEidByName(varname);
}

/**
 * Relation manipulation.
 */
extern "C" eid_t readExpRelation(string& file) {
  Assert(false, "Unexpected reading relation.");
  ExpNode* new_node = new RelationNode(file);

  ExpNode* cur = BackendIface::getEnv()->new_node(NULL, NULL, new_node);
  return cur->get_name();
}

extern "C" void bindEnvRelationVal(vname rname, int pkey, int field, eid_t eid) {
#ifdef __DEBUG_1
  cout << "BackendIface::bindEnvRelationVal: " << rname << endl;
#endif
  ExpNode* relnode = BackendIface::getEnv()->getExpByName(rname);
  RelationNode* cur = NULL;
  if (relnode == NULL) {
#ifdef __DEBUG_1
    printf("BackendIface: Creating new relation %d.\n", rname);
#endif
#ifdef __TEST
    Assert(BackendIface::getEnv() != NULL, "env should not be null");
#endif
    cur =
      (RelationNode*)
      BackendIface::getEnv()->new_node(NULL, NULL, ExpNode::RELATION);
#ifdef __TEST
    Assert(cur != NULL, "Our new relation node is not null.");
    Assert( cur->getType() == ExpNode::RELATION
          , "Should have created relation node.");
#endif
    BackendIface::getEnv()->setExpByName(rname, cur);
  } else {
    Assert( relnode->getType() == ExpNode::RELATION
          , "Node we got from the environment should be a relation.");
    cur = (RelationNode*)relnode;
  }
#ifdef __DEBUG_1
  cout << "BackendIface::bindEnvRelationVal: Created relation with ID "
       << cur->get_name() << " for variable " << rname << "." << endl;
#endif
  Assert(cur->getType() == ExpNode::RELATION, "Should have relation.")
  
  /* Make a new record. */
  ExpNode* new_record =
    BackendIface::getEnv()->new_node(NULL, NULL, ExpNode::RECORD);
  Assert(new_record->getType() == ExpNode::RECORD, "Should have record.");
  cur->setRecord(pkey, (RecordNode*)new_record);
  cur->setValue(pkey, field, eid);
}

extern "C" eid_t lookupEnvRelationVal(vname rname, int pkey, int field) {
#ifdef __DEBUG_1
  printf( "BackendIface: Looking up relation value %d[%d][%d].\n"
        , rname, pkey, field );
  Assert( BackendIface::getEnv()->hasExpByName(rname)
        , "Must look up relation bound in environment.");
#endif
  RelationNode* cur =
    (RelationNode*)BackendIface::getEnv()->getExpByName(rname);
  Assert( cur != NULL
        , "Cannot look up value in null relation.");
  return cur->lookupRelationVal(pkey, field);
}

extern "C" eid_t lookupEnvFieldIndex(eid_t rname, int field) {
#ifdef __DEBUG_1
  cout << "BackendIface::lookupEnvFieldIndex: " << rname << "." << field << endl;
#endif
  ExpNode* cur = BackendIface::getEnv()->getExpByName(rname);
  Assert( cur->getType() == ExpNode::RECORD
        , "Should be looking up field of record." );
  return ((RecordNode*)cur)->lookupField(field);
}

/**
 * Given an expression ID corresponding to a relation, returns an iterator ID
 * corresponding to a new iterator for that relation.
 */
extern "C" int getRelIterator(eid_t relid) {
#ifdef __DEBUG_1
  printf("BackendIface::getRelIterator: Getting iterator for %d\n", relid);
#endif
  ExpNode* cur = BackendIface::getEnv()->get_node(relid);
  Assert(cur != NULL, "Cannot bind value in null relation.");
  Assert( cur->getType() == ExpNode::RELATION
        , "Must be relation to have iterator.");
  RelationNode* relnode = (RelationNode*)cur;
  return relnode->getIterator();
}

extern "C" eid_t getNextRecord(eid_t relid, int relit) {
#ifdef __DEBUG_1
  cout << "BackendIface::getNextRecord: EID " << relid << endl;
  Assert(relid > -1, "Relation ID should be valid.");
#endif
  ExpNode* cur = BackendIface::getEnv()->get_node(relid);
#ifdef __TEST
  Assert(cur != NULL, "Relation node should not be null");
  Assert(cur->get_name() == relid, "Relation node should be what we asked for.");
#endif

  Assert(cur->getType() == ExpNode::RELATION, "Should have relation type.");
  cout << "Assertions done" << endl;
  RelationNode* relnode = (RelationNode*)cur;
  int next_id = relnode->nextRecord(relit);
  cout << "Next record ID is " << next_id << endl;
  if (next_id > -1) {
    ExpNode* currec = BackendIface::getEnv()->get_node(next_id);
#ifdef __TEST
    Assert(currec != NULL, "Current record in iterator should not be null");
#endif
    if (currec->getOtype() == ExpNode::RECORDTY) {
      cout << "I get here!!" << endl;
      RecordNode* rec = (RecordNode*)currec;
      BackendIface::getEnv()->setExpByName(relit, rec);
      cout << "Returning..." << endl;
      return rec->get_name();
    }
  }
  return -1;
}

/**
 * Equality.
 */
extern "C" bool eqExpsInEnv(eid_t eid1, eid_t eid2) {
  ExpNode* e1 = BackendIface::getEnv()->get_node(eid1);
  ExpNode* e2 = BackendIface::getEnv()->get_node(eid2);
  return ExpNode::eqNodes(e1, e2);
}

extern "C" bool eqExpToValInEnv(int val, eid_t eid) {
  ExpNode* e = BackendIface::getEnv()->get_node(eid);
  if (e->getType() == ExpNode::CONST) {
    return val == ((ConstNode*)e)->getVal();
  } else {
    return false;
  }
}

/**
 *
 */
extern "C" void assertExpEnv(eid_t eid) {
#ifdef __DEBUG_1
  cout << "BackendIface::assertExpEnv: Asserting " << eid << endl;
#endif
  ExpNode* aexp = BackendIface::getEnv()->get_node(eid);
#ifdef __TEST
  Assert(aexp != NULL, "Node we're asserting should not be null.");
  cout << "BackendIface::assertExpEnv: Got node " << aexp->get_name() << endl;
#endif

  aexp->addVars(BackendIface::getEnv()->getVars());

  AssertNode* cur = new AssertNode(aexp);
  BackendIface::getEnv()->addAssertion(cur);

#ifdef __DEBUG_1
  cout << "BackendIface::assertExpEnv: Finished asserting " << eid << "." << endl;
#endif
}

extern "C" void writeEnvAssertions() {
  string file("constraints.txt");
  BackendIface::getEnv()->writeAssertionsToFile(file);
}

extern "C" void clearBackendEnv() {
  BackendIface::getEnv()->clear();
}
