/**
 *  This file defines different kinds of Boolean nodes.
 */
#ifndef EXPNODES_H_
#define EXPNODES_H_

#include <map>
#include <set>
#include <cstdlib>
#include <typeinfo>
#include "BackendTypes.h"
#include "FastSet.h"
#include "NodeVisitor.h"
#include "Dllist.h"


#ifdef CONST
#undef CONST
#endif

using namespace std;

class ExpDAG;

//#define SCHECKMEM

class ExpNode {

  private:
    /** The unique ID to be assigned to the next ExpNode created. */
    static int __s_next_global_id;
#ifdef SCHECKMEM
    static set<ExpNode*> allocated;
#endif

  protected:
    ExpNode();
    ExpNode(const ExpNode& bn);
    eid_t __name;

  public:
    typedef enum{ AND, OR, NOT
                , PLUS, TIMES, DIV, NEG
                , CONST, BOOLCONST, SYM, RELATION, RECORD
                , TERNARY
                , GT, GE, LT, LE, EQ
                , ASSERT} Type;
    typedef enum{ BOTTOM
                , BOOL, INT
                , RELTY, RECORDTY } OutType;

    int layer;
    /**
     * The globally unique ID of this node.
     *
     * (Warning: limits number of nodes to 4Gi).
     */
    int globalId;
    /** 
     * The unique ID of this node within a DAG, not guaranteed to be globally
     * unique.
     * (Implementation detail: the DAG contains a vector of nodes, and this 'id'
     * is this nodes position within that vector.
     */
    int id;	/* TODO: Ask Armando what ID is used for. */
    mutable int flag;
    int ion_pos;

    Type type;
    mutable OutType otype;
    ExpNode* mother;
    ExpNode* father;

    virtual ~ExpNode();

    void resetId();
    OutType joinOtype(OutType t1, OutType t2) const;

    void setMotherFather(ExpNode* mother, ExpNode* father) {
      this->mother = mother;
      this->father = father;
    }
    virtual int do_dfs(int idx);
    virtual int back_dfs(int idx);
    virtual void replace_parent(const ExpNode * oldpar, ExpNode* newpar);
    virtual void outDagEntry(ostream& out) const;
    virtual void redirectParentPointers(
        ExpDAG& oribdag, const vector<const ExpNode*>& bdag, bool setChildrn,
        ExpNode* childToInsert);
    virtual void redirectPointers(
        ExpDAG& oribdag, const vector<const ExpNode*>& bdag);
    virtual void switchInputs(
        ExpDAG& bdag, map<ExpNode*, ExpNode*>& replacements);
    virtual string get_tname() const;
    virtual string get_sym() const;
    virtual void printval();
    virtual string lprint();
    virtual string lid();
    eid_t get_name();
    void set_name(vname name);

	ExpNode* get_mother() { return mother; }
	ExpNode* get_father() { return father; }

    /**
     * Serialization.
     */
    virtual string get_namestr() const;
    virtual string get_yicesstr() const;

    void set_layer(bool isRecursive);
    virtual void accept(NodeVisitor& visitor) =0;
    virtual ExpNode* clone(bool copyChildren = true)=0;
    virtual void printSubDAG(ostream& out, set<const ExpNode* >& s)const;
    virtual OutType getOtype() const;
    virtual Type getType() const;
    void neighbor_replace(ExpNode* replacement);

    virtual void addVars(set<int>& varset);

    static bool eqNodes(ExpNode* n1, ExpNode* n2);
};


/**
 * Special FastSet hash function that hashes on the
 * ExpNode's ID (assumed unique) instead of its memory address.
 */
template<>
struct FastSetTraits<ExpNode>
{
  static inline unsigned
    hash (ExpNode *bn, int sz)
    {
      unsigned tmp = bn->globalId; // (unsigned) in>>2;  // (generic version)

      tmp = tmp * (tmp + 4297);
      tmp = tmp + (tmp >> 2) + (tmp >> 5) + (tmp >> 21) + (tmp >> 28);		

      unsigned m = 1 << sz;
      m = m-1;
      return (tmp & m)<<2;
    }
};

class AndNode: public ExpNode{ 	
  public: 
    AndNode(const AndNode& b);
    AndNode();
    virtual void accept(NodeVisitor& visitor);
    virtual ExpNode* clone(bool copyChildren);
    OutType getOtype() const;
    Type getType() const;
    void addVars(set<int>& varset);

    string get_yicesstr() const;
};

class OrNode: public ExpNode{	
  public: 
    OrNode();
    OrNode(const OrNode& bn);
    virtual void accept(NodeVisitor& visitor);
    virtual ExpNode* clone(bool copyChildren = true);
    OutType getOtype()const;
    Type getType() const;
    void addVars(set<int>& varset);

    string get_yicesstr() const;
};

/* TODO */
class NOT_node: public ExpNode{	
  public: 
    NOT_node(){ type = NOT; }  
    NOT_node(const NOT_node& bn): ExpNode(bn){ }  
    virtual void accept(NodeVisitor& visitor)  { visitor.visit( *this ); }
    virtual ExpNode* clone(bool copyChildren = true){return new NOT_node(*this);  };
    OutType getOtype() const { return BOOL; }
    Type getType() const { return NOT; }
    virtual string lid(){
      stringstream str;
      str<<"(!"<<mother->lid()<<")";
      return str.str();
    }
};

class PlusNode: public ExpNode{	
  private:
    bool __is_minus;
  public:
    PlusNode();
    PlusNode(bool minus);
    PlusNode(const PlusNode& bn);

    void set_minus(bool minus);

    virtual void accept(NodeVisitor& visitor);
    virtual ExpNode* clone(bool copyChildren);
    OutType getOtype()const;
    Type getType() const;

    string get_yicesstr() const;
    void addVars(set<int>& varset);
};
class TimesNode: public ExpNode {
public:
	TimesNode();
	TimesNode(const TimesNode& bn);
	virtual void accept(NodeVisitor& visitor);
	virtual ExpNode* clone(bool copyChildren);
	OutType getOtype() const;
	Type getType() const;

	string get_yicesstr() const;
	void addVars(set<int>& varset);
};

class DivNode: public ExpNode{	
  public: 
    DivNode();
    DivNode(const DivNode& bn);
    virtual void accept(NodeVisitor& visitor);
    virtual ExpNode* clone(bool copyChildren);
    OutType getOtype()const;
    Type getType() const;

    string get_yicesstr() const;
    void addVars(set<int>& varset);
};

class NEG_node: public ExpNode{	

  public: 
    NEG_node(){ type = NEG; }
    NEG_node(const NEG_node& bn): ExpNode(bn){ }
    virtual void accept(NodeVisitor& visitor)  { visitor.visit( *this ); }
    virtual ExpNode* clone(bool copyChildren = true){return new NEG_node(*this);  };
    OutType getOtype()const { return INT; }
    Type getType() const { return NEG; }
};


class ConstNode: public ExpNode{ 
  int val;
  public:
  ConstNode();
  ConstNode(int n);
  ConstNode(const ConstNode& bn);
  virtual void accept(NodeVisitor& visitor);
  virtual void setVal(int v);
  virtual int getVal() const;

  string get_namestr() const;
  string get_yicesstr() const;

  virtual string lid();
  virtual string lprint();
  virtual ExpNode* clone(bool copyChildren = true);
  OutType getOtype()const;
  Type getType() const;
};

class BoolNode: public ExpNode{
  bool __val;
  public:
  BoolNode();
  BoolNode(bool n);
  BoolNode(const BoolNode& bn);
  virtual void accept(NodeVisitor& visitor);
  virtual void setVal(bool v);
  virtual int getVal() const;

  string get_namestr() const;
  string get_yicesstr() const;

  virtual string lid();
  virtual string lprint();
  virtual ExpNode* clone(bool copyChildren = true);
  OutType getOtype()const;
  Type getType() const;
};

typedef std::vector<eid_t> record_t;
typedef std::vector<eid_t>::iterator record_it;
class RecordNode: public ExpNode{
 private:
  record_t* __record;
  int __num_fields;

 public:
  RecordNode();
  ~RecordNode();
  RecordNode(int numfields);
  RecordNode(const RecordNode& bn);
  virtual void accept(NodeVisitor& visitor);
  string get_namestr() const;
  virtual string lid();
  virtual string lprint();
  virtual ExpNode* clone(bool copyChildren = true);
  OutType getOtype() const;
  Type getType() const;

  eid_t lookupField(int field);
  void setValue(int field, eid_t eid);

  void resize(int newsize);
};

typedef map<int, RecordNode*> relation_t;
typedef map<int, RecordNode*>::iterator relation_it;
class RelationNode: public ExpNode{
 private:
  relation_t __rel;
  int __num_fields;
  map<int, relation_it> __iterators;
  int __next_itid;

 public:
  RelationNode();
  ~RelationNode();
  RelationNode(const RelationNode& bn);
  RelationNode(int numfields);
  RelationNode(int numfields, const char* file_name);

  virtual void accept(NodeVisitor& visitor);
  string get_namestr() const;
  virtual string lid();
  virtual string lprint();
  virtual ExpNode* clone(bool copyChildren = true);
  OutType getOtype() const;
  Type getType() const;

  /* Iterating. */
  map<int, RecordNode*>& get_relation();
  int getIterator();
  eid_t nextRecord(int it_id);

  /* Handling record values. */
  eid_t lookupRelationVal(int pkey, int field);
  void setRecord(int pkey, RecordNode* record); 
  void setValue(int pkey, int field, eid_t eid);

  int get_num_fields();
};

/* TODO: Move this to C file. */
class InterNode: public ExpNode{
  protected:
    InterNode(const InterNode& bn);
    InterNode();
    int nbits;
  public:
    virtual ~InterNode(){}
    int get_nbits() const;
    void set_nbits(int n);
    OutType getOtype() const;
    Type getType() const { Assert(false, "Inter node?"); }
    string get_namestr() const;
    virtual string lprint();
    virtual string lid();
};

class SymNode: public InterNode{
  public:
    SymNode();
    SymNode(const SymNode& bn);
    SymNode(const vname& nm);
    virtual void accept(NodeVisitor& visitor);
    virtual ExpNode* clone(bool copyChildren = true);
    Type getType() const;

    string get_yicesstr() const;
    void addVars(set<int>& vars);
};

class GtNode: public ExpNode{	
  public: 
    GtNode();
    GtNode(const GtNode& bn);
    virtual void accept(NodeVisitor& visitor);
    virtual ExpNode* clone(bool copyChildren);
    OutType getOtype() const;
    Type getType() const;
    string get_yicesstr() const;
    void addVars(set<int>& varset);
};
class GeNode: public ExpNode{	
  public: 
    GeNode();
    GeNode(const GeNode& bn);
    virtual void accept(NodeVisitor& visitor);
    virtual ExpNode* clone(bool copyChildren);
    OutType getOtype() const;
    Type getType() const;
    string get_yicesstr() const;
    void addVars(set<int>& varset);
};
class LtNode: public ExpNode{	
  public: 
    LtNode();
    LtNode(const LtNode& bn);
    virtual void accept(NodeVisitor& visitor);
    virtual ExpNode* clone(bool copyChildren);
    OutType getOtype() const;
    Type getType() const;
    string get_yicesstr() const;
    void addVars(set<int>& varset);
};
class LeNode: public ExpNode{	
  public: 
    LeNode();
    LeNode(const LeNode& bn);
    virtual void accept(NodeVisitor& visitor);
    virtual ExpNode* clone(bool copyChildren);
    OutType getOtype()const;
    Type getType() const;
    string get_yicesstr() const;
    void addVars(set<int>& varset);
};
class EqNode: public ExpNode{	
  public: 
    EqNode();
    EqNode(const EqNode& bn);
    virtual void accept(NodeVisitor& visitor);
    virtual ExpNode* clone(bool copyChildren);
		OutType getOtype() const;
    Type getType() const;
    string get_yicesstr() const;
    void addVars(set<int>& varset);
};

class NeqNode: public ExpNode{
  public:
    NeqNode();
    NeqNode(const NeqNode& bn);
    virtual void accept(NodeVisitor& visitor);
    virtual ExpNode* clone(bool copyChildren);
		OutType getOtype() const;
    Type getType() const;
    string get_yicesstr() const;
    void addVars(set<int>& varset);
};

/* TODO: */
class AssertNode: public ExpNode, virtual public DllistNode{
  bool isHardAssert;
  ExpNode* __assert_exp;

  string msg;
  public:
  AssertNode ();
  AssertNode(ExpNode* aexp);
  AssertNode(const AssertNode& bn);
  Type getType() const;
  string get_yicesstr() const;

  virtual void accept (NodeVisitor &visitor);
  virtual ExpNode* clone(bool copyChildren = true);
  virtual void makeHardAssert();
  virtual bool isHard() const;
  virtual void setMsg(const string& pmsg);
  virtual const string& getMsg()const;
  virtual ~AssertNode(){  }
  virtual string lprint();

  void addVars(set<int>& varset);
};

/* Predicated nodes. */
class TernaryNode: public ExpNode {
 public:
  TernaryNode ();
  TernaryNode(const TernaryNode& bn);
  TernaryNode(ExpNode* guard);
  void set_guard(ExpNode* g) { __guard = g; }
  /* TODO: Make destructor? */
  Type getType() const;
  string get_yicesstr() const;

  virtual void accept (NodeVisitor &visitor);
  virtual ExpNode* clone(bool copyChildren = true);
  virtual ~TernaryNode(){  } 
  virtual string lprint();
  void addVars(set<int>& varset);

 private:
  ExpNode* __guard;
};

/* TODO: Change function name. */
/*
     typedef enum{ AND, OR, NOT
                , PLUS, TIMES, DIV, NEG
                , CONST, BOOLCONST, SYM, RELATION, RECORD
                , TERNARY
                , GT, GE, LT, LE, EQ
                , ASSERT} Type;
*/
inline ExpNode* newBoolNode( ExpNode::Type type){
	switch(type){
		case ExpNode::AND: return new AndNode();
		case ExpNode::OR: return new OrNode();
		case ExpNode::NOT: return new NOT_node();
		case ExpNode::ASSERT: return new AssertNode ();
		case ExpNode::PLUS: return new PlusNode();
		case ExpNode::TIMES: return new TimesNode();
		case ExpNode::DIV: return new DivNode();
		case ExpNode::NEG: return new  NEG_node();
		case ExpNode::CONST: return new ConstNode();
    case ExpNode::SYM: return new SymNode();
    case ExpNode::RELATION: return new RelationNode();
    case ExpNode::RECORD: return new RecordNode();
    case ExpNode::TERNARY: return new TernaryNode();
    case ExpNode::GT: return new GtNode();
		case ExpNode::GE: return new GeNode();
		case ExpNode::LT: return new LtNode();
		case ExpNode::LE: return new LeNode();
		case ExpNode::EQ: return new EqNode();
    default:
      Assert(false, "Creating node of unknown type.");
	}
}

inline bool isDllnode(ExpNode* bn){
	return bn->type == ExpNode::ASSERT;
}

inline DllistNode* getDllnode(ExpNode* bn){
	return dynamic_cast<DllistNode*>(bn);
}
#endif
