(*
 * Module       : BackendInterface
 * Maintainer   : jeanyang@csail.mit.edu
 * Last updated : 2/1/10
 *
 * Notes on the OCaml/C interface:
 * - An object of type value is either:
 *    - an unboxed integer
 *    - a pointer to a block inside the heap
 *    - a pointer to an object outside to the heap (allocated with malloc, or
 *      a C variable
 * - Heap blocks a strict structure with a header containing the size of the
 *   block and a tag saying how the contents are structured.
 *
 * Some questions:
 *
 *)
module BackendInterface =
struct
  (* Some type synonyms *)
  type rit = int      (* Relation iterator. *)
  type arithop = int
  type boolop = int

  type bexp =
    | Int of int
    | Sym of int
    | Bool of bool
    | String of string
    | GuardedExp of bexp        (* Guard *)
                  * bexp * bexp (* True case, false case *)
    | Nothing

  external pop : unit -> unit = "pop"
  external dup : unit -> unit = "dup"
  external swap : unit -> unit = "swap"
  external pushUnit : unit -> unit = "pushUnit"

  (***************************)
  (* Manipulating relations. *)
  (***************************)
  (* Takes a file name and returns a record ID. *)
  external readRelation : int -> string -> unit = "readRelation"
  external mkRelation : int -> int -> unit = "mkRelation"

  external bindRelationVal : int -> int -> unit =
    "bindRelationVal"
  external lookupRelationVal : int -> int -> unit =
    "lookupRelationVal"
  external lookupFieldIndex : int -> int -> unit = "lookupFieldIndex"
  external lookupFieldIndexHidden : int -> int -> unit = "lookupFieldIndexHidden"

  (* TODO: Do we also want a separate createRelation function? *)
  external relIterator : unit -> unit = "relIterator"
  external relPkeyIterator : int -> unit = "relPkeyIterator"
  external nextRecord : unit -> bool = "nextRecord"

  external recordSelect : int -> int = "recordSelect"
  external listElem : unit -> unit = "listElem"

  (***************************)
  (* Evaluating expressions. *)
  (***************************)
  (* TODO: Figure out if this should be a list. *)
  
  (* Stores a basicArithop in the back-end and returns and expression ID.
   * We expect the two bexp arguments to either be constaints, symbolic, or
   * expression ID's. *)
  external basicArithop : arithop -> unit = "basicArithop"
  (* Analogous to basicArithop. *)
  external basicBoolop : boolop -> unit = "basicBoolop"

  external createAtomEid : bexp -> unit = "getAtomEid"
  external createCondEid : unit -> unit = "getCondEid"

  external bindVal : int -> unit = "bindVal"
  external lookupVal : int -> unit = "lookupVal"
  external lookupValHidden : int -> unit = "lookupValHidden"

  external concretizeSym : int -> int -> unit = "concretizeSym"

  (************************)
  (* Expression equality. *)
  (************************)
  external eqExps : unit -> bool = "eqExps"
  external eqExpToVal : int -> bool = "eqExpToVal"

  external isSym : unit -> bool = "isSym"
                                   (* Raises Failure *)
  external getConcreteBool : unit -> bool = "getConcreteBool"
  external getConcreteInt : unit -> int = "getConcreteInt"

  (***********************)
  (* Adding constraints. *)
  (***********************)
  external assertExp : unit -> unit = "assertExp"
  external writeAssertions : unit -> unit = "writeAssertions"
  external resolveWithContext : int -> unit = "resolveWithContext"

  (************)
  (* Printing *)
  (************)
  external printEnv : unit -> unit = "printEnv"
  external printExp : unit -> unit = "printExp"

  (**********************)
  (* Clear environment. *)
  (**********************)
  external initEnv : unit -> unit = "initEnv"
  external clearEnv : unit -> unit = "clearEnv"
  external printLive : unit -> unit = "printLive"
end
