(*
  Marriage inference example.
  Last updated: 11/11/09
 *)
open AstUtils
open Error
open Evalenv

module Ast = Ast.Ast
module Evaluator = Evaluator.Evaluator
module TestUtils = TestUtils.TestUtils

let census_fields =
  [ (0, Ast.IntType)  (* household *)
  ; (1, Ast.IntType)  (* line *)
  ; (2, Ast.IntType)  (* age *)
  ; (3, Ast.IntType)  (* married *)
  ; (4, Ast.IntType)  (* spouse *)
  ]
(* Fields *)
let household = 0
let line = 1
let age = 2
let married = 3
let spouse = 4

let infer_married_ty_name = TestUtils.gen_var ()
let infer_married_ty =
  { Ast.dataname = infer_married_ty_name
  ; Ast.fieldinfo = census_fields }

(*
  data infer_married_ty { ... }
  where (rel) {
    forall e in rel .
      forall e' in rel.
        (e.household == e'.household &&
          e.spouse == e'.line)==>
          (e'.spouse == e.line &&
            e.married == e.married)
  }
 *)

(* Define a data type. *)
(*
 *)
let rel = TestUtils.gen_var ()
(* Variables *)
let bound_constraint =
  let e = TestUtils.gen_var () in
  let e' = TestUtils.gen_var () in
  let spouse_match_expr =
    (*
    let household_match =
      Ast.Boolop
        ( Ast.Equiv
          , ref (Ast.VarExp (Ast.FieldIndexVar (e, 0)))
          , ref (Ast.VarExp (Ast.FieldIndexVar (e', 0)))) in *)
    let line_match =
      Ast.Boolop
        ( Ast.Equiv
          , ref (Ast.VarExp (Ast.FieldIndexVar (e, 4)))
          , ref (Ast.VarExp ((Ast.FieldIndexVar (e', 1)))) ) in
      ( Ast.Boolop
          ( Ast.And
            , ref (Ast.Boolop
                     ( Ast.Nequiv
                       , ref (Ast.VarExp (Ast.FieldIndexVar (e, 1)))
                       , ref (Ast.VarExp (Ast.FieldIndexVar (e', 1))) ))
            , ref line_match ) ) in
  let married_same_as_spouse =
    Ast.Boolop ( Ast.Equiv
                 , ref (Ast.VarExp (Ast.FieldIndexVar (e, 3)))
               , ref (Ast.VarExp (Ast.FieldIndexVar (e', 3))) ) in
  let spouse_symmetric =
    Ast.Boolop ( Ast.Equiv
               , ref (Ast.VarExp (Ast.FieldIndexVar (e', 4)))
               , ref (Ast.VarExp (Ast.FieldIndexVar (e, 1))) ) in
  let married_constraint =
    Ast.Boolop ( Ast.And
               , ref married_same_as_spouse
               , ref spouse_symmetric ) in
     [ Ast.ConstraintForMatch
          ( e, rel
          , e', rel
          , ref spouse_match_expr
          , ref married_constraint) ]
let tydecl = Ast.TypeDecl (infer_married_ty, rel, bound_constraint)

(* Declare a relation. *)
let indata_name = TestUtils.gen_var ()
let data file =
  Ast.FunAppExp
    ( -1
      , [ ref (Ast.PrimExp (Ast.String file)) ] )
let assign_in file =
  Ast.AssignDecl ( indata_name
                 , Ast.RelationType infer_married_ty_name
                 , Ast.RhsExp (data file) )

(*
 * Declare a function that sums over things.
 *
 * procedure sum_helper (entry : census_data, sum : int) =
 *  if entry.married
 *    then sum + 1
 *    else sum
 *
 * procedure sum_ages ( r : census_data relation ):
 *  let sum_age = fold e cond_sum_married 0 r in
 *    sum_age
 *)
let sum_helper = TestUtils.gen_var ()
let sum_helper_fun =
  let sum = TestUtils.gen_var () in
  let census_entry = TestUtils.gen_var () in
  let age_exp =
        ref (Ast.VarExp (Ast.FieldIndexVar (census_entry, 2))) in
  Ast.FunDecl
    ( (sum_helper, Ast.IntType)
    , { Ast.funargs = [ (census_entry, Ast.RecordType infer_married_ty_name)
                      ; (sum, Ast.IntType) ]
      ; Ast.funbody =
      ref
        (Ast.Arithop
        ( Ast.Plus
        , ref (Ast.VarExp (Ast.Var sum))
        , ref
               (Ast.CondExp
                  ( ref
                    (Ast.Boolop
                      ( Ast.Equiv
                      , ref
                          (Ast.VarExp
                             (Ast.FieldIndexVar (census_entry, married)))
                      , ref (Ast.PrimExp (Ast.Integer 1)) ))
                  , age_exp
                  , ref (Ast.PrimExp (Ast.Integer 0)))) ))
    } )
(* TODO: test let *)
let sum_ages = TestUtils.gen_var ()
let sum_ages_fun =
  let r = TestUtils.gen_var () in
  Ast.FunDecl
    ( (sum_ages, Ast.IntType)
    , { Ast.funargs = [ (r, Ast.RelationType infer_married_ty_name) ]
      ; Ast.funbody =
        ref (Ast.FoldExp ( sum_helper
                    , ref (Ast.PrimExp (Ast.Integer 0))
                    , r)) } )

(* Call the function we just made. *)
let sum_data_name = TestUtils.gen_var ()
let sum_data =
  Ast.AssignDecl
    ( sum_data_name, Ast.IntType
    , Ast.RhsExp
        (Ast.FunAppExp (sum_ages, [ ref (Ast.VarExp (Ast.Var indata_name))])) )

(* Program. *)
let program files =
  [ tydecl
  ; assign_in (List.nth files 0)
  ; sum_helper_fun
  ; sum_ages_fun
  ; sum_data
  ; TestUtils.mkConcreteResult TestUtils.result_name Ast.IntType sum_data_name ]
