(*
    File          : constraintUtils.ml
    Last updated  : 10.7.09
    Maintainer    : jeanyang@csail.mit.edu
    Modules       : ConstraintSimplify
                    ConstraintConverter
                    ConstraintPrinter
 *)
open AstUtils
open Error

module Ast = Ast.Ast;;

(**********************)
(* ConstraintSimplify *)
(**********************)
module ConstraintSimplify =
struct
  let is_constrained c =
    match !c with
    | Ast.ACTrue -> false
    | _ -> true

  let eq_lists lst1 lst2 compf =
    List.fold_left2
      (fun acc ->
         fun elt1 -> fun elt2 -> acc && (compf elt1 elt2)) true lst1 lst2

  let eq_ints i1 i2 : int =
    if i1 == i2 then 0 else if i1 < i2 then -1 else 1

  let cvarind_eq (n1, r1, v1) (n2, r2, v2) =
    (String.compare n1 n2 == 0) && (String.compare r1 r2 == 0) && (v1 == v2)

  let crecind_eq (p1, n1, r1, i1) (p2, n2, r2, i2) : int =
    let order_p = compare p1 p2 in
      if order_p == 0
      then
        let order_n = compare n1 n2 in
          if order_n == 0
          then let order_r = compare r1 r2 in
            if order_r == 0
            then eq_ints i1 i2
            else order_r
            else order_n
            else order_p

  let cboolop_eq (n1 : Ast.boolop) (n2 : Ast.boolop) : int =
    match n1, n2 with
      | Ast.Equiv, Ast.Equiv -> 0
      | Ast.Nequiv, Ast.Nequiv -> 0
      | Ast.Lt, Ast.Lt -> 0
      | Ast.Gt, Ast.Gt -> 0
      | Ast.And, Ast.And -> 0
      | Ast.Or, Ast.Or -> 0
      (* Arbitrary orderings. *)
      | Ast.Equiv, _ -> -1
      | Ast.Nequiv, Ast.Equiv -> 1
      | Ast.Nequiv, _ -> -1
      | Ast.Lt, Ast.Equiv -> 1
      | Ast.Lt, Ast.Nequiv -> 1
      | Ast.Lt, _ -> -1
      | Ast.Gt, Ast.And -> -1
      | Ast.Gt, Ast.Or -> -1
      | Ast.Gt, _ -> 1
      | Ast.And, Ast.Or -> -1
      | Ast.And, _ -> 1
      | Ast.Or, _ -> 1
  let cbinop_eq (b1 : Ast.binop) (b2 : Ast.binop) : int =
    match b1, b2 with
      | Ast.Plus, Ast.Plus -> 0
      | Ast.Minus, Ast.Minus -> 0
      | Ast.Plus, Ast.Minus -> -1
      | Ast.Minus, Ast.Plus -> 1
  (* Compare constraints to make it an ordered type. *)
  let rec cexp_eq (e1 : Ast.cexp ref) (e2 : Ast.cexp ref) : int =
    let binop_eq eq_binops e1_1 e1_2 e2_1 e2_2 =
      if eq_binops == 0
      then let order_e1 = cexp_eq e1_1 e2_1 in
        if order_e1 == 0
        then cexp_eq e1_2 e2_2
        else order_e1
      else eq_binops in
    let listop_eq eq_ops e1's e2's =
      if eq_ops == 0
      then
        if (eq_lists e1's e2's (fun c1 -> fun c2 -> cexp_eq c1 c2 == 0))
        then 0
        else 1  (* TODO *)
      else eq_ops in
      match !e1, !e2 with
        | Ast.ACVarExp ve1, Ast.ACVarExp ve2 -> eq_ints ve1 ve2
        | Ast.ACPrim i1, Ast.ACPrim i2 -> eq_ints i1 i2
        | Ast.ACArithop (binop1, e1's), Ast.ACArithop (binop2, e2's) ->
            listop_eq (cbinop_eq binop1 binop2) e1's e2's
        | Ast.ACBoolop (boolop1, e1's), Ast.ACBoolop (boolop2, e2's) ->
            listop_eq (cboolop_eq boolop1 boolop2) e1's e2's
        | Ast.ACGuard (c1, e1_1, e1_2), Ast.ACGuard (c2, e2_1, e2_2) ->
            binop_eq (cexp_eq c1 c2) e1_1 e1_2 e2_1 e2_2
        | Ast.ACTrue, Ast.ACTrue -> 0
        | Ast.ACFalse, Ast.ACFalse -> 0
        (* Ast.ACVarExp < Ast.ACPrim < Ast.ACArithop < Ast.ACBoolop <
           Ast.ACGuard < Ast.ACTrue < Ast.ACFalse *)
        | Ast.ACVarExp _, _ -> -1
        | Ast.ACPrim _, Ast.ACVarExp _ -> 1
        | Ast.ACPrim _, _ -> -1
        | Ast.ACArithop (_, _), Ast.ACVarExp _
        | Ast.ACArithop (_, _), Ast.ACPrim _ -> 1
        | Ast.ACArithop (_, _), _ -> -1
        | Ast.ACBoolop (_, _), Ast.ACVarExp _ | Ast.ACBoolop (_, _), Ast.ACPrim _
        | Ast.ACBoolop (_, _), Ast.ACArithop (_, _) -> 1
        | Ast.ACBoolop (_, _), _ -> -1
        | Ast.ACGuard (_, _, _), Ast.ACVarExp _
        | Ast.ACGuard (_, _, _), Ast.ACPrim _
        | Ast.ACGuard (_, _, _), Ast.ACArithop (_, _)
        | Ast.ACGuard (_, _, _), Ast.ACBoolop (_, _) -> 1
        | Ast.ACGuard (_, _, _), _ -> -1
        | Ast.ACTrue, Ast.ACFalse -> -1
        | Ast.ACTrue, _ -> 1
        | Ast.ACFalse, _ -> 1
        | _, _ -> -1 (* TODO *)

  let remove_duplicates eq_fn lst =
    let rec rmdup_helper accum cur_elt =
      (try
         let _ = List.find (eq_fn cur_elt) accum in
           accum
       with Not_found -> cur_elt::accum) in
      List.rev (List.fold_left rmdup_helper [] lst)

  let mkACAnd c1 c2 =
    if c1 == Ast.ref_false || c2 == Ast.ref_false
      then Ast.ref_false
      else
      if c1 == Ast.ref_true
      then c2
      else
        if c2 == Ast.ref_true
        then c1
        else ref (Ast.ACBoolop (Ast.And, [c1; c2]))

  (* Simplify constraints. *)
  let rec csimplify (ce : Ast.cexp ref) : Ast.cexp ref =
    let prim e =
      match !e with
        | Ast.ACPrim _ | Ast.ACTrue | Ast.ACFalse -> true
        | _ -> false in
    let get_int e =
      match !e with
        | Ast.ACPrim i -> i
        | _ ->
            raise
              (Unexpected
                 "Internal error: Something other than a primitive integer.") in
    let get_bool e =
      match !e with
        | Ast.ACTrue -> true
        | Ast.ACFalse -> false
        | _ ->
            raise
              (Unexpected
                 "Internal error: Something other than true/false.") in
    let eval_bool e = if e then Ast.ref_true else Ast.ref_false in
    match !ce with
      | Ast.ACId _ -> ce
      | Ast.ACVarExp _ -> ce
      | Ast.ACPrim _ -> ce 
      | Ast.ACArithop (op, ce's) ->
          let simplified = List.map csimplify ce's in
          if List.for_all prim simplified
          then
            let ints = List.map get_int simplified in
            (match op with
              | Ast.Plus -> ref (Ast.ACPrim (List.fold_left (+) 0 ints))
              | Ast.Minus -> ref (Ast.ACPrim (List.fold_left (-) 0 ints))
              | _ -> raise (Unexpected "op"))
          else ref (Ast.ACArithop (op, simplified))
      | Ast.ACBoolop (op, ce's) ->
          let simplified = List.map csimplify ce's in
          if List.for_all prim simplified
          then
            (match op with
              | Ast.And ->
                  let bools = List.map get_bool simplified in
                  eval_bool (List.fold_left (&&) true bools)
              | Ast.Or ->
                  let bools = List.map get_bool simplified in
                    eval_bool (List.fold_left (||) false bools)
              | Ast.Equiv ->
                  if List.length simplified == 2
                  then 
                    let ints = List.map get_int simplified in
                      if List.nth simplified 0 == List.nth simplified 1
                      then Ast.ref_true
                      else Ast.ref_false
                  else raise (InternalError "equiv without 2 args")
              | Ast.Nequiv ->
                  if List.length simplified == 2
                  then 
                    let ints = List.map get_int simplified in
                      if List.nth simplified 0 != List.nth simplified 1
                      then Ast.ref_true
                      else Ast.ref_false
                      else raise (InternalError "equiv without 2 args")
              | _ -> ref (Ast.ACBoolop (op, simplified))) (* TODO *)
          else ref (Ast.ACBoolop (op, simplified))
      | Ast.ACGuard (ce, ce1, ce2) ->
          let ce' = csimplify ce in
          let ce1' = csimplify ce1 in
          let ce2' = csimplify ce2 in
            if ce' == Ast.ref_true
            then ce1'
            else
              if ce' == Ast.ref_false
              then ce2'
              else
                if ce2' == Ast.ref_true
                then
                  if ce1' == Ast.ref_true || ce1' == Ast.ref_false
                  then ce1'
                  else ref (Ast.ACGuard (ce', ce1', ce2'))
                else ref (Ast.ACGuard (ce', ce1', ce2'))
      | Ast.ACTrue -> Ast.ref_true
      | Ast.ACFalse -> Ast.ref_false
end

(*********************)
(* ConstraintPrinter *)
(*********************)
module ConstraintPrinter =
  struct
    let pickle_cboolop nc = match nc with
        Ast.Equiv -> "=="
      | Ast.Nequiv -> "!="
      | Ast.Lt -> "<"
      | Ast.Gt -> ">"
      | Ast.And -> "and"
      | Ast.Or -> "or"
    let pickle_binop binop = match binop with
      | Ast.Plus -> "+"
      | Ast.Minus -> "-"
    let pickle_vrecordindex (n, f, v) =
      String.concat "" ["("; n; ", "; f; ", "; v; ")"]
    let pickle_recordindex (n, f, i) =
      String.concat "" ["("; n; ", "; string_of_int f; ", "; string_of_int i; ")"]

    (* TODO: Just make this return a string. *)
    let rec pickle_exp exp : string list =
      match !exp with
        | Ast.ACId id -> [string_of_int id]
        | Ast.ACVarExp ve -> [string_of_int ve]
        | Ast.ACPrim i -> [string_of_int i]
        | Ast.ACArithop (binop, e's) ->
            ("("::
              (List.append
                 ((pickle_binop binop)::
                  (List.concat (List.map pickle_exp e's))) [")"]))
        | Ast.ACBoolop (boolop, e's) ->
            "("::
              (List.append
                 ((pickle_cboolop boolop)::
                  (List.concat (List.map pickle_exp e's))) [")"])
        | Ast.ACGuard (p, e_t, e_f) ->
            List.append (pickle_exp p)
              ("?"::(List.append (pickle_exp e_t) (":"::(pickle_exp e_f))))
        | Ast.ACTrue -> ["True"]
        | Ast.ACFalse -> ["False"]

    let print_cexps (cdecls : Ast.cexp ref list) =
      let cdecls' =
          List.map ConstraintSimplify.csimplify
            (List.filter ConstraintSimplify.is_constrained cdecls) in
      let print_cdecl cdecl =
        let cdecl_str = String.concat " " (pickle_exp cdecl) in
         Printf.printf "%s\n\n" cdecl_str in
      List.map print_cdecl cdecls'
end

(***********************)
(* ConstraintConverter *)
(***********************)
module ConstraintConverter =
struct
  (* Conversion of arithmetic operators. *)
  let convert_arithop arithop = match arithop with
    | Ast.Plus -> Ast.Plus
    | Ast.Minus -> Ast.Minus
  (* Conversion of boolean operators. *)
  let convert_boolop boolop = match boolop with
    | Ast.Equiv -> Ast.Equiv
    | Ast.Nequiv -> Ast.Nequiv
    | Ast.Lt -> Ast.Lt
    | Ast.Gt -> Ast.Gt
    | Ast.And -> Ast.And
    | Ast.Or -> Ast.Or

  let rec ast2constraint (aexp : Ast.exp ref) : Ast.cexp ref =
      (* Body. *)
      match !aexp with
        | Ast.ExpId id -> ref (Ast.ACId id)
        | Ast.PrimExp pe ->
            (match pe with
               | Ast.Integer i -> ref (Ast.ACPrim i)
               | Ast.Bool b -> raise (Unexpected "ast2constraint: boolean")
               | Ast.String _ ->
                   raise (Unexpected "ConstraintConverter: string"))
        | Ast.FunAppExp _ ->
            raise (Unexpected "ConstraintConverter: function application")
        | Ast.VarExp ve ->
            (match ve with
               | Ast.Var v -> ref (Ast.ACVarExp v)
               | _ -> raise (Unexpected "ConstraintConverter: Non-variable."))
        | Ast.Arithop (opr, e's) ->
            ref (Ast.ACArithop
                   ( opr
                   , List.map (fun e -> ast2constraint e) e's))
        | Ast.Boolop (opr, e's) ->
            ref (Ast.ACBoolop
                   ( opr
                   , List.map (fun e -> ast2constraint e) e's))
        | Ast.LetExp (_, _, _, _) ->
            raise (Unexpected "ConstraintConverter: unevaluated let")
        | Ast.FoldExp (_, _, _) ->
            raise (Unexpected "ConstraintConverter: unevaluated fold")
        | Ast.CondExp (ec, et, ef) ->
            ref
              (Ast.ACGuard
                 ( ast2constraint ec
                 , ast2constraint et
                 , ast2constraint ef))
        | Ast.Symbolic v -> ref (Ast.ACVarExp v)

  (* Converts the constraint environment. *)
  let convert_constraints (cexps : Ast.cexp ref list)
                        : string list =
    let ybool2string b = 
      match b with
        | Ast.Equiv -> "="
        | Ast.Nequiv -> "/="
        | Ast.Lt -> "<"
        | Ast.Gt -> ">"
        | Ast.Or -> "or"
        | Ast.And -> "and" in
    let yarith2string a =
      match a with
        | Ast.Plus -> "+"
        | Ast.Minus -> "-" in
    (* expression *)
    let rec yc2string cexp =
      match !cexp with
        | Ast.ACPrim i -> string_of_int i
        | Ast.ACVarExp ve -> "v" ^ string_of_int ve
        | Ast.ACBoolop (ybool, e's) ->
            String.concat " "
              (List.append
                ["("; ybool2string ybool]
                (List.append (List.map yc2string e's) [")"]))
        | Ast.ACArithop (yarith, e's) ->
            String.concat " "
              (List.append
                ["("; yarith2string yarith]
                (List.append (List.map yc2string e's) [")"]))
        | Ast.ACGuard (pred, texp, fexp) ->
            String.concat " "
              ["("; "if"; yc2string pred; yc2string texp; yc2string fexp; ")"]
        | Ast.ACTrue -> "true"
        | Ast.ACFalse -> "false" in
    let assertify s = String.concat "" ["(assert "; s; ")"] in
    List.map assertify (List.map yc2string cexps)

let get_name (v : Ast.var) : string = match v with
  | Ast.Var vname -> string_of_int vname
  | Ast.RecordIndexVar (p, r, f) -> Utils.to_record_name (p, r, f)
  | Ast.FieldIndexVar (r, f) -> string_of_int r ^ "." ^ string_of_int f
end
