(* 
    Module      : driver.ml
    Maintainer  : jeanyang@csail.mit.edu
    Updated     : 10.26.10

    This contains the LogLog main loop.
*)
open AstUtils
open Error
open Vars

module Ast = Ast.Ast;;
module BackendInterface = BackendInterface.BackendInterface;;
module Evalenv = Evalenv.Evalenv;;
module Evaluator = Evaluator.Evaluator;;

(* Runs the interpreter.
   Unless -debug is set to some level (0-n), the only thing we print is the
   cumulative sum total constraint solving time. *)
let run_interpreter () : unit =
  let i_start_time = Sys.time () in

  (* Perform initialization. *)
  let _ = BackendInterface.initEnv () in
  let _ = Init.init_eval_env () in

  (* 1. Open file. *)
  let cin = open_in !file_name in

  (* 2. Lex. *)
  let lexbuf = Lexing.from_channel cin in

  (* 3. Parse. *)
  let topDecls = Parser.input Lexer.token lexbuf in
  let _ =
    if !print_ast
      then AstPrint.print_topdecls topDecls
      else () in

  let env = ref Evalenv.empty_evalenv in

  (* OPTIONAL: Type-check. *)
  let _ =
    if !type_check
    then Typecheck.check env topDecls
    else () in
    
  (* 4. Evaluate. *)
    Evaluator.evaluate env topDecls;
    if debug_level 3
      then BackendInterface.printEnv ()
      else ();
    
    (* Print the time. *)
    if !print_times
      then
        let i_end_time = Sys.time () in
          print_endline
            ("Running time: " ^ string_of_float (i_end_time -. i_start_time));
          Time.print_time ()
      else ()

(* Specifications for command-line arguments. *)
let cmd_specs =
  [ ("-file",       Arg.Set_string file_name, "Jeeves file name")
  (* DEBUG: Print the debug level. *)
  ; ("-debug",      Arg.Int set_debug,        "debug?") 
  ; ("-printast",   Arg.Set print_ast,        "Print the AST after parsing")
  ; ("-times",      Arg.Set print_times,      "Print running times")
  ; ("-typecheck",  Arg.Set type_check,       "Typecheck") ]
let anon_fun x = print_endline x
let usage_message = "Please use the following options."

(* Main stuff. *)
let main () =
  Arg.parse cmd_specs anon_fun usage_message;
  if (String.length (!file_name) > 0)
  then run_interpreter ()
  else print_endline "Please specify a valid file."

let _ = Printexc.print main ()

