(*
  File          : evalExp.ml
  Last updated  : 3.15.10
  Maintainer    : jeanyang@csail.mit.edu

  This file contains the function for expression evaluation.
*)
open AstBackend
open AstUtils
open BackendInterface
open Error
open Evalenv

module Ast = Ast.Ast

(*
 * Function evaluation.
 *)
let rec evalFunction
      (env : Evalenv.evalenv ref) (result_type : Ast.typerep)
      (f    : string)
      : unit =
  (* Look up the function in the environment. *)
  match Evalenv.lookupFunDecl env (Varenv.lookup_var f) with
      (* Function comes with a definition and a closure. *)
    | Some ( { Ast.funargs = fargs
             ; Ast.funbody = funbody }
           , fun_closure_env ) ->
        (* Go through and bind each of the arguments. *)
        (* TODO: type check! *)
        let bind_arg
              ((name, ty) : Ast.varname * Ast.typerep)
              : unit =
          (* We evaluate the function in the closure environment. *)
          (* evalExp fun_closure_env arg ty; *)
          (* TODO: Make sure the functions are being evaluated in the right
           * environments. *)
          Evalenv.bindValueEnv env (Ast.Var name);
          Evalenv.bindTypeEnv env name ty in
          List.iter bind_arg fargs;
          evalExp env funbody result_type
    | None -> raise (Unbound (f, "EvalStatement.evalFunction"))

(*
 * Expression evaluation.
 *
 * Evaluation of an expression either yields a concrete expression or a
 * symbolic expression.
 *
 * Invariants:
 * - All evaluated expressions are Ast.ExpId values *unless* they are primitive
 *   expressions or symbolic expressions. (TODO: Fix the part about the symbolic
 *   guys?)
 *)
and evalExp (env : Evalenv.evalenv ref) (e : Ast.exp ref) (t : Ast.typerep)
                : unit =
  debug_print 3 ("Evaluating " ^ AstPrint.pickle_exp e);
  (* Type-checking helper function. *)
  let checkType t0 t1 =
    if AstEq.eq_types t0 t1
    then ()
    else raise
           (TypeError ("Expected type " ^ AstPrint.pickle_type t0 ^ "; " ^
                       "found type " ^ AstPrint.pickle_type t1 ^ " in exp " ^
                       AstPrint.pickle_exp e ^ ".")) in
      (* evalExp body *)
      match !e with
      (* PRIMITIVES.
       * Do type-checking. *)
      (* TODO: Make sure we're applying type-level constraints whenever we're
       * dealing with relations. *)
      | Ast.PrimExp pe -> AstBackend.mkAstExpNode e
      (* VARIABLES.
       * For variable, look it up and evaluate it. *)
      | Ast.VarExp varexp ->
          (match varexp with
             | Ast.RecordIndexVar (pkey, _, _) ->
                 evalExp env pkey Ast.IntType
             | _ -> ());
          Evalenv.lookupValueEnv env varexp
      (* Functions. *)
      | Ast.Lambda ((varname, varty), lambdaexp) ->
          let fdecl = { Ast.funargs = [(varname, varty)]
                      ; Ast.funbody = lambdaexp } in
          let env_copy = !env in
            Evalenv.addFunDecl env varname fdecl (ref env_copy)
      (* FUNCTION APPLICATION. *)
      | Ast.FunAppExp (f, args) ->
        (match f with
          | "input" -> (* input *)
          if (List.length args == 2)
            then
              let file_arg =
                    match !(List.nth args 0) with
                    | Ast.PrimExp (Ast.String s) -> s
                    | _ -> raise (Unexpected "non-string") in
              let numfields =
                match !(List.nth args 1) with
                  | Ast.PrimExp (Ast.Integer i) -> i
                  | _ -> raise (Unexpected "non-integer") in
              let file_name = "data/" ^ file_arg ^ ".txt" in
              let start_time = Sys.time () in
              BackendInterface.readRelation numfields file_name;
              let total_time = (Sys.time ()) -. start_time in
               print_endline ("Read from " ^ file_name^ " (interval time): " ^
                              string_of_float total_time);
            else raise (Unexpected "input applied to wrong number of args")8
          | _ ->
              List.iter (fun e -> evalExp env e Ast.UnknownType) (List.rev args);
              (try EvalFun.evalBuiltinFun f
               with FunctionUndefined _ -> evalFunction env t f))
      (* ARITHMETIC OPERATIONS. *)
      | Ast.Arithop (op, e1, e2) ->
          evalExp env e1 Ast.IntType;
          evalExp env e2 Ast.IntType;
          BackendInterface.basicArithop (AstBackend.get_arithop op)
      (* BOOLEAN OPERATIONS. *)
      | Ast.Boolop (op, e1, e2) ->
          evalExp env e1 Ast.UnknownType;
          evalExp env e2 Ast.UnknownType;
          BackendInterface.basicBoolop (AstBackend.get_boolop op)
      (* LET-BOUND EXPRESSION. *)
      | Ast.LetExp (var, varty, assignexp, expbody) ->
          evalExp env assignexp varty;
          Evalenv.bindValueEnv env (Ast.Var var);
          Evalenv.bindTypeEnv env var varty;
          evalExp env expbody t
      (* FOLD. *)
      | Ast.FoldExp (foldfun, init, foldvar) ->
          debug_print 2 "Evaluating fold expression.";
          debug_print 3 "Getting relation iterator.";
          let _ = Evalenv.lookupValueEnv env (Ast.Var foldvar) in
          let iter_id = BackendInterface.relIterator () in
          debug_print 3 "Getting first record.";
          let has_cur_elt = ref (BackendInterface.nextRecord ()) in
          evalExp env init Ast.UnknownType;
          BackendInterface.swap ();
          while !has_cur_elt do
            (* Need this swap to get the record on top and the accumulator
             * below.  Coming into the function, we should have
             *    | next record      |
             *    | accumulator      |  *)
            evalFunction env t foldfun;
            debug_print 3 "EvalExp.evalExp: Getting next record.";
            (* Need this swap to get the relation on top and the accumulator
             * below it. *)
            has_cur_elt := BackendInterface.nextRecord ()
            (* After this, we have:
             *    | next record      |
             *    | accumulator      |  *)
          done;
      (* CONDITIONAL EXPRESSION. *)
      (* We try to evaluate the condition.  If the condition is concrete, then
       * we only evaluate one of the branches.  Otherwise, we evaluate the
       * both branches. *)
        | Ast.CondExp (cond, tbr, fbr) ->
            evalExp env cond Ast.BoolType;
            let (csym, ccond) =
              (try (false, BackendInterface.getConcreteBool ())
               with Failure _ -> (true, true)) in
              (* If the condition is symbolic, then we create a new conditional
               * expression. *)
              if csym
              then
                (evalExp env tbr t;
                 evalExp env fbr t;
                 BackendInterface.createCondEid ())
              (* Otherwise we resolve the condition. *)
              else 
                (BackendInterface.pop ();
                if ccond then evalExp env tbr t else evalExp env fbr t)
        (* TODO: Make it so symbolic variables in the front-end don't
         * store anything. *)
        | Ast.Symbolic (vname, cexp)  ->
            (* TODO: Evaluate the constraint. *)
            AstBackend.mkAstExpNode e
