(*
  File          : evalSym.ml
  Last updated  : 4.13.10
  Maintainer    : jeanyang@csail.mit.edu

  This file contains the function for expression evaluation.
*)
open BackendInterface
open Error
open Evalenv

module Ast = Ast.Ast

let rec evalSym (env : Evalenv.evalenv ref) (e : Ast.exp ref)
      : bool =
  match !e with
    | Ast.PrimExp pe -> false
    | Ast.VarExp varexp ->
        (Evalenv.lookupValueEnv env varexp;
         BackendInterface.isSym ())
    | Ast.CondExp (cond, tbr, fbr) ->
        (evalSym env cond) || (evalSym env tbr) || (evalSym env fbr)
    | Ast.Symbolic (vname, _) -> true
    | Ast.Arithop (_op, e1, e2) -> (evalSym env e1) || (evalSym env e2)
    | Ast.Boolop (_op, e1, e2) -> (evalSym env e1) || (evalSym env e2)
    | Ast.FunAppExp (_, args) -> List.exists (fun e -> evalSym env e) args
    | Ast.Lambda ((varname, varty), lambdaexp) ->
        raise (Unexpected "lambda expression")
    | Ast.LetExp (var, varty, assignexp, expbody) ->
        raise (Unexpected "let exp")
    | Ast.FoldExp (foldfun, init, foldvar) ->
        raise (Unexpected "fold exp")

type sval =
  | SInt of int
  | SBool of bool

let rec sEval (env : Evalenv.evalenv ref) (e : Ast.exp ref) : sval =
  let getInt e =
    match sEval env e with
      | SInt i -> i
      | SBool _b -> raise (Unexpected "bool arg") in
  let getBool e =
    match sEval env e with
      | SInt _i -> raise (Unexpected "int arg")
      | SBool b -> b in
  match !e with
    | Ast.PrimExp pe ->
        (match pe with
           | Ast.Integer i -> SInt i
           | Ast.Bool b -> SBool b
           | _ -> raise (Unexpected "EvalSym.sEval: Other primitive exp."))
    | Ast.VarExp varexp ->
        (try (Evalenv.lookupValueEnv env varexp;
              SInt (BackendInterface.getConcreteInt ()))
         with Failure _ -> raise SymbolicExp)
    | Ast.CondExp (cond, tbr, fbr) -> raise Undefined
        (match sEval env cond with
           | SBool b -> (if b then evalSym env tbr else evalSym env fbr)
           | SInt _i -> raise (Unexpected "EvalSym.sEval: Integer condition"))
    | Ast.Symbolic (vname, _) -> raise SymbolicExp
    | Ast.Arithop (op, e1, e2) ->
        let e1' = getInt e1 in
        let e2' = getInt e2 in
        (match op with
           | Ast.Plus -> SInt (e1' + e2')
           | Ast.Minus -> SInt (e1' - e2')
           | Ast.Times -> SInt (e1' * e2')
           | Ast.Divide -> SInt (e1' / e2'))
    | Ast.Boolop (op, e1, e2) ->
        let boolargs = op = Ast.Or || op = Ast.And in
          if boolargs
          then
            (match op with
                | Ast.Or ->
                    if getBool e1
                    then SBool true
                    else SBool (getBool e2)
                | Ast.And -> 
                    if getBool e1
                    then SBool (getBool e2)
                    else SBool false
                | _ -> raise (Unexpected "EvalSym.sEval: Internal error"))
          else
            (let e1' = getInt e1 in
             let e2' = getInt e2 in
               match op with
                 | Ast.Equiv -> SBool (e1' == e2')
                 | Ast.Nequiv -> SBool (e1' != e2')
                 | Ast.Lt -> SBool (e1' < e2')
                 | Ast.Gt -> SBool (e1' > e2')
                 | _ -> raise (Unexpected "EvalSym.sEval: Internal error"))
    | Ast.FunAppExp (_, args) -> raise (Unexpected "EvalSym.sEval: funapp")
    | Ast.Lambda ((varname, varty), lambdaexp) ->
        raise (Unexpected "EvalSym.sEval: lambda expression")
    | Ast.LetExp (var, varty, assignexp, expbody) ->
        raise (Unexpected "EvalSym.sEval: let exp")
    | Ast.FoldExp (foldfun, init, foldvar) ->
        raise (Unexpected "EvalSym.sEval: fold exp")

(* Returns (cond_not_false, cond_symbolic) *)
let shallowEval (env : Evalenv.evalenv ref) (e : Ast.exp ref) : bool * bool =
  try (match sEval env e with
         | SBool b ->
             debug_print 2 ("Concrete condition: " ^ string_of_bool b);
             (b, false)
         | SInt _ -> raise (Unexpected "EvalSym.shallowEval: Integer condition"))
  with SymbolicExp -> (true, true)
