(**
  Module      : Evalenv
  Maintainer  : jeanyang@csail.mit.edu
  Updated     : 1.12.10
*)
open AstBackend
open AstUtils
open BackendInterface
open Error

module Ast = Ast.Ast

module VarEnv =
  Hashtbl.Make
    (struct type t = Ast.varname
      let equal = (==)
      let hash v = v
    end)
module TypeEnv =
  Hashtbl.Make
    (struct type t = Ast.typename
      let equal v1 v2 = v1 == v2
      let hash v = v
    end)

(** Module for evaluation environment. *)
module Evalenv =
struct
  let max_env = 50

  (** Evaluation environment. *)
  type tyenv = Ast.typerep VarEnv.t
  and evalenv = { typeconstraints    : ((Ast.varname * Ast.cdecl list) VarEnv.t) ref;
                   (** Type-level constraints for corner-case values.
                       relation name |-> relation -> env -> env *)
                   tyenv              : (Ast.typerep VarEnv.t) ref;
                   (** variable name |-> type *)
                   tydefenv           : (Ast.userty VarEnv.t) ref;
                   (** type name |-> type representation *)
                   functions          : ((Ast.fundecl * evalenv ref)
                                           VarEnv.t) ref }

  let default_tyenv = VarEnv.create max_env
  let default_valenv = VarEnv.create max_env
  let empty_evalenv = { typeconstraints = ref (VarEnv.create max_env)
                      ; tyenv = ref (VarEnv.create max_env)
                      ; tydefenv = ref (VarEnv.create max_env)
                      ; functions = ref (VarEnv.create max_env) }

  (*************************)
  (* Function declarations *)
  (*************************)
  let addFunDecl env fname fdecl envref =
    VarEnv.replace !(!env.functions) fname (fdecl, envref)
  let lookupFunDecl env (fdeclname : Ast.varname)
        : (Ast.fundecl * evalenv ref) option =
    try Some (VarEnv.find !(!env.functions) fdeclname)
    with Not_found -> None

  (* Manipulating the global constraint environment. *)
  let addConstraintEnv (ce : Ast.exp ref) : unit =
    debug_print 3 "Evalenv.addConstraintEnv: Adding constraint";
    AstBackend.mkAstExpNode ce;
    BackendInterface.assertExp ()
  let addConstraintsEnv (constraints  : Ast.exp ref list) : unit =
    List.iter addConstraintEnv constraints

  (*
   * Binding a new value involves interacting with the back-end.
   *)
  let bindValueEnv (env : evalenv ref) (name : Ast.var) : unit =
    match name with
      | Ast.Var v -> BackendInterface.bindVal v
      | Ast.RecordIndexVar(_p, r, f) ->
          raise (Unexpected "Record index bind")
           (*  BackendInterface.bindRelationVal r f *)
      | _ -> raise (Unexpected "binding field index")
  
  (* Expressions in the variable environment. *)
  (* Value lookup--treat record field lookup in a special way. *)
  let lookupValueEnv (env : evalenv ref) (name : Ast.var) : unit =
    try match name with
      | Ast.Var v -> BackendInterface.lookupVal v
      | Ast.FieldIndexVar (r, f) -> BackendInterface.lookupFieldIndex r f
      | Ast.RecordIndexVar (_p, r, f) -> BackendInterface.lookupRelationVal r f
      | Ast.HiddenVarAccess v ->
         (match v with
             | Ast.Var v -> BackendInterface.lookupValHidden v
             | Ast.FieldIndexVar (r, f) ->
                 BackendInterface.lookupFieldIndexHidden r f
             | _ -> raise (Unexpected "Hidden access of record index var"))
    with Not_found ->
      raise (Unbound (AstPrint.pickle_var name, "Ast.lookupValueEnv"))

  (* Adds a type declaration to the environment. *)
  let addTypeDecl
        (env : evalenv ref) (tdecl : Ast.userty)
        (constraints : Ast.varname * Ast.cdecl list)
        : unit =
    let ty_name = tdecl.Ast.dataname in
    debug_print 3 ("Evalenv.addTypeDecl: " ^ string_of_int ty_name);
    VarEnv.replace !(!env.tydefenv) ty_name tdecl;
    VarEnv.replace !(!env.typeconstraints) ty_name constraints

  (* Get the type representation based on the data name. *)
  let lookupTypeDeclByData
        (env : evalenv ref) (data_name : Ast.varname)
        : Ast.userty option =
    try match VarEnv.find !(!env.tyenv) data_name with
      | Ast.RelationType type_name ->
          Some (VarEnv.find !(!env.tydefenv) type_name)
      | _ -> None
    with Not_found ->
      debug_print 1
        ("Evalenv.addTypeDecl: Did not find either the type or the type
          definition for " ^ string_of_int data_name ^ " in the environment.");
      None
  let lookupNumFieldsByData env (data_name : Ast.varname) : int =
    match lookupTypeDeclByData env data_name with
      | Some uty -> List.length uty.Ast.fieldinfo
      | None -> 0


  (*********************)
  (* Type environment. *)
  (*********************)
  let bindTypeEnv
        (env : evalenv ref)
        (name : Ast.varname)
        (ty : Ast.typerep)
        : unit =
    debug_print 2
      ("Evalenv.bindTypeEnv: Binding " ^ string_of_int name ^ " in type env.");
    VarEnv.replace !(!env.tyenv) name ty
  (* Looks up the type of a variable. *)
  let lookupType (env : evalenv ref)
                 (data_name : Ast.var)
                : Ast.typerep option =
    match data_name with
    | Ast.Var v ->
        (try Some (VarEnv.find !(!env.tyenv) v)
         with Not_found -> None)
    | Ast.RecordIndexVar (p, r, f) ->
        (try let reltype = VarEnv.find !(!env.tyenv) r in
              match reltype with
              | Ast.RelationType rtype ->
                  Some Ast.IntType (*  TODO *)
              | _ -> raise (Unexpected "Relation doesn't have relation type.")
         with Not_found -> None)
    | Ast.FieldIndexVar (r, f) ->
         (try let reltype = VarEnv.find !(!env.tyenv) r in
              match reltype with
              | Ast.RecordType rtype ->
                  Some Ast.IntType (*  TODO *)
              | _ -> raise (Unexpected "Relation doesn't have relation type.")
         with Not_found -> None)
end
