(**
    File         : evaluator.ml
    Last updated : 12.14.09
      
    Evaluator top loop.
*)
open AstUtils
open BackendInterface
open Error
open Vars

module Ast = Ast.Ast
module Evalenv = Evalenv.Evalenv;;
module Builtins = Builtins.Builtins;;
module Oracle = Oracle.Oracle;;

module Evaluator =
struct
    (* 
     * Evaluation function.
     *
     * We will go along the dataorder, creating the data transformations and
     * checking the constraints.  After each stage, we go through and refine
     * the symbolic values based on constraints?
     *)
    let evaluate
          (env : Evalenv.evalenv ref)
          (topdecls : Ast.topdecl list)
          : unit =
      (* Process top-level RHS of a data declaration. *)
      let processTopExp
            (data_name : Ast.varname) data_exp data_ty
            : unit =
        (Evalenv.bindTypeEnv env data_name data_ty;
         EvalExp.evalExp env (ref data_exp) data_ty) in
      (* Process top-level declarations.
       Assume data declarations occur before their associated constraint
       declarations. *)
      let processTopDecl () decl =
        match decl with
          (* Add type declaration information to type declaration environment. *)
          | Ast.TypeDecl (trep, v, constraints) ->
              debug_print 3
                ("Evaluator.processTopDecl: adding type decl for " ^
                 string_of_int trep.Ast.dataname);
              Evalenv.addTypeDecl env trep (v, constraints)
          (* Add function declaration information to fun. decl. environment. *)
          | Ast.FunDecl ((fname, _fty), fdecl) ->
              debug_print 3
                ("Evaluator.processTopDecl: adding function decl");
              Evalenv.addFunDecl env fname fdecl env
          (* Process data assignment appropriately. *)
          | Ast.AssignDecl (name, ty, rhs) ->
              debug_print 3
                ("Evaluator.processTopDecl: adding assignment");
              Evalenv.bindTypeEnv env name ty;
              match rhs with
                | Ast.RhsExp exp ->
                    processTopExp name exp ty;
                    Evalenv.bindValueEnv env (Ast.Var name);
                      debug_print 2
                      ("Evaluator.processTopDecl: adding type constraints " ^
                       "in assignment.");
                    ProcessConstraints.addTypeConstraints env name
                | Ast.Concretize (ctxt, exp) ->
                    debug_print 1
                      ("Evaluator.evaluate: concretizing " ^
                       AstPrint.pickle_exp (ref exp));
                    Oracle.concretize env ctxt (ref exp) ty;
                    Evalenv.bindValueEnv env (Ast.Var name) in
        (* Initialize the evaluation environment. *)
        Stats.reset_constraint_time;
        Stats.reset_exp_time;

        (* Process declarations. *)
        List.fold_left processTopDecl () topdecls;

        if !print_times
        then
          (print_endline ("Constraint processing time: " ^
                         string_of_float !(Stats.constraint_proc_time));
          print_endline ("Expression processing time: " ^
                          string_of_float !(Stats.exp_proc_time)))
        else ()
end
