(*
    File          : inputHandler.ml
    Last updated  : 10.7.09
    Maintainer    : jeanyang@csail.mit.edu

    This file contains the built-in procedure for reading from file.

    We read in from file, turning all non-integer values into symbolic AST
    expressions.

    (When do we apply constraints?)
 *)
open AstUtils
open Error
open Evalenv

module Ast = Ast.Ast

let missing_delimiter = "!!"

(* Count number of missing things. *)
let number_missing_lines = ref 0
let number_missing_entries = ref 0

let get_max_pkey (filename : string) (cur_max : int ref) : unit =
  let data_in = open_in filename in
  try
    while true; do
      let curLine = input_line data_in in
      try let curPkey = int_of_string (List.nth (Utils.split curLine) 0) in
        if curPkey > !cur_max
        then cur_max := curPkey
        else ()
      with _ -> ()
    done; ()
  with End_of_file -> close_in_noerr data_in

(* Converts a line from the input file to a record and adds it to the
 * current relation. *)
let convertLine
      (rel_name : Ast.varname)   (* Name of our relation.*)
      ({Ast.dataname = n; Ast.fieldinfo = rfields} : Ast.userty) (* Template *)
      (line : string)
      (accum : Ast.relation ref)
      : unit =
         let line_has_missing = ref false in
         (* Get each of the elements in the line. *)
         let line_elts = Utils.split line in
         let primary_key =
           try int_of_string (List.nth line_elts 0)
           with _ -> raise (Unexpected "Non-integer primary key") in
         let count = List.length (Ast.IntMap.find_all !accum primary_key) in
         let processValue (fname : int) (elt : string) : Ast.taggedExp ref =
           (try ref
                  ( ref (Ast.PrimExp (Ast.Integer (int_of_string elt)))
                    , false)
            with e ->
              let index_var_name =
                Ast.RecordIndexVar (primary_key, rel_name, fname, count) in
                line_has_missing := true;
                number_missing_entries := !number_missing_entries + 1;
                (* TODO *)
                ref (ref (Ast.Symbolic 0), true)) in
         (* If the line is well-formed, then process the line and add it
          * to the relation.  Otherwise, fail silently. TODO? *)
         let fields =
           try Array.init (List.length rfields)
                (fun i -> processValue i (List.nth line_elts i))
           with Failure _ -> raise (InputError "line length not as expected") in
           Ast.IntMap.add !accum primary_key (ref fields);
           if !line_has_missing
           then number_missing_lines := !number_missing_lines + 1
           else ()

(* Read from file based on information about file records. *)
let readFromFile
      (env : Evalenv.evalenv ref)
      (input_name : Ast.varname)
      (filename : string)
      : Ast.relation ref =
  (* Look up structure associated with input type. *)
  let input_recordty = match Evalenv.lookupTypeDeclByData !env input_name with
    | Some t -> t
    | None ->
        raise (Unexpected "input doesn't have record type")  in
  (* Read in thing from file *)
  let max_pkey = ref 0 in
  let _ = get_max_pkey filename max_pkey in
  let _ = max_pkey := 1 + !max_pkey in
  let data_in = open_in filename in
  let resultMap = ref (Ast.IntMap.create !max_pkey) in
    try
      (try
         while true; do
           let curLine = input_line data_in in
             convertLine input_name input_recordty curLine resultMap;
          done; resultMap
        with End_of_file ->
          print_endline ("Number lines with missing values: " ^
                         string_of_int !number_missing_lines);
          print_endline ("Number of missing values: " ^
                         string_of_int !number_missing_entries);
          resultMap)
    with e -> close_in_noerr data_in;
              raise e
