(* 
    Module      : driver.ml
    Maintainer  : jeanyang@csail.mit.edu
    Updated     : 10.26.10

    This contains the LogLog main loop.
*)
open AstUtils
open Error

module Ast = Ast.Ast;;
module BackendInterface = BackendInterface.BackendInterface;;
module Evalenv = Evalenv.Evalenv;;
module Evaluator = Evaluator.Evaluator;;

(* Runs the interpreter.
   Unless -debug is set to some level (0-n), the only thing we print is the
   cumulative sum total constraint solving time. *)
let run_interpreter (file_name : string) : unit =
  let i_start_time = Sys.time () in
  let _ = BackendInterface.initEnv () in
  (* 1. Open file. *)
  let cin = open_in file_name in
  (* 2. Lex. *)
  let lexbuf = Lexing.from_channel cin in
  (* 3. Parse. *)
  let topDecls = Parser.input Lexer.token lexbuf in
  let _ =
    if debug_level 2
      then AstPrint.print_topdecls topDecls
      else () in
  (* 4. Evaluate. *)
  let denv = ref Evalenv.empty_evalenv in
    Evaluator.evaluate denv topDecls;
    if debug_level 3
      then BackendInterface.printEnv ()
      else ();
    if debug_level 2
      then
        (Printf.printf "Driver.run_interpreter: solving for result.\n";
         BackendInterface.lookupVal (0);
         BackendInterface.printExp ())
      else ();
    let i_end_time = Sys.time () in
    print_endline
      ("Running time: " ^ string_of_float (i_end_time -. i_start_time));
    Time.print_time ()
