(* Header section *)
{
  open Printf
  open Parser
  open Lexing
  let incr_lineno lexbuf =
    let pos = lexbuf.lex_curr_p in
    lexbuf.lex_curr_p <- { pos with
      pos_lnum = pos.pos_lnum + 1;
      pos_bol = pos.pos_cnum;
    }
}

(*
    Definitions section
    Simplify scanner specification with expressions of the form
      let ident = regexp
*)
let digit = ['0'-'9']
let id = ['a'-'z'] ['a'-'z' '0'-'9' '/' '_']*
let constid = ['A'-'Z']*

(*
    Rules are of the form
      rule entrypoint [arg1 ... argn] = parse
        | pattern { action }
        | ...
        | pattern { action }
      and ...
*)
rule token = parse
  | "(*"       { comments 0 lexbuf; COMMENT }
  (* Top-level keywords *)
  | "function" { FUN }
  | "data"  { RELTYPE }
  | "with"     { WITH }
  (* Constraint syntax. *)
  | "concretize" { CONCRETIZE }
  (* Types *)
  | "int"      { INTTY }
  | "bool"     { BOOLTY }
  | "float"    { FLOATTY }
  | "string"   { STRINGTY }
  | "relation" { RELATION }
  | "record"   { RECORD }
  | "list"     { LIST }
  | "unit"     { UNITTY }
  (* Punctuation *)
  | ";" { SEMI }
  | "," { COMMA }
  | "." { PERIOD }
  | ":" { COLON }
  | "::" { DOUBLECOLON }
  | "\"" { QUOTE }
  | "?" { QMARK }
  | "_" { UNDERSCORE }
  (* Arithmetic *)
  | digit digit* as num
    { INTEGER (int_of_string num) }
  | '+'   { PLUS }
  | '-'   { MINUS }
  | '*'   { MULTIPLY }
  | '/'   { DIVIDE }
  | '^'   { CARET }
  (* Booleans *)
  | "true"  { TRUE }
  | "false" { FALSE }
  (* Keywords *)
  | "let" { LET }
  | "forall" { FORALL }
  | "in" { IN }
  | "if" { IF }
  | "then" { THEN }
  | "else" { ELSE }
  | "fold" { FOLD }
  | id as ident { IDENTIFIER ident }
  (* Whitespace *)
  | [' ' '\t']  { token lexbuf }
  | '\n'  { incr_lineno lexbuf;
            NEWLINE }
  (* Equality *)
  | "==" { DEQUALS }
  | "=" { EQUALS }
  | "!=" { NEQUALS }
  | "<" { LT }
  | "<=" { LTE }
  | ">" { GT }
  | ">=" { GTE }
  | "&&" { AND }
  | "||" { OR }
  (* Enclosing symbols *)
  | "(" { LPAREN }
  | ")" { RPAREN }
  | "{" { LBRACE }
  | "}" { RBRACE }
  | "[" { LBRACKET }
  | "]" { RBRACKET }
  | _ { token lexbuf }
  | eof   { EOF }
and comments level = parse
  | "*)" {
          if level = 0 then token lexbuf
          else comments (level-1) lexbuf
         }
  | "(*" {
          comments (level+1) lexbuf
         }
  | "\n" { incr_lineno lexbuf;
           comments level lexbuf }
  | _    { comments level lexbuf}
  | eof  { print_endline "Unclosed comment";
           raise End_of_file }

(* Trailer section *)
{

}

