(*
    Module        : oracle.ml
    Last updated  : 
    Maintainer    : jeanyang@csail.mit.edu
 *)
open AstUtils
open BackendInterface
open Error
open Evalenv

module Ast = Ast.Ast

module Oracle =
struct
  (*
   * This function takes a list of Yices expressions and binds the variables
   * back into the environment.  This happens by replacing the symbolic
   * nodes in the backend.
   *)
  let add_yout_to_env
        (env: Evalenv.evalenv ref)
        (y_out : Yices.yexp list)
        : unit =
    let add_to_env _accum yexp = match yexp with
      | Yices.YEqual (v, value) ->
          debug_print 2
            ("Oracle.add_yout_to_env: Adding " ^ v ^ " = " ^
             (string_of_int value) ^ " back to env");
          let cur_var = Var.str2var v in
            BackendInterface.concretizeSym cur_var value in
      List.fold_left add_to_env () y_out;
      debug_print 2 "Oracle.add_yout_to_env: Finished adding variables."

  (*
   * This function does the following:
   * 1) Writes the constraints to the file constraints.txt.
   * 2) Invokes the Yices on these constraints.
   * 3) Reads the output of Yices from standard output.
   * 4) Lexes and parses this output, returning it in a standard form.
   *)
  let run_yices () : Yices.yexp list =
    BackendInterface.writeAssertions ();
    (* Read output. *)
    let (begin_time : float) = Sys.time () in
    (* TODO: Call Yices from back-end. *)
    let oc_in, oc_out = Unix.open_process "yices -e < constraints.txt" in
    let input_of_inchannel = IO.input_channel oc_in in
    (* Deal with input channel *)
    let yicesbuff = IO.read_all input_of_inchannel in
    let _ = Unix.close_process (oc_in, oc_out) in
    let end_time = Sys.time () in
    let _ = if debug_level 2
    then print_endline yicesbuff
    else () in
    (* Lex and parse. *)
    let lexbuf = Lexing.from_string yicesbuff in
    try
      let yices_output = Yparser.input Ylexer.token lexbuf in
        Time.update_time (end_time -. begin_time);
      yices_output
    with _ -> [] 

  (*
   * Concretize an expression.  All variables that appear in the expression
   * need to be mapped to a concrete value.
   *)
  (* TODO: Add context. *)
  let concretize
        (env : Evalenv.evalenv ref) (e : Ast.exp ref) (t : Ast.typerep)
        : unit =
    EvalExp.evalExp env e t;
    BackendInterface.resolveWithContext 0
end
