(**
  Module      : Evalenv
  Maintainer  : jeanyang@csail.mit.edu
  Updated     : 9.28.09
*)
open AstUtils
open ConstraintUtils
open Error
open MapUtils

module Ast = Ast.Ast
module Constraint = Constraint.Constraint

(* TODO: change this to a reference list! *)
module ConstraintSet =
  Set.Make (struct type t = Constraint.cexp
              (* For now, compare references. *)
              (* TODO: Figuer out if this is what we want. *)
              let compare = ConstraintSimplify.cexp_eq
            end)


(** Module for evaluation environment. *)
module Evalenv =
struct
  
  (** For storing information about data values/user-defined types along with
      their associated constraints. *)
  type constraintenv = (Constraint.cexp list) ref
 
  (** Evaluation environment. *)
  type tyenv = Ast.var -> Ast.typerep
  type valenv = Ast.var -> Ast.taggedExp
  type typeconstraint = Ast.taggedExp
                        -> evalenv
                        -> constraintenv
  and evalenv = { typeconstraints    : typeconstraint StringMap.t;
                   (** Type-level constraints for corner-case values.
                       relation name |-> relation -> env -> env *)
                   tyenv              : tyenv;
                   (** variable name |-> type *)
                   tydefenv           : Ast.userty StringMap.t;
                   (** type name |-> type representation *)
                   functions          : Ast.fundecl StringMap.t;
                   (** function name |-> function def *)
                   valbindings     : valenv }

  (* Some defaults. *)
  let default_tyconstraint = fun _exp -> fun env -> ref []
  let empty_constraintenv = ref []
  let default_tyenv : tyenv = fun _ -> raise Not_found
  let default_valenv : valenv = fun _ -> raise Not_found
  let empty_evalenv = { typeconstraints = StringMap.empty
                      ; tyenv = default_tyenv
                      ; tydefenv = StringMap.empty
                      ; functions = StringMap.empty
                      ; valbindings = default_valenv }

  (*************************)
  (* Function declarations *)
  (*************************)
  let addFunDecl env fdecl =
    { typeconstraints = env.typeconstraints
    ; tyenv = env.tyenv
    ; tydefenv = env.tydefenv
    ; functions = StringMap.add fdecl.Ast.funname fdecl env.functions
    ; valbindings = env.valbindings }
  let lookupFunDecl env fdeclname : Ast.fundecl option =
    try Some (StringMap.find fdeclname env.functions)
    with Not_found -> None

  (* Manipulating the global constraint environment. *)
  let addConstraintEnv
        (cenv : constraintenv)
        (ce : Constraint.cexp)
        : constraintenv =
    match ce with
      | Constraint.CFalse -> cenv
      | _ -> ref ((ConstraintSimplify.csimplify ce)::(!cenv))
  let addConstraintListEnv
        (cenv : constraintenv)
        (cdecls : Constraint.cexp list)
        : constraintenv =
      List.fold_left
        (fun cur_cenv -> fun cur_ce -> addConstraintEnv cur_cenv cur_ce)
        cenv cdecls

    (* Binding a new value *)
  let bindValueEnv (env : evalenv)
        (name : Ast.var)
        ((dexp, symb) : Ast.taggedExp)
        : evalenv =
    if debug_level 2
    then print_endline
           ("Evalenv.bindValueEnv: Binding " ^ AstPrint.pickle_var name ^ ".")
    else ();
    (* Make new bindings. *)
    let makeNewBindings (curBindings : valenv) : valenv =
      fun v ->
        (* If the name matches, then return our thing. *)
        if AstEq.eq_vars v name == 0
        then (dexp, symb)
        else
          (* Otherwise there is this other case where we could be
           * indexing into our expression, which is a relation. *)
          match name, v with
            | Ast.Var r, Ast.RecordIndexVar (r', f, i) ->
                if String.compare r r' == 0
                then
                  (match dexp with
                     | Ast.PrimExp (Ast.RelationExp (_n, _ty, relation)) ->
                         let field_exp =
                           RelationUtils.get_index_field f i relation in
                           (match field_exp with
                              | Some e -> e
                              | None -> raise Not_found)
                     | _ ->
                         raise (Unexpected
                                  ("Looking up " ^ r ^ ", but it is " ^
                                   "not a relation.")))
                    else curBindings v
                  (* Otherwise look in the rest of our environment. *)
              | _ -> curBindings v in

    { typeconstraints = env.typeconstraints
    ; tyenv = env.tyenv
    ; tydefenv = env.tydefenv
    ; functions = env.functions
    ; valbindings = makeNewBindings env.valbindings }

  (* Expressions in the variable environment. *)
  (* Value lookup--treat record field lookup in a special way. *)
  let lookupValueEnv (env : evalenv) (name : Ast.var)
        : Ast.taggedExp option =
    try match name with
      | Ast.FieldIndexVar (r, f) ->
          (match env.valbindings (Ast.Var r) with
             | (Ast.PrimExp (Ast.RecordExp (_n, _ty, record)), _sym) ->
                 Some (StringMap.find f record)
             | _ -> None)
      | _ -> Some (env.valbindings name)
    with Not_found -> None

  (* addConstraint called by addTypeDecl *)
  let addTypeLevelConstraint
        (env : evalenv) (type_name : string) 
        (constraint_decls : Ast.constraintbody list)
        : evalenv =
    (* Some utility functions. *)
    let mkSingleNoPred (ce : Constraint.cexp) =
      ref [(Constraint.CGuard (Constraint.CTrue, ce, Constraint.CTrue))] in
    (* Gets the cross-product pairs of things in a binary operation. *)
    let getPairs (cset1 : constraintenv) (cset2 : constraintenv)
          : (Constraint.cexp * Constraint.cexp * Constraint.cexp) list =
      let addPairs
            (cset : constraintenv)
            (accum :
               (Constraint.cexp * Constraint.cexp * Constraint.cexp) list)
            (cur_cexp : Constraint.cexp) =
        let addElts
              (pred1 : Constraint.cexp) (ce1 : Constraint.cexp)
              (accum :
                (Constraint.cexp * Constraint.cexp * Constraint.cexp) list)
              (cexp2 : Constraint.cexp ) =
          match cexp2 with
            | Constraint.CGuard (pred2, ce2, Constraint.CTrue) ->
                (ConstraintSimplify.conjoin pred1 pred2, ce1, ce2)::accum
            | _ -> raise (Unexpected "constraint format 2") in
      match cur_cexp with
        | Constraint.CGuard (pred, ce1, Constraint.CTrue) ->
            List.fold_left (addElts pred ce1) accum !cset
        | _ -> raise (Unexpected "constraint format") in
      List.fold_left (addPairs cset2) [] !cset1 in
    (*********************************************************)
    (* Processing application of AST constraint expressions. *)
    (* INVARIANT: All constraints coming out of this guy should have format
     * CGuard (_, _, _) *)
    let rec processCexp (env : evalenv) (cexp : Ast.constraintexp)
          : constraintenv =
      (* Process variable lookup. *)
      let processVarExp vexp =
        match lookupValueEnv env vexp with
          | Some (e, sym) ->
              mkSingleNoPred (ConstraintConverter.ast2constraint e)
          | None ->
              raise (Unbound ( AstPrint.pickle_var vexp
                             , "Evalenv.addTypeLevelConstraint")) in
      (* Process "match" expression. *)
    match cexp with
      | Ast.ACBoolop (op, ce1, ce2) ->
          let ce1c's = processCexp env ce1 in
          let ce2c's = processCexp env ce2 in
          let pairs = getPairs ce1c's ce2c's in
          let addBoolopConstraint cset (pred, ce1, ce2) =
            ref ((Constraint.CGuard
                 ( pred
                 , Constraint.CBoolop
                     (ConstraintConverter.convert_boolop op, ce1, ce2)
                 , Constraint.CTrue ))::!cset) in
            List.fold_left addBoolopConstraint empty_constraintenv pairs
      | Ast.ACArithop (op, ce1, ce2) ->
         let ce1c's = processCexp env ce1 in
          let ce2c's = processCexp env ce2 in
          let pairs = getPairs ce1c's ce2c's in
          let addArithopConstraint cset (pred, ce1, ce2) =
            ref ((Constraint.CGuard
                 ( pred
                 , Constraint.CArithop
                     (ConstraintConverter.convert_arithop op, ce1, ce2)
                 , Constraint.CTrue ))::!cset) in
            List.fold_left addArithopConstraint empty_constraintenv pairs 
      | Ast.ACPrim i -> mkSingleNoPred (Constraint.CInt i)
      | Ast.ACVarExp vexp -> processVarExp vexp
      (* Go through and see what matches. *)
      | Ast.ACMatch (rvar, paramvar, exp, field) ->
          (* Gets matches from the current element. *)
          let get_matches n ty _idx cur_elt cur_matches =
            let addMatch accum cur_ce =
              match cur_ce with
                | Constraint.CGuard (_, Constraint.CFalse, _) -> accum
                | Constraint.CGuard (_, cond, _) ->
                    let newc =
                      ConstraintConverter.ast2constraint
                        (fst (StringMap.find field cur_elt)) in
                    let guardedc =
                      Constraint.CGuard (cond, newc, Constraint.CTrue) in
                      ref (guardedc::!accum)
                | _ -> raise (Unexpected "bad thing") in
            let env' = bindValueEnv env (Ast.Var paramvar)
                          ( (Ast.PrimExp (Ast.RecordExp (n, ty, cur_elt)))
                          , true) in
              (* If ce' isn't the set containing false--or the empty set,
               * we should add the converted version of the current constraint
               * to the accumulated list of constraints. *)
              List.fold_left addMatch cur_matches !(processCexp env' exp) in
          (match lookupValueEnv env rvar with
             | Some (e, _sym) ->
                 (match e with
                      (* If we have a record, want to go through and 
                       * discover everything that matches. *)
                    | Ast.PrimExp (Ast.RelationExp (n, ty, elts)) ->
                        IntMap.fold (get_matches n ty) elts empty_constraintenv
                    | _ -> raise (Unimplemented "match on non-relation"))
             | None ->
                 raise (Unbound ("match var", "addTypeLevelConstraints")))
      (* Processing a forall. *)
      (* Go through and apply constraints for all things in list/relation. *)
      | Ast.ConstraintForall (v, relvar, ce) ->
          (* Processing relation records. *)
          let processRecord n ty idx (elt : Ast.datafields) accum =
            let env' = bindValueEnv
                         env (Ast.Var v)
                         ((Ast.PrimExp (Ast.RecordExp (n, ty, elt))), true) in
              (* TODO: Put whether record is actually symbolic. *)
            let constraints = processCexp env' ce in
              addConstraintListEnv accum !constraints in
          (* Processing list elements. *)
          let processListElt accum elt =
            let env' = bindValueEnv env (Ast.Var v) (elt, true) in
            let constraints = processCexp env' ce in
            addConstraintListEnv accum !constraints in
          (match lookupValueEnv env relvar with
             | Some (e, sym) ->
                 (* For now, we are applying constraints no matter if the
                  * thing is symbolic or not. *)
                  (match e with
                     | Ast.PrimExp (Ast.RelationExp (n, ty, rel)) ->
                        IntMap.fold (processRecord n ty) rel empty_constraintenv
                     | Ast.PrimExp (Ast.ListExp (ty, lst)) ->
                        List.fold_left processListElt empty_constraintenv lst
                     | _ -> raise (Unimplemented "Other cases"))
             | None ->
                raise
                  (Unbound
                     ("Constraint var", "Evalenv.addTypeLevelConstraint"))) in
    (* For each record, update the record based on the constraints *)
    let processConstraint
          (e : Ast.taggedExp)
          (env : evalenv)
          (cset : constraintenv)
          (cnst : Ast.constraintbody)
          : constraintenv =
      let processAssumption
            (cset : constraintenv)
            ((v, cexp) : string * Ast.constraintexp) =
        let env' = bindValueEnv env (Ast.Var v) e in
          addConstraintListEnv
            cset !(processCexp env' cexp) in
      match cnst with
          | Ast.ConstraintAssume assumptions ->
              List.fold_left processAssumption cset assumptions in
      (* First try looking for a type unit with this name. *)
      try
        (* Current thing mapping fields to constraints. *)
        let (cur_constraint : typeconstraint) =
          try StringMap.find type_name env.typeconstraints
          with Not_found -> default_tyconstraint in
        (* New thing mapping fields to constraints. *)
        let (new_decls : typeconstraint) =
          fun e -> fun cur_env ->
            let new_constraints =
              List.fold_left
                (processConstraint e cur_env)
                (cur_constraint e cur_env)
                constraint_decls in
              new_constraints in
          if debug_level 3
          then print_endline
                 ("EvalExp.evalExp: Adding type constraints for " ^
                  "relation " ^ type_name ^ ".")
          else ();
          { typeconstraints =
              StringMap.add type_name new_decls env.typeconstraints
          ; tyenv = env.tyenv
          ; tydefenv = env.tydefenv
          ; functions = env.functions
          ; valbindings = env.valbindings }
      with Not_found -> raise (Unbound (type_name, "Evalenv.addConstraint"))

  (* Adds a type declaration to the environment. *)
  let addTypeDecl
        env (tdecl : Ast.userty) (constraints : Ast.constraintbody list) =
    let ty_name = tdecl.Ast.dataname in
    let _ = if debug_level 3
    then Printf.printf "Evalenv.addTypeDecl: %s\n" ty_name
    else () in
    let new_env =
      { typeconstraints = env.typeconstraints
      ; tyenv = env.tyenv
      ; tydefenv = StringMap.add ty_name tdecl env.tydefenv
      ; functions = env.functions
      ; valbindings = env.valbindings } in
      addTypeLevelConstraint new_env ty_name constraints
  let lookupTypeDeclByData env (data_name : string) : Ast.userty option =
    try match env.tyenv (Ast.Var data_name) with
      | Ast.RelationType type_name ->
          Some (StringMap.find type_name env.tydefenv)
      | _ -> None
    with Not_found -> None

  (*********************)
  (* Type environment. *)
  (*********************)
  let bindTypeEnv
        (env : evalenv)
        (name : string)
        (ty : Ast.typerep)
        : evalenv =
    if debug_level 2
    then print_endline
           ("Evalenv.bindTypeEnv: Binding " ^ name ^ " in type env.")
    else ();
    { typeconstraints = env.typeconstraints
    ; tyenv =
        (fun v ->
          if AstEq.eq_vars (Ast.Var name) v == 0
               then ty
                else env.tyenv v)
    ; tydefenv = env.tydefenv
    ; functions = env.functions
    ; valbindings = env.valbindings }
  (* Looks up the type of a variable. *)
  let lookupType (env : evalenv)
                 (data_name : Ast.var)
                : Ast.typerep option =
    match data_name with
    | Ast.Var v ->
        (try Some (env.tyenv data_name)
         with Not_found -> None)
    | Ast.RecordIndexVar (r, f, i) ->
        (try let reltype = env.tyenv (Ast.Var r) in
              match reltype with
              | Ast.RelationType rtype ->
                  Some Ast.IntType (*  TODO *)
              | _ -> raise (Unexpected "Relation doesn't have relation type.")
         with Not_found -> None)
    | Ast.FieldIndexVar (r, f) ->
         (try let reltype = env.tyenv (Ast.Var r) in
              match reltype with
              | Ast.RecordType rtype ->
                  Some Ast.IntType (*  TODO *)
              | _ -> raise (Unexpected "Relation doesn't have relation type.")
         with Not_found -> None)

  (* Look up constraints associated with a type. *)
  let lookupTypeConstraints
        (env : evalenv)
        (tyname : string)
        : typeconstraint option =
    try Some (StringMap.find tyname env.typeconstraints)
    with Not_found -> None

  (* Applies type-level constraints to a given expression. *)
  let addTypeConstraints
        (env : evalenv)
        (cenv : constraintenv)
        (name : string)
        ((e, sym) : Ast.taggedExp)
        : constraintenv =
    if sym
    then
      match lookupTypeDeclByData env name with
        | Some ty ->
            let ty_name = ty.Ast.dataname in
              (match lookupTypeConstraints env ty_name with
                 | Some tc ->
                     let newConstraints = tc (e, sym) env in
                       ref (List.append !newConstraints !cenv)
                 | None ->
                  raise (Unbound (ty_name, "Evalenv.lookupTypeConstraints")))
      | None -> cenv
    else cenv

  let lookupRelationGlobalEnv env name =
  match lookupValueEnv env name with
  | Some (Ast.PrimExp (Ast.RelationExp (_n, _ty, r)), _sym) -> Some r
  | _ -> None

  end
