/* file: parser.mly */
/* SketchPADS parser */

/* Header section */
%{
open Error
open Lexing
open List
open Printf
%}

%token NEWLINE COMMENT EOF
%token LPAREN RPAREN LBRACE RBRACE LBRACKET RBRACKET
%token SEMI PERIOD COMMA COLON DOUBLECOLON QUOTE
%token UNDERSCORE
%token QMARK
%token EQUALS NEQUALS DEQUALS
%token CONCRETIZE
/* Types */
%token INTTY FLOATTY STRINGTY BOOLTY UNITTY
/* Boolean */
%token TRUE FALSE
%token LT LTE GT GTE
%token AND OR
/* Arithmetic */
%token PLUS MINUS MULTIPLY DIVIDE NEG CARET
/* Keywords */
%token LET
%token RELATION RECORD LIST
%token FORALL IN
%token IF THEN ELSE
%token RELTYPE FUN FOLD
%token WITH
/* Basics */
%token <float> FLOAT
%token <int> INTEGER
%token <string> IDENTIFIER

%left PLUS MINUS
%left MULTIPLY DIVIDE
%left NEG   /* negation -- unary minus */
%right CARET  /* exponentiation */

%start input
%type <Ast.Ast.topdecl list> input
%type <Ast.Ast.topdecl> topdecl
%type <Ast.Ast.exp> exp
%type <Ast.Ast.typerep> typerep
%type <Ast.Ast.varname * Ast.Ast.typerep> typedvardecl
%type <string list> usertyrep

/* Grammar follows */
%%
input: topdecls EOF { $1 }

topdecls: optionalvwsp topdecl vwhitespace { [$2] }
        | topdecls topdecl optionalvwsp { List.append $1 [$2] }
;
vwhitespace: /* empty */ { }
           | NEWLINE  { }
           | COMMENT { }
           | vwhitespace NEWLINE { }
           | vwhitespace COMMENT { }
optionalvwsp: /* empty */ { }
            | vwhitespace { }

topdecl: typedecl { $1 }
       | fundecl { $1 }
       | assigndecl { $1 }
       | error { let start_pos = Parsing.symbol_start_pos () in
                 let end_pos = Parsing.symbol_start_pos () in
                 printf "Parsing error\n";
                 printf "%d.%d - %d.%d: Error\n"
                    start_pos.pos_lnum
                    (start_pos.pos_cnum - start_pos.pos_bol)
                    end_pos.pos_lnum
                    (end_pos.pos_cnum - end_pos.pos_bol);
                 raise Parse_error }
;

typedvardecl : IDENTIFIER COLON typerep
          { (Varenv.map_var $1, $3) }

typerep: INTTY      { Ast.Ast.IntType }
       | BOOLTY     { Ast.Ast.BoolType }
       | UNITTY     { Ast.Ast.UnitType }
       | RELATION IDENTIFIER
       { let type_as_int =
            try Varenv.lookup_type $2
            with Not_found -> raise (ParseUndefinedType $2) in
          Ast.Ast.RelationType type_as_int }
      | RECORD IDENTIFIER
       { let type_as_int =
            try Varenv.lookup_type $2
            with Not_found -> raise (ParseUndefinedType $2) in
          Ast.Ast.RecordType type_as_int }
     | LIST typerep { Ast.Ast.ListType $2 }
usertyrep: IDENTIFIER { [$1] }

/*********************/
/* Data declarations */
/*********************/
typedecl : RELTYPE IDENTIFIER LBRACE databody RBRACE optionalvwsp WITH
           LPAREN vardecl RPAREN
           LBRACE optionalvwsp constraints optionalvwsp RBRACE
            { let data_name = Varenv.map_type $2 in
              let typeinfo = { Ast.Ast.dataname = data_name
                             ; Ast.Ast.fieldinfo = $4 } in
                Varenv.reset_field_counter ();
                Ast.Ast.TypeDecl (typeinfo, $9, $13) }
         | RELTYPE IDENTIFIER LBRACE databody RBRACE optionalvwsp
            { let data_name = Varenv.map_type $2 in
              let typeinfo = { Ast.Ast.dataname = data_name
                             ; Ast.Ast.fieldinfo = $4 } in
                Varenv.reset_field_counter ();
                Ast.Ast.TypeDecl (typeinfo, 0, []) }
databody: IDENTIFIER COLON typerep
          { let field_int = Varenv.map_field $1 in
            [(field_int, $3)] }
        | databody optionalvwsp SEMI IDENTIFIER COLON typerep
          { let field_int = Varenv.map_field $4 in
              List.append $1 [(field_int, $6)]  }

/* Function declarations */
fundecl: FUN IDENTIFIER COLON typerep LPAREN funargs RPAREN EQUALS
           optionalvwsp LBRACE optionalvwsp exp optionalvwsp RBRACE
          { let var_as_int = Varenv.map_var $2 in
            (* TODO: Make sure the body is not evaluated lazily. *)
            (* TODO: Unmap function arguments. *)
            Ast.Ast.FunDecl ( (var_as_int, $4)
                            , { Ast.Ast.funargs = $6;
                                Ast.Ast.funbody = ref $12 } ) }
funargs : typedvardecl { [$1] }
        | funargs COMMA typedvardecl { List.append $1 [$3] }

/* TODO: Other right-hand sides. */
datarhs : exp { Ast.Ast.RhsExp $1 }
        | CONCRETIZE INTEGER exp { Ast.Ast.Concretize ($2, $3) }
assigndecl: LET IDENTIFIER COLON typerep EQUALS datarhs
              { let var_as_int = Varenv.map_var $2 in
                  Ast.Ast.AssignDecl (var_as_int, $4, $6) }
          | LET UNDERSCORE COLON typerep EQUALS datarhs
              { let var_as_int = Varenv.map_var "_" in
                  Ast.Ast.AssignDecl (var_as_int, $4, $6) }

/***************/
/* Constraint.Constraints */
/***************/
constraints : cdecl { [$1] }
            | constraints PERIOD optionalvwsp cdecl { List.append $1 [$4] }
cdecl : exp { Ast.Ast.CExp (ref $1) }
      | FORALL vardecl IN varexp optionalvwsp LPAREN exp RPAREN
        optionalvwsp LPAREN cdecl RPAREN
        { Ast.Ast.ConstraintForall ($2, $4, ref $7, $11) }

vardecl : IDENTIFIER { Varenv.map_var $1 }
varexp : IDENTIFIER
          { let var_as_int =
              try Varenv.lookup_var $1
              with Not_found -> raise (ParseUndefinedVar $1) in
            Ast.Ast.Var var_as_int }
       | IDENTIFIER LBRACKET exp RBRACKET PERIOD IDENTIFIER
        { let r =
            try Varenv.lookup_var $1
            with Not_found -> raise (ParseUndefinedVar $1) in
          let field =
            try Varenv.lookup_var $6
            with Not_found -> raise (ParseUndefinedVar $6) in
            Ast.Ast.RecordIndexVar (ref $3, r, field) }
       | IDENTIFIER PERIOD IDENTIFIER
        { let v =
            try Varenv.lookup_var $1
            with Not_found -> raise (ParseUndefinedVar $1) in
          let f =
            try Varenv.lookup_var $3
            with Not_found -> raise (ParseUndefinedVar $3) in
            Ast.Ast.FieldIndexVar (v, f) }
      | QMARK varexp { Ast.Ast.HiddenVarAccess $2 }
str :
  | IDENTIFIER { $1 }
  | str IDENTIFIER { $1 ^ " " ^ $2 }
prim :
  INTEGER { Ast.Ast.PrimExp (Ast.Ast.Integer $1) }  
  | TRUE { Ast.Ast.PrimExp (Ast.Ast.Bool true) }
  | FALSE { Ast.Ast.PrimExp (Ast.Ast.Bool false) }
  | QUOTE str QUOTE { Ast.Ast.PrimExp (Ast.Ast.String $2) }
exp: prim      { $1 }
  | IF LPAREN exp RPAREN optionalvwsp THEN exp optionalvwsp ELSE exp
    { Ast.Ast.CondExp (ref $3, ref $7, ref $10) }
  | LET IDENTIFIER COLON typerep EQUALS exp IN exp
    { let v = Varenv.map_var $2 in
        Ast.Ast.LetExp (v, $4, ref $6, ref $8) }
  | LET UNDERSCORE COLON typerep EQUALS exp IN exp
    { let v = Varenv.map_var "_" in
        Ast.Ast.LetExp (v, $4, ref $6, ref $8) }
  | FOLD IDENTIFIER exp IDENTIFIER
    { (* let f =
        try Varenv.lookup_var $2
        with Not_found -> raise (ParseUndefinedVar $2) in *)
      let e =
        try Varenv.lookup_var $4
        with Not_found -> raise (ParseUndefinedVar $4) in
        Ast.Ast.FoldExp ($2, ref $3, e) }
  | IDENTIFIER LPAREN args RPAREN
    { (* let f_int =
        try Varenv.lookup_var $1
        with Not_found -> raise (ParseUndefinedVar $1) in *)
    Ast.Ast.FunAppExp ($1, (List.map (fun e -> ref e) $3)) }
  | varexp { Ast.Ast.VarExp $1 }
  | exp PLUS exp
  { Ast.Ast.Arithop (Ast.Ast.Plus, ref $1, ref $3) }
  | exp MINUS exp
  { Ast.Ast.Arithop (Ast.Ast.Minus, ref $1, ref $3) }
  | exp MULTIPLY exp
  { Ast.Ast.Arithop (Ast.Ast.Times, ref $1, ref $3) }
  | exp DIVIDE exp
  { Ast.Ast.Arithop (Ast.Ast.Divide, ref $1, ref $3) }
  | exp DEQUALS exp
  { Ast.Ast.Boolop (Ast.Ast.Equiv, ref $1, ref $3) }
  | exp NEQUALS exp
  { Ast.Ast.Boolop (Ast.Ast.Nequiv, ref $1, ref $3) }
  | exp LT exp
  { Ast.Ast.Boolop (Ast.Ast.Lt, ref $1, ref $3) }
  | exp GT exp
  { Ast.Ast.Boolop (Ast.Ast.Gt, ref $1, ref $3) }
  | exp AND exp
  { Ast.Ast.Boolop (Ast.Ast.And, ref $1, ref $3) }
  | exp OR exp
  { Ast.Ast.Boolop (Ast.Ast.Or, ref $1, ref $3) }
  | LPAREN exp RPAREN { $2 }
;
args:
  | exp { [$1] }
  | args COMMA exp { List.append $1 [$3] }

%%

