(**
  Module      : Evalenv
  Maintainer  : jeanyang@csail.mit.edu
  Updated     : 3.15.10
*)
open AstBackend
open AstUtils
open BackendInterface
open Error
open Evalenv

module Ast = Ast.Ast

module ConstraintSet =
  Set.Make (struct type t = (Ast.exp ref)
              let compare c1 c2 =
                if AstEq.eq_exps c1 c2 then 1 else 0 (* TODO *)
            end)


type constraintenv = ConstraintSet.t
let print_constraintenv cenv : unit =
  ConstraintSet.iter
    (fun c ->
       print_endline (AstPrint.pickle_exp c))
      !cenv

  (* Manipulating the global constraint environment. *)
  let addConstraintEnv
        (cenv : constraintenv ref)
        (ce : Ast.exp ref)
        : unit =
    cenv := ConstraintSet.add ce !cenv
  let addConstraintsEnv
        (cenv : constraintenv ref)
        (new_constraints : constraintenv ref)
        : unit =
    ConstraintSet.iter
      (fun newc -> addConstraintEnv cenv newc) !new_constraints

  (***************************************)
  (* addConstraint called by addTypeDecl *)
  let addTypeLevelConstraints
        (env : Evalenv.evalenv ref) (constraint_decls : Ast.cdecl list)
        : unit =
    debug_print 2 "ProcessConstraints.addTypeLevelConstraints: Adding";
    (*********************************************************)
    (* Processing application of AST constraint expressions. *)
    (* INVARIANT: All constraints coming out of this guy should have format
     * CGuard (_, _, _) *)
    let conjoin c1 c2 = ref (Ast.Boolop (Ast.And, c1, c2)) in
    let constraintSet = ref ConstraintSet.empty in
      (* Processing a forall. *)
      (* Go through and apply constraints for all things in list/relation. *)
    let rec processCDecl
          (decl : Ast.cdecl) (condition : Ast.exp ref) (ignore_concrete : bool)
          : bool =
      match decl with
      (* If we have an expression, we evaluate the condition.  If the
       * condition is not false, then we evaluate the record.  If either the
       * constraint has a symbolic component or we are not ignoring concrete,
       * then we add the constraint to the set. *)
      | Ast.CExp cexp ->
          debug_print 2 "ProcessConstraints: processing cexp";
          debug_print 2 (AstPrint.pickle_exp condition);
          (* EvalExp.evalExp env condition Ast.UnknownType; *)
          let (cond_not_false, cond_symbolic) =
            EvalSym.shallowEval env condition in
            (* try (BackendInterface.getConcreteBool (), true)
            with Failure _ -> (false, true) in *)
            if (cond_not_false && ignore_concrete) || (not ignore_concrete)
            then
              (if cond_symbolic
               then
                 (print_endline ("evaluating condition: " ^ AstPrint.pickle_exp condition);
                  EvalExp.evalExp env condition Ast.UnknownType;
                  EvalExp.evalExp env cexp Ast.UnknownType;
                  AstBackend.mkAstExpNode (ref (Ast.PrimExp (Ast.Integer 1)));
                  BackendInterface.createCondEid ();
                  BackendInterface.assertExp ();
                  true)
               else
                 (if EvalSym.evalSym env cexp
                  then
                    (EvalExp.evalExp env cexp Ast.UnknownType;
                     BackendInterface.assertExp ();
                     true)
                  else false))
            else false
        | Ast.ConstraintForall (v, relvar, cond, cdecl) ->
            Evalenv.lookupValueEnv env relvar;
            debug_print 2
              "ProcessConstraints: processing forall";
            BackendInterface.relIterator ();
          let get_outer_rec =
            ref (BackendInterface.nextRecord ()) in
          while !get_outer_rec do
            BackendInterface.bindVal v;
            (* Process the constraint. *)
            processCDecl cdecl (conjoin condition cond) ignore_concrete;
            get_outer_rec := BackendInterface.nextRecord ()
          done;
          debug_print 3 "ProcessExp: done with ConstraintForall";
          true
      (* Matching on primary key. *)
      | Ast.ConstraintForMatch (v, relvar, v', relvar', cond, cexp) ->
          debug_print 3 "ProcessExp: ConstraintForMatch";
          (* Get the outer relation iterator. *)
          Evalenv.lookupValueEnv env (Ast.Var relvar);
          let _outer_rel = BackendInterface.relIterator () in
          let get_outer_rec = ref (BackendInterface.nextRecord ()) in
          while !get_outer_rec do
            BackendInterface.bindVal v;
            (* Get the inner relation iterator. *)
            BackendInterface.lookupFieldIndex v 0;
            BackendInterface.relPkeyIterator relvar';
            let get_inner_rec =
              ref (BackendInterface.nextRecord ()) in
            while !get_inner_rec do
              BackendInterface.bindVal v';
              processCDecl (Ast.CExp cexp) cond true;
              get_inner_rec := BackendInterface.nextRecord ();
            done;
            get_outer_rec := BackendInterface.nextRecord ()
          done;
          true
      | Ast.ConstraintExistsMatch
          (v, relvar, v', relvar', applyvar, cond, cexp) ->
          debug_print 3 "ProcessExp: ConstraintForMatch";
          (* Get the outer relation iterator. *)
          let _ = Evalenv.lookupValueEnv env (Ast.Var relvar) in
          let _outer_rel = BackendInterface.relIterator () in
          let get_outer_rec = ref (BackendInterface.nextRecord ()) in
          while !get_outer_rec do
            BackendInterface.bindVal v;
            (* Get the inner relation iterator. *)
            BackendInterface.lookupFieldIndex v 0;
            BackendInterface.relPkeyIterator relvar';
            let get_inner_rec = ref (BackendInterface.nextRecord ()) in
            let found_match = ref false in
            while (!get_inner_rec)
            do
              BackendInterface.bindVal v';
              (* Process the declaration only if the variable we care about is
               * not symbolic. *)
              (match applyvar with
                 | Ast.FieldIndexVar (var, field) ->
                    BackendInterface.lookupFieldIndex var field;
                    if BackendInterface.isSym ()
                      then ()
                      else
                        found_match := processCDecl (Ast.CExp cexp) cond true
                 | _ -> raise (Unexpected "non field var"));
              get_inner_rec := BackendInterface.nextRecord ();
            done;
            (* When inner loop is done, add things to the constraint
             * environment. *)
            get_outer_rec := BackendInterface.nextRecord ()
          done;
          true
    in
    (* For each record, update the record based on the constraints *)
    let processConstraint (cdecl : Ast.cdecl) : unit  =
      let _ = processCDecl cdecl (Ast.ref_true) true in
        () in
      (* First try looking for a type unit with this name. *)
        List.iter processConstraint constraint_decls;
        if debug_level 2
        then (ConstraintSet.iter
                (fun c ->
                   print_endline (AstPrint.pickle_exp c))
                !constraintSet;
              ())
          else ()

  (* Applies type-level constraints to a given expression.
   * *)
  let addTypeConstraints
        (env : Evalenv.evalenv ref) (name : Ast.varname)
        : unit =
    Evalenv.lookupValueEnv env (Ast.Var name);
    let sym =
      BackendInterface.dup ();
      BackendInterface.isSym () in
    debug_print 2
      ("ProcessConstraints.addTypeConstraints: adding constraints for " ^
       string_of_int name);
    (* TODO: Take this out again. *)
    if sym
    then
      (match Evalenv.lookupTypeDeclByData env name with
        | Some ty ->
            let ty_name = ty.Ast.dataname in
              (try
                 let (asname, typeconstraints) =
                   VarEnv.find !(!env.Evalenv.typeconstraints) ty_name in
                 let _ =
                    BackendInterface.dup ();
                    Evalenv.bindValueEnv env (Ast.Var asname) in
                 (* Update the time here. *)
                 let start_time = Sys.time () in
                 let _ = addTypeLevelConstraints env typeconstraints in
                 let end_time = Sys.time () in
                   Stats.log_constraint_time (end_time -. start_time)
               with Not_found ->
                 debug_print 3
                   ("WARNING: Did not get constraints for " ^
                    string_of_int name ^ " because couldn't look up type name");
                 ())
        | None ->
            print_endline ("WARNING: Did not get constraints for " ^
                           (string_of_int name) ^ " because couldn't get " ^
                           "type description."))
  else ();
  BackendInterface.pop ()
